# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# This Source Code Form is "Incompatible With Secondary Licenses", as
# defined by the Mozilla Public License, v. 2.0.

package Bugzilla::Auth::Verify::RedHat;

use 5.10.1;
use strict;
use warnings;

use parent qw(Bugzilla::Auth::Verify);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::Constants;
use Bugzilla::Token;
use Bugzilla::Util;
use Bugzilla::User;
use Authen::Radius;

use List::Util qw(any);
use Data::Dumper;

# Bug 1262651 - Radius authentication for admin users.
#
# The purpose of this is to force users with administrative permissions to log
# in using RADIUS (pin+otp) authentication.
#
# This isn't enforced on RPC logins due to practical concerns.  This should be
# reviewed when bugzilla auth tokens are introduced.

sub check_credentials {
  my ($self, $params) = @_;

  my $username = $params->{username};
  my $user     = Bugzilla::User->new({name => $username});

  return {failure => AUTH_NO_SUCH_USER} unless $user;

  my $group_str = Bugzilla->params->{'rh_radius_groups'} // ''; # List of groups that require RADIUS auth.
  my @groups    = split(/,/, $group_str);
  my $in_sensitive_group = any { $user->in_group($_) } @groups;

  if (
    (Bugzilla->params->{'RADIUS_secret'} ne '')    # If we are using radius
    && (Bugzilla->usage_mode == USAGE_MODE_BROWSER) # and is a web interface login attempt
    && $in_sensitive_group
    )    # and they are a group that requires RADIUS auth
  {
    ## RED HAT EXTENSION 1657869
    # Then we need to enforce RADIUS authentication
    my $res = $self->check_credentials_radius($params);

    # If raidus login failed to validate then check local passwd
    if ($res->{failure} && $res->{failure} == AUTH_RH_RADIUS_LOGINFAILED) {
      my $res2 = $self->check_credentials_password($params);

      # If local login failed bail with local error
      return ($res2) if ($res2->{failure});
    }

    return ($res);
  }

  # Otherwise, use password auth
  return $self->check_credentials_password($params);
}

sub check_credentials_radius {
  my ($self, $params) = @_;
  my $dbh            = Bugzilla->dbh;
  my $address_suffix = Bugzilla->params->{'RADIUS_email_suffix'};
  my $username       = $params->{username};


  # If we're using RADIUS_email_suffix, we may need to cut it off from
  # the login name.
  if ($address_suffix) {
    $username =~ s/\Q$address_suffix\E$//i;
  }

  # Create RADIUS object.
  my $radius = new Authen::Radius(
    Host   => Bugzilla->params->{'RADIUS_server'},
    Secret => Bugzilla->params->{'RADIUS_secret'}
    )
    || return {
    failure => AUTH_ERROR,
    error   => 'radius_preparation_error',
    details => {errstr => Authen::Radius::strerror()}
    };

  # Check the password.
  unless ($radius->check_pwd(
    $username, $params->{password}, Bugzilla->params->{'RADIUS_NAS_IP'} || undef
  ))
  {
    Bugzilla->logger->info(
      "Radius Auth failure: " . $radius->get_error . " - " . $radius->strerror);
    return {failure => AUTH_RH_RADIUS_LOGINFAILED};
  }

  # Build the user account's e-mail address.
  $params->{bz_username} = $username . $address_suffix;

  return $params;
}

sub check_credentials_password {
  my ($self, $login_data) = @_;
  my $dbh = Bugzilla->dbh;

  my $username = $login_data->{username};
  my $user     = new Bugzilla::User({name => $username});

  return {failure => AUTH_NO_SUCH_USER} unless $user;

  $login_data->{user}        = $user;
  $login_data->{bz_username} = $user->login;

  if ($user->account_is_locked_out) {
    return {failure => AUTH_LOCKOUT, user => $user};
  }

  my $password              = $login_data->{password};
  my $real_password_crypted = $user->cryptpassword;

  # Using the internal crypted password as the salt,
  # crypt the password the user entered.
  my $entered_password_crypted = bz_crypt($password, $real_password_crypted);

  if ($entered_password_crypted ne $real_password_crypted) {

    # Record the login failure
    $user->note_login_failure();

    # Immediately check if we are locked out
    if ($user->account_is_locked_out) {
      return {failure => AUTH_LOCKOUT, user => $user, just_locked_out => 1};
    }

    return {
      failure       => AUTH_LOGINFAILED,
      failure_count => scalar(@{$user->account_ip_login_failures}),
    };
  }

  # Force the user to change their password if it does not meet the current
  # criteria. This should usually only happen if the criteria has changed.
  if ( Bugzilla->usage_mode == USAGE_MODE_BROWSER
    && Bugzilla->params->{password_check_on_login})
  {
    my $check = validate_password_check($password);
    if ($check) {
      return {
        failure    => AUTH_ERROR,
        user_error => $check,
        details    => {locked_user => $user}
      };
    }
  }

  # The user's credentials are okay, so delete any outstanding
  # password tokens or login failures they may have generated.
  Bugzilla::Token::DeletePasswordTokens($user->id, "user_logged_in");
  $user->clear_login_failures();

  my $update_password = 0;

  # If their old password was using crypt() or some different hash
  # than we're using now, convert the stored password to using
  # whatever hashing system we're using now.
  my $current_algorithm = PASSWORD_DIGEST_ALGORITHM;
  $update_password = 1 if ($real_password_crypted !~ /{\Q$current_algorithm\E}$/);

  # If their old password was using a different length salt than what
  # we're using now, update the password to use the new salt length.
  if ($real_password_crypted =~ /^([^,]+),/) {
    $update_password = 1 if (length($1) != PASSWORD_SALT_LENGTH);
  }

  # If needed, update the user's password.
  if ($update_password) {

    # We can't call $user->set_password because we don't want the password
    # complexity rules to apply here.
    $user->{cryptpassword} = bz_crypt($password);
    $user->update();
  }

  return $login_data;
}

sub change_password {
  my ($self, $user, $password) = @_;
  my $dbh           = Bugzilla->dbh;
  my $cryptpassword = bz_crypt($password);
  $dbh->do("UPDATE profiles SET cryptpassword = ? WHERE userid = ?",
    undef, $cryptpassword, $user->id);
  Bugzilla->memcached->clear({table => 'profiles', id => $user->id});
}

1;
