# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::AuthJWT;

use 5.10.1;
use strict;
use warnings;

use parent qw(Bugzilla::Extension);

# This code for this is in ../extensions/AuthJWT/lib/Util.pm
use Bugzilla::Extension::AuthJWT::Util;
use Bugzilla::Extension::AuthJWT::Source;

our $VERSION = '0.01';

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  $schema->{authjwt_source} = {
    FIELDS => [
      id       => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      kid      => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      cert     => {TYPE => 'LONGTEXT',     NOTNULL => 1},
      comment  => {TYPE => 'LONGTEXT',     NOTNULL => 1, DEFAULT => "''"},
      isactive => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 'TRUE'},
    ],
    INDEXES => [authjwt_source_kid_idx => {FIELDS => [qw(kid)], TYPE => 'UNIQUE'},],
  };

  # authjwt groups, specifies who is NOT allowed to use it!
  $schema->{authjwt_groups} = {
    FIELDS => [
      authjwt_source_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'authjwt_source', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      authjwt_authjwt_groups_idx =>
        {FIELDS => [qw(authjwt_source_id group_id)], TYPE => 'UNIQUE'},
    ],
  };

  return;
}

sub auth_login_methods {
  my ($self, $args) = @_;
  my $modules = $args->{'modules'};
  if (exists $modules->{'AuthJWT'}) {
    $modules->{'AuthJWT'} = 'Bugzilla/Extension/AuthJWT/Login.pm';
  }

  return;
}

sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};
  my $vars = $args->{vars};
  my $cgi  = Bugzilla->cgi;
  my $p    = $cgi->Vars;

  # If we aren't viewing a AuthJWT page, we don't need to do anything
  return unless ($page =~ m{^authjwt/});

  if ($page eq 'authjwt/settings.html') {
    my $action = $cgi->param('action') || "";
    my $params;
    foreach my $key (keys(%$p)) {
      next if ($key eq 'id');
      $params->{$key} = $p->{$key};
    }
    delete($params->{action});

    if ($action eq 'create') {
      $params->{kid} = delete($params->{source_kid});
      my $groups = delete $params->{groups};
      my $source = Bugzilla::Extension::AuthJWT::Source->create($params);
      $source->set_groups($groups);
      $source->update();
    }
    elsif ($action eq 'save') {
      $params->{isactive} = 0 unless defined $params->{isactive};
      my $id = delete($params->{source_id});
      $params->{kid} = delete($params->{source_kid});
      my $groups = delete $params->{groups};
      my $source = Bugzilla::Extension::AuthJWT::Source->new($id);
      $source->set_all($params);
      $source->set_groups($groups);
      $source->update();
    }
    elsif ($action eq 'delete') {
      my $id     = delete($params->{source_id});
      my $source = Bugzilla::Extension::AuthJWT::Source->new($id);
      $source->remove_from_db();
    }

    $vars->{'groups'} = [Bugzilla::Group->get_all];
    $args->{vars}->{sources} = Bugzilla::Extension::AuthJWT::Source->match();
    $args->{vars}->{'doc_section'}
      = 'extensions/AuthJWT/index-admin.html#bugzilla-extension-authjwt-identity-providers';

  }
  else {
    ThrowUserError('authjwt_no_such_page');
  }

  return;
}

__PACKAGE__->NAME;

__END__

=head1 Name

Bugzilla::Extension::AuthJWT - JWT authentication support for Bugzilla

=head1 Version

Version 0.01

=head1 Description

This module will allow you to set up your Bugzilla server as an SP and configure
multiple Sources to authenticate against.

