# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2013 Jolla Ltd.
# Contact: Pami Ketolainen <pami.ketolainen@jollamobile.com>

package Bugzilla::Extension::BayotBase::Util;
use strict;
use warnings;
use 5.10.1;

use Bugzilla::Error;
use open ':encoding(utf8)';

use base qw(Exporter);

our @EXPORT = qw(
  cache_base_dir
  cache_timestamp
  cache_ts_file
  cache_user_file
  get_field_defs
  clean_params
);

use Bugzilla::Constants;
use Bugzilla::Component;
use Bugzilla::Keyword;
use Bugzilla::Milestone;
use Bugzilla::Status;
use Bugzilla::Util;

use Scalar::Util qw(blessed);
use List::MoreUtils qw(none);

use constant PRODUCT_SPECIFIC =>
  {version => 1, target_milestone => 1, component => 1,};

use constant FIELD_OVERRIDES => {
  alias       => {type => 1, is_on_bug_entry => 1,},
  assigned_to => {
    type            => 11,    # USER
    is_on_bug_entry => 1,
  },
  blocked =>
    {name => 'blocks', type => 6, multivalue => 1, is_on_bug_entry => 1,},
  cc => {type => 11, multivalue => 1, is_on_bug_entry => 1,},
  classification => {immutable       => 1, is_on_bug_entry => 0,},
  component      => {is_on_bug_entry => 1, value_field     => 'product',},
  longdesc    => {name => 'comment', type => 4, is_on_bug_entry => 1,},
  creation_ts => {
    name            => 'creation_time',
    immutable       => 1,
    is_on_bug_entry => 1,

    # It's a datetime, but as immutable the type doesn't matter
  },
  reporter => {name            => 'creator', type => 11, immutable => 1,},
  deadline => {is_on_bug_entry => 1,},
  dependson =>
    {name => 'depends_on', type => 6, is_on_bug_entry => 1, multivalue => 1,},
  estimated_time => {is_on_bug_entry => 1,},
  bug_group      => {
    name            => 'groups',
    type            => 0,
    multivalue      => 1,
    is_on_bug_entry => 1,

    # Should this be multiselect?
    # Would require resolving the available groups
  },
  bug_id            => {name => 'id', type => 6, immutable => 1,},
  cclist_accessible => {
    name            => 'is_cc_accessible',
    type            => 12,                   # BOOLEAN
    is_on_bug_entry => 1,
  },
  reporter_accessible =>
    {name => 'is_creator_accessible', type => 12, is_on_bug_entry => 1,},
  keywords => {multivalue => 1,                  is_on_bug_entry => 1,},
  delta_ts => {name       => 'last_change_time', immutable       => 1,},
  op_sys   => {
    is_on_bug_entry => 1,
    is_mandatory    => Bugzilla->params->{defaultopsys} ? 0 : 1,
  },
  rep_platform => {
    name            => 'platform',
    is_on_bug_entry => 1,
    is_mandatory    => Bugzilla->params->{defaultplatform} ? 0 : 1,
  },
  priority       => {is_on_bug_entry  => 1,},
  product        => {is_on_bug_entry  => 1,},
  qa_contact     => {type             => 11, is_on_bug_entry => 1,},
  remaining_time => {is_on_bug_entry  => 1,},
  resolution     => {visibility_field => 'status',},
  see_also       => {multivalue       => 1,},
  bug_severity   => {name             => 'severity', is_on_bug_entry => 1,},
  bug_status => {name => 'status', is_on_bug_entry => 1, value_field => 'status'},
  short_desc => {name => 'summary', is_on_bug_entry => 1,},
  target_milestone =>
    {is_on_bug_entry => 1, type => 2, value_field => 'product',},
  bug_file_loc => {name => 'url', is_on_bug_entry => 1,},
  version      => {
    type            => 2,
    is_mandatory    => 1,
    is_on_bug_entry => 1,
    value_field     => 'product',
  },
  status_whiteboard => {name => 'whiteboard', is_on_bug_entry => 1,},
  work_time         => {},
  rh_sub_components =>
    {is_on_bug_entry => 1, value_field => 'component', type => 2,},
};

use constant EXTRA_FIELDS => (

# Not in fields() result
  {
    id               => 0,
    name             => 'dupe_of',
    internal_name    => 'dup_id',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    type             => 6,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'is_confirmed',
    internal_name    => undef,
    display_name     => 'Confirmed',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 12,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'is_open',
    internal_name    => undef,
    display_name     => 'Open',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 12,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'update_token',
    internal_name    => undef,
    display_name     => 'Update Token',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 0,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'actual_time',
    internal_name    => undef,
    display_name     => 'Actual Hours',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 0,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
);

## REDHAT EXTENSION START 1775514
# Add relationships to user specific JS to allow client side field mappings

sub _add_related_value {
  my ($rel_fields, $changed_field, $cid, $affected_field, $aid) = @_;

  $rel_fields->{$changed_field} = {}
    unless (exists $rel_fields->{$changed_field});
  $rel_fields->{$changed_field}->{'values'} = {}
    unless (exists $rel_fields->{$changed_field}->{'values'});
  $rel_fields->{$changed_field}->{'values'}->{$cid} = {}
    unless (exists $rel_fields->{$changed_field}->{'values'}->{$cid});
  $rel_fields->{$changed_field}->{'values'}->{$cid}->{$affected_field} = []
    unless (
    exists $rel_fields->{$changed_field}->{'values'}->{$cid}->{$affected_field});

  push(
    @{$rel_fields->{$changed_field}->{'values'}->{$cid}->{$affected_field}},
    $aid
  );

  return;
}

sub _add_related_field {
  my ($rel_fields, $changed_field, $affected_field, $ids) = @_;

  $rel_fields->{$changed_field} = {}
    unless (exists $rel_fields->{$changed_field});
  $rel_fields->{$changed_field}->{'fields'} = {}
    unless (exists $rel_fields->{$changed_field}->{'fields'});

  $rel_fields->{$changed_field}->{'fields'}->{$affected_field} = $ids;

  return;
}
## REDHAT EXTENSION END 1775514

sub _generate_field_defs {
  my ($user, $user_values) = @_;
  $user ||= Bugzilla->user;

  my %fields;
  my %related_fields;
  my %all_classifications;
  my $field_descs = template_var('field_descs');
  my %selectable_products
    = map { $_->id => $_ } @{$user->get_selectable_products};

  foreach my $field (@{Bugzilla->fields}) {
    next unless $field->custom or defined FIELD_OVERRIDES->{$field->name};

    my @values;
    if ($field->name eq 'product') {
      ## REDHAT EXTENSION START 1775514
      foreach my $product (values %selectable_products) {
        $all_classifications{$product->name} = $product->classification->name;
      }
      ## REDHAT EXTENSION END 1775514

      ## REDHAT EXTENSION START 1298042
      #foreach my $product (values %selectable_products) {
      #    push @values, {
      #        name     => $product->name,
      #        sort_key => 0,
      #        visibility_values => [],
      #        is_default => 0,
      #        allows_unconfirmed => $product->allows_unconfirmed ? 1 : 0,
      #    };
      #}
      ## REDHAT EXTENSION END 1298042
    }
    elsif ($field->name eq 'keywords') {
      foreach my $keyword (Bugzilla::Keyword->get_all) {
        push @values,
          {
          name              => $keyword->name,
          sort_key          => 0,
          visibility_values => [],
          is_default        => 0,
          };
      }
    }
    elsif (PRODUCT_SPECIFIC->{$field->name}) {
      my @list;
      if ($field->name eq 'version') {
        @list = Bugzilla::Version->get_all;
      }
      elsif ($field->name eq 'component') {
        ## REDHAT EXTENSION START 1298042
        #@list = Bugzilla::Component->get_all;
        ## REDHAT EXTENSION END 1298042
      }
      else {
        @list = Bugzilla::Milestone->get_all;
      }
      foreach my $value (@list) {
        my $product = $selectable_products{$value->product_id};
        next unless defined $product;

        ## REDHAT EXTENSION START 1317777
        next if (exists($value->{isactive}) && (!$value->{isactive}));
        ## REDHAT EXTENSION END 1317777

        my $sortkey = $value->can('sortkey') ? $value->sortkey : 0;
        push @values,
          {
          name              => $value->name,
          sort_key          => $sortkey || 0,
          visibility_values => [$product->name],
          is_default        => 0,
          };

        _add_related_value(\%related_fields, 'product', $product->name, $field->name,
          $value->name);
      }
    }
    elsif ($field->name eq 'bug_status') {
      my %statuses;

      # BUGBUG this limits selections to valid status changes, noice
      # we should use it ...
      foreach my $value (@{$field->legal_values}) {
        foreach (@{$value->can_change_to}) {
          $statuses{$_->name} ||= {visibility_values => []};
          push(@{$statuses{$_->name}->{visibility_values}}, $value->name);
        }
        $statuses{$value->name} ||= {visibility_values => []};
        $statuses{$value->name}->{name}       = $value->name;
        $statuses{$value->name}->{sort_key}   = $value->sortkey;
        $statuses{$value->name}->{is_default} = 0;
      }
      @values = values %statuses;
    }
    elsif ($field->is_select) {
      foreach my $value (@{$field->legal_values}) {
        my $vis_val = $value->visibility_value;
        push @values,
          {
          name              => $value->name,
          sort_key          => $value->sortkey || 0,
          visibility_values => [defined $vis_val ? $vis_val->name : ()],
          is_default        => $value->is_default,
          };

        _add_related_value(\%related_fields, $field->value_field->name,
          $vis_val->name, $field->name, $value->name)
          if ($vis_val);
      }
    }
    @values
      = sort { $a->{sort_key} <=> $b->{sort_key} || $a->{name} cmp $b->{name} }
      @values;

    my $visibility_field
      = $field->visibility_field ? $field->visibility_field->name : undef;
    my $value_field = $field->value_field ? $field->value_field->name : undef;
    my @visibility_values;
    if ($field->name eq 'resolution') {
      @visibility_values = map { $_->name } Bugzilla::Status::closed_bug_statuses();
      $visibility_field  = 'bug_status';
    }
    else {
      @visibility_values = map { $_->name } @{$field->visibility_values};
    }

    my $name = FIELD_OVERRIDES->{$field->name}->{name} || $field->name;
    my $display_name = $field_descs->{$field->name} || $field->description;

    $fields{$name} = {
      name              => $name,
      internal_name     => $field->name,
      display_name      => $display_name,
      id                => scalar $field->id,
      is_custom         => $field->custom ? 1 : 0,
      values            => \@values,
      visibility_values => \@visibility_values,

      # OVERRIDABLE
      type         => FIELD_OVERRIDES->{$field->name}->{type} || $field->type,
      is_mandatory => FIELD_OVERRIDES->{$field->name}->{is_mandatory}
        || $field->is_mandatory ? 1 : 0,
      is_on_bug_entry => FIELD_OVERRIDES->{$field->name}->{is_on_bug_entry}
        || $field->enter_bug ? 1 : 0,
      value_field => FIELD_OVERRIDES->{$field->name}->{value_field} || $value_field,
      visibility_field => FIELD_OVERRIDES->{$field->name}->{visibility_field}
        || $visibility_field,

      # EXTRA
      immutable  => FIELD_OVERRIDES->{$field->name}->{immutable} || 0,
      multivalue => FIELD_OVERRIDES->{$field->name}->{multivalue}
        || $field->type == FIELD_TYPE_MULTI_SELECT,
    };

    _add_related_field(\%related_fields, $visibility_field, $field->name,
      \@visibility_values)
      if (@visibility_values);    # REDHAT EXTENSION 1775514
  }
  foreach (EXTRA_FIELDS) {
    $fields{$_->{name}} = $_;
  }

  my %teams;
  foreach my $team (@{$user->visible_agile_teams}) {
    my $new_team = {name => $team->name, id => $team->id,};

    my $pools;
    foreach my $pool (@{$team->pools(1)}) {
      my $new_pool = {
        name       => $pool->name,
        id         => $pool->id,
        is_sprint  => $pool->is_sprint,
        is_backlog => $pool->is_backlog,
      };
      $pools->{$pool->name} = $new_pool;
    }

    $new_team->{pools} = $pools;
    $teams{$team->name} = $new_team;
  }

  $user_values->{fields}              = \%fields;
  $user_values->{relations}           = \%related_fields;
  $user_values->{all_classifications} = \%all_classifications;
  $user_values->{teams}               = \%teams;

  return;    # REDHAT EXTENSION 1775514
}

sub cache_base_dir {
  return bz_locations()->{'datadir'} . '/extensions/bayotbase_cache/';
}

sub cache_ts_file {
  my $filename = cache_base_dir() . 'lastupdate';

  # Make sure timestamp file exists
  -e $filename
    || open(my $fh, ">", $filename)
    || warn "Failed to create $filename: $!";
  return $filename;
}

sub cache_user_file {
  my $uid  = shift;
  my $name = shift // '.json';    # REDHAT EXTENSION 1775514

  $uid = $uid->id if (blessed $uid);
  $uid ||= Bugzilla->user->id || 0;
  return cache_base_dir() . "$uid$name";
}

sub cache_timestamp { return (stat cache_ts_file())[9] || time; }

sub get_field_defs {
  my %args     = @_;
  my $cache_ts = cache_timestamp();
  my $as_json  = $args{as_json} || 0;
  my $user     = $args{user} || Bugzilla->user;

  my $json_string     = "{}";
  my $user_cache_file = cache_user_file($user);
  my $user_cache_ts   = (stat($user_cache_file))[9] || 0;
  my $pretty          = Bugzilla->params->{bb_use_debug_js};
  my $json            = JSON->new->allow_blessed->allow_nonref;

## REDHAT EXTENSION START 1775514
  my %user_files = (
    fields              => '',
    relations           => '.rel',
    all_classifications => '.ac',
    teams               => '.teams',
  );

  my %user_values;
## REDHAT EXTENSION END 1775514

  my ($fh);
  if ($cache_ts > $user_cache_ts) {
    _generate_field_defs($user, \%user_values);
    $json_string = $json->pretty($pretty)->encode($user_values{fields});

    foreach my $type (keys %user_files) {
      my $file_name = cache_user_file($user, $user_files{$type});
      open($fh, ">", $file_name) and do {
        print $fh $json->encode($user_values{$type});
        }
        or warn "Failed to write $file_name: $!";
      close $fh if defined $fh;
    }
  }
  else {
    eval {
      foreach my $type (keys %user_files) {
        my $file_name = cache_user_file($user, $user_files{$type});
        open($fh, "<", $file_name) and do {
          my $input = join('', <$fh>);
          $user_values{$type} = $json->decode($input);
          $json_string = $input if ($type eq 'fields');
          }
          or warn "Failed to write $file_name: $!";
        close $fh if defined $fh;
      }
    };
  }

  if ($@) {
    _generate_field_defs($user, \%user_values);
    $json_string = $json->pretty($pretty)->encode($user_values{fields}) if $as_json;
  }

  return (
    $as_json ? $json_string : $user_values{fields},
    $json->pretty($pretty)->encode($user_values{relations}),
    $user_values{all_classifications},
    $json->pretty($pretty)->canonical->encode($user_values{teams}),
  );
}

sub clean_params {
  my $url   = shift;

  my $query = $url;
  my $start = '';

  # peel off start of URL if it's included
  if ($url =~ /^(.*\?)(.*)$/) {
    $start = $1;
    $query = $2;
  }
  my @params = qw/
    Bugzilla_api_key
    Bugzilla_api_token
    Bugzilla_login
    Bugzilla_login_token
    Bugzilla_password
    GoAheadAndLogIn
    api_key
    api_token
    password
    target
    token
    /;

  my @parts = grep {
    my $param = $_;
    none { $param =~ /^$_=/ } @params
  } split('&', $query);

  return $start . join('&', @parts);
}

1;
