# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;
use 5.10.1;

package Bugzilla::Extension::DependentProducts::Bug;

use Scalar::Util qw(blessed);

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::Util;
use Bugzilla::Error;

use base qw(Bugzilla::Object);

use Bugzilla::Extension::DependentProducts::Util;

###############################
####    Initialization     ####
###############################

use constant DB_TABLE   => 'bug_dependent_products';
use constant DB_COLUMNS => qw(
  id
  value
  bug_id
  enabled_by
);

use constant UPDATE_COLUMNS => qw(
  value
  bug_id
  enabled_by
);

use constant VALIDATORS => {
  value      => \&_check_value,
  bug_id     => \&_check_bug_id,
  enabled_by => \&_check_enabled_by,
};

###############################
####      Validators       ####
###############################

sub _check_value {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->value : 0;
  $value = trim($value);

  my $object = Bugzilla::Product->check({id => $value});
  if ((!$old_id || $object->id != $old_id) && !$object->is_active) {
## Breaks upgrade as it prevents old links being updated
## TODO not sure this is actually required
##        ThrowUserError('value_inactive', {class => ref($object), value => $object->name});
  }

  # BUGBUG needs to check Product::dependent_product as well

  return ($object->id);
}

sub _check_bug_id {
  my ($invocant, $bug_id) = @_;
  $bug_id = trim($bug_id);

  my $bug_obj = Bugzilla::Bug->check($bug_id);
  $bug_obj->check_is_visible();

  my $privs;
  if (!$bug_obj->check_can_change_field('dependent_product', 0, 1, \$privs)) {
    ThrowUserError('illegal_change',
      {field => 'dependent_product', privs => $privs});
  }

  return ($bug_id);
}

sub _check_enabled_by {
  my ($invocant, $value) = @_;
  my $old_id = blessed($invocant) ? $invocant->enabled_by : 0;
  $value = trim($value);

  my $object = Bugzilla::User->check({id => $value});
  if (!$old_id || $object->id != $old_id) {
    ThrowUserError('value_inactive',
      {class => ref($object), value => $object->name});
  }
  return ($object->id);
}

###############################
####       Methods         ####
###############################

sub set_value      { $_[0]->set('value',      $_[1]); return; }
sub set_bug_id     { $_[0]->set('bug_id',     $_[1]); return; }
sub set_enabled_by { $_[0]->set('enabled_by', $_[1]); return; }

sub delete {
  my ($self, $args) = @_;
  my $query = "DELETE FROM bug_dependent_products WHERE bug_id = ? and value = ?";
  Bugzilla->dbh->do($query, undef, $self->bug_id, $self->value);

  return;
}

###############################
####      Accessors        ####
###############################

sub value      { return ($_[0]->{value}); }
sub bug_id     { return ($_[0]->{bug_id}); }
sub enabled_by { return ($_[0]->{enabled_by}); }

1;

__END__

=head1 Description

Bugzilla::Extension::DependentProducts::Bug -This modules represents the relationship
between a bug and the dependent products that are affected by it.

=head1 Fields

=over 4

=item value

The dependent product id.

=item bug_id

The bug id.

=item enabled_by

The user who set this relationship on a bug.

=back

=head1 Methods

These methods allow you to set the specified field.

=over 4

=item set_value

=item set_bug_id

=item set_enabled_by

=item delete

Delete the relationship between a bug and a dependent product.

=back

=head1 Accessors

These methods allow you to get the specified field.

=over 4

=item value

=item bug_id

=item enabled_by

=back

=cut

