# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Bug;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Object);

use Bugzilla::Bug;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Constants;

use Bugzilla::Extension::ExternalBugs::Type;
use Bugzilla::Extension::ExternalBugs::Util qw(perform_substs);

use Date::Parse;
use Scalar::Util qw(blessed);

# Time in seconds that we need to refresh the bug data from the external tracker
use constant BUG_REFRESH_AGE => 360;
use constant AUDIT_CREATES   => 0;

###############################
####    Initialization     ####
###############################

use constant DB_COLUMNS => qw(
  id
  bug_id
  ext_bz_id
  ext_bz_bug_id
  ext_status
  ext_priority
  ext_description
  ext_last_updated
);

use constant DB_TABLE   => 'ext_bz_bug_map';
use constant LIST_ORDER => 'bug_id';
use constant ID_FIELD   => 'id';

use constant REQUIRED_CREATE_FIELDS => qw(
  bug_id
  ext_bz_id
  ext_bz_bug_id
  ext_last_updated
);

use constant VALIDATOR_DEPENDENCIES => {ext_bz_bug_id => ['ext_bz_id']};

use constant VALIDATORS => {
  bug_id        => \&_check_bug_id,
  ext_bz_id     => \&_check_ext_bz_id,
  ext_bz_bug_id => \&_check_ext_bz_bug_id,
};

use constant UPDATE_COLUMNS => qw(
  ext_description
  ext_last_updated
  ext_priority
  ext_status
);

###############################

## REDHAT EXTENSION BEGIN 821293
sub new {
  my $invocant = shift;
  my $class    = ref($invocant) || $invocant;
  my $param    = shift;

  my $self = $class->SUPER::new($param);

  if (  defined $self
    and $self->ext_bz_id
    and not Bugzilla->user->can_see_ext_bz($self->ext_bz_id))
  {
    # User cannot see the tracker, so treat it like it doesn't exist
    return;
  }

  return $self;
}

# Modify the where value to restrict visibility
sub _do_list_select {
  my $invocant = shift;

  my $group_list = Bugzilla->user->groups_as_string;
  if ($_[0]) {
    $_[0] .= " AND ";
  }

  $_[0] .= "
ext_bz_id IN (
    SELECT DISTINCT id
    FROM external_bugzilla b
        LEFT JOIN ext_bz_group_map m ON (m.ext_bz_id = b.id)
    WHERE m.group_id IS NULL OR m.group_id IN ($group_list)
)";

  return $invocant->SUPER::_do_list_select(@_);
}
## REDHAT EXTENSION END 821293

sub create {
  my ($class, $args) = @_;
  my $dbh = Bugzilla->dbh;

  $dbh->bz_start_transaction();
  $class->check_required_create_fields($args);
  my $params = $class->run_create_validators($args);

  # Set the last updated date.
  $params->{'ext_last_updated'} = Bugzilla->dbh->selectrow_array('SELECT NOW()');

  my $ext_bug = $class->insert_create_data($params);

  # Update the status, and then do any type specific activiation
  $ext_bug->update_ext_info(1);
  $ext_bug->type->activiate($ext_bug);

  $dbh->bz_commit_transaction();

  return $ext_bug;
}

sub remove_from_db {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  ## REDHAT EXTENSION BEGIN 821293
  if ($self->ext_bz_id and not Bugzilla->user->can_see_ext_bz($self->ext_bz_id)) {

    # User cannot see the tracker, so give them an error
    ThrowUserError('external_bug_id_not_valid');
  }
  ## REDHAT EXTENSION END 821293

  $dbh->bz_start_transaction();

  $self->SUPER::remove_from_db();

  # We may need to remove a link to the external tracker
  my $tracker = Bugzilla::Extension::ExternalBugs::Type->new($self->ext_bz_id);
  $tracker->deactivate($self);

  $dbh->bz_commit_transaction();

  return;
}

#####################################################################
# Validators
#####################################################################

sub _check_bug_id {
  my ($invocant, $bug_id) = @_;

  my $bug_obj = Bugzilla::Bug->check($bug_id);
  $bug_obj->check_is_visible();

  my $privs;
  if (!$bug_obj->check_can_change_field('external_bugs', 0, 1, \$privs)) {
    ThrowUserError('illegal_change', {field => 'external_bugs', privs => $privs});
  }

  return $bug_id;
}

sub _check_ext_bz_id {
  my ($invocant, $ext_bz_id) = @_;
  my $ext_bz_obj = Bugzilla::Extension::ExternalBugs::Type->new($ext_bz_id);
  $ext_bz_obj
    || ThrowUserError('ext_bz_invalid_ext_bz_id', {badvalue => $ext_bz_id});
  return $ext_bz_id;
}

sub _check_ext_bz_bug_id {
  my ($invocant, $ext_bz_bug_id, undef, $params) = @_;
  $ext_bz_bug_id = trim($ext_bz_bug_id);

  $ext_bz_bug_id
    || ThrowUserError('ext_bz_invalid_ext_bug_id', {badvalue => $ext_bz_bug_id});

  my $ext_bz_obj
    = blessed($invocant)
    ? $invocant->type
    : Bugzilla::Extension::ExternalBugs::Type->new($params->{ext_bz_id});

  if (my $regex = $ext_bz_obj->regex_pattern) {
    unless ($ext_bz_bug_id =~ qr{$regex}) {
      ThrowUserError('ext_bz_invalid_regex',
        {badvalue => $ext_bz_bug_id, description => $ext_bz_obj->regex_description});
    }
  }

  # SFDC issues need to be padded with leading zeros
  if (  $ext_bz_obj->regex
    and $ext_bz_obj->regex eq 'SFDC Case'
    and length($ext_bz_bug_id) < 8)
  {
    my $x = 8 - length($ext_bz_bug_id);
    return ('0' x $x) . $ext_bz_bug_id;
  }

  return $ext_bz_bug_id;
}

###############################
####     Accessors       ######
###############################

sub id               { return $_[0]->{'id'}; }
sub bug_id           { return $_[0]->{'bug_id'}; }
sub ext_bz_id        { return $_[0]->{'ext_bz_id'}; }
sub ext_bz_bug_id    { return $_[0]->{'ext_bz_bug_id'}; }
sub ext_status       { return $_[0]->{'ext_status'} || "None"; }
sub ext_priority     { return $_[0]->{'ext_priority'} || "None"; }
sub ext_description  { return $_[0]->{'ext_description'} || "None"; }
sub ext_last_updated { return $_[0]->{'ext_last_updated'} || "Never"; }

sub set_ext_description { $_[0]->set('ext_description', $_[1]); return; }
sub set_ext_priority    { $_[0]->set('ext_priority',    $_[1]); return; }
sub set_ext_status      { $_[0]->set('ext_status',      $_[1]); return; }

sub type {
  my ($self) = @_;
  return $self->{type} if blessed($self->{type});
  $self->{type} = Bugzilla::Extension::ExternalBugs::Type->new($self->ext_bz_id);
  return $self->{type};
}

sub full_url {
  my ($self) = @_;
  return $self->{full_url} if $self->{full_url};
  $self->{full_url}
    = perform_substs($self->type->full_url, {id => $self->ext_bz_bug_id});

## REDHAT ENTENSION BEGIN 653915 1510439
  if ($self->type->sfdc_hostname) {
    if (Bugzilla->user->settings->{view_customer_cases_on_sfdc}{value} eq 'as_sfdc')
    {
      $self->{full_url}
        = 'https://'
        . $self->type->sfdc_hostname
        . '/apex/Case_View?sbstr='
        . $self->ext_bz_bug_id;
    }
    elsif (Bugzilla->user->settings->{view_customer_cases_on_sfdc}{value} eq
      'as_ascension' and Bugzilla->params->{ascension_url})
    {
      $self->{full_url} = Bugzilla->params->{ascension_url} . $self->ext_bz_bug_id;
    }
  }

## REDHAT EXTENSION END 653915 1510439

  return $self->{full_url};
}

sub choices {
  my ($self) = @_;
  return $self->{choices} if $self->{choices};
  $self->{choices} = Bugzilla::Extension::ExternalBugs::Type->match();
  return $self->{choices};
}

sub needs_refresh {
  my ($self) = @_;

  # don't forget time is in milli-seconds
  if (
    $self->type->can_get()
    && ($self->ext_last_updated eq 'Never'
      || (time() - str2time($self->ext_last_updated) > (BUG_REFRESH_AGE * 1000)))
    )
  {
    return 1;
  }
  return 0;
}

###############################
####     Subroutines     ######
###############################

sub update_ext_info {
  my ($self, $ok_to_skip) = @_;
  my $dbh = Bugzilla->dbh;

  my $tracker = Bugzilla::Extension::ExternalBugs::Type->new($self->ext_bz_id);
  my $bug_data;
  if ($tracker->can_get()) {
    my $mode = Bugzilla->error_mode();
    Bugzilla->error_mode(ERROR_MODE_DIE) if ($mode == ERROR_MODE_WEBPAGE);
    eval { $bug_data = $tracker->get_data($self); };
    Bugzilla->error_mode($mode);
  }
  elsif ($ok_to_skip) {

    # Sync attempt was not user generated.
    return;
  }
  else {
    ThrowUserError("ext_bz_cant_sync_tracker",
      {site_name => $tracker->description});
  }

  my $timestamp = $dbh->selectrow_array(q{SELECT NOW()});

  if ($bug_data) {
    my @fields = qw /status description priority/;
    foreach my $field (@fields) {
      if ($bug_data->{$field}) {
        trick_taint($bug_data->{$field});
        $self->{"ext_$field"} = $bug_data->{$field};
      }
    }

    $self->{'ext_last_updated'} = $timestamp;

    ## REDHAT EXTENSION BEGIN 1206001
    # Use accessors so we get defaults
    $dbh->do(
      "UPDATE ext_bz_bug_map SET ext_status = ?, ext_description = ?, "
        . "ext_priority = ?, ext_last_updated = ? WHERE id = ?",
      undef,
      $self->ext_status,
      $self->ext_description,
      $self->ext_priority,
      $timestamp,
      $self->id
    );
    ## REDHAT EXTENSION End 1206001
  }
  else {
    # update the time so we don't spam the server
    $self->{'ext_last_updated'} = $timestamp;
    $dbh->do("UPDATE ext_bz_bug_map SET ext_last_updated = ? WHERE id = ?",
      undef, $timestamp, $self->id);
  }

  return;
}

sub update {
  my $self = shift;

  $self->{'ext_last_updated'} = Bugzilla->dbh->selectrow_array('SELECT NOW()');

  $self->SUPER::update(@_);

  return;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs::Bug - External Bugs Class.

=head1 SYNOPSIS

    use Bugzilla::Extension::ExternalBugs::Bug;

    my $external_bugs = Bugzilla::Extension::ExternalBugs::Bug->match({ bug_id => $bug_id });

    # Public methods
    my $ext_bug = Bugzilla::Extension::ExternalBugs::Bug->create({ bug_id => $bug_id,
                                              ext_bz_id => $ext_bz_id,
                                              ext_bz_bug_id => $ext_bz_bug_id });
    $ext_bug->remove_from_db();


=head1 DESCRIPTION

ExternalBugs::Bug represents an external bug object. It is an implementation
of L<Bugzilla::Object>, and thus provides all methods that
L<Bugzilla::Object> provides.

The methods that are specific to C<Bugzilla::Extension::ExternalBugs::Bug> are listed
below.

=head1 METHODS

=over

=item C<Need Documentation>

=back

=head1 SEE ALSO

L<Bugzilla::Object>

=cut
