# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Object);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(diff_arrays trick_taint trim);

use File::Basename;
use JSON;

###############################
####    Initialization     ####
###############################

use constant DB_COLUMNS => qw(
  id
  url
  description
  full_url
  type
  can_get
  can_send
  must_send
  send_once
  ext_bz_regex_id
  matchrx
  format
);

use constant DB_TABLE   => 'external_bugzilla';
use constant LIST_ORDER => 'description';
use constant ID_FIELD   => 'id';
use constant NAME_FIELD => 'description';

use constant REQUIRED_CREATE_FIELDS => qw(
  url
  description
  full_url
);

use constant UPDATE_COLUMNS => qw(
  url
  description
  full_url
  type
  can_get
  can_send
  must_send
  send_once
  ext_bz_regex_id
  matchrx
  format
);

use constant VALIDATORS => {
  url             => \&_check_url,
  description     => \&_check_description,
  full_url        => \&_check_full_url,
  type            => \&_check_type,
  can_get         => \&Bugzilla::Object::check_boolean,
  can_send        => \&Bugzilla::Object::check_boolean,
  must_send       => \&Bugzilla::Object::check_boolean,
  send_once       => \&Bugzilla::Object::check_boolean,
  ext_bz_regex_id => \&_check_regex,
  matchrx         => \&_check_match,
  format          => \&_check_format,
};

use constant TYPE => 'None';

###############################
####     Constructors     #####
###############################

sub new {
  my $invocant = shift;
  my $class    = ref($invocant) || $invocant;
  my $param    = shift;

  my $self = $class->SUPER::new($param);

  ## REDHAT EXTENSION BEGIN 821293
  if (  defined $self
    and $self->id
    and not Bugzilla->user->can_see_ext_bz($self->id))
  {
    # User cannot see tracker, so treat it like it doesn't exist
    return;
  }
  ## REDHAT EXTENSION END 821293

  if ($self && $self->type && $self->type ne 'None') {
    my $type_module = __PACKAGE__ . "::" . $self->type;
    ## no critic (ProhibitStringyEval)
    # type protected in _check_type
    eval("require $type_module");
    ## use critic
    if ($@) {
      Bugzilla->logger->logdie("Error loading type module: $@");
    }
    else {
      # Re-bless into the new rule module
      bless($self, $type_module);
    }
  }

  return $self;
}

sub remove_from_db {
  my ($self) = @_;
  my $dbh = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  # We need to remove all references to this external tracker first
  foreach my $bug_id (@{$self->bug_ids}) {
    my $ext_bugs
      = Bugzilla::Extension::ExternalBugs::Bug->match({ext_bz_id => $self->id});
    foreach my $ext_bug (@$ext_bugs) {
      $ext_bug->remove_from_db();
    }
  }

  # We need to remove the groups that restrict this tracker
  my $query = "DELETE FROM ext_bz_group_map WHERE ext_bz_id = ?";
  $dbh->do($query, undef, $self->id);

  $self->SUPER::remove_from_db();

  $dbh->bz_commit_transaction();

  return;
}

## REDHAT EXTENSION BEGIN 821293
# Modify the where value to restrict visibility
sub _do_list_select {
  my $invocant = shift;

  unless (Bugzilla->user->in_group('admin')) {
    my $group_list = Bugzilla->user->groups_as_string;
    if ($_[0]) {
      $_[0] .= " AND ";
    }

    $_[0] .= "
id IN (
    SELECT DISTINCT id
    FROM external_bugzilla b
        LEFT JOIN ext_bz_group_map m ON (m.ext_bz_id = b.id)
    WHERE m.group_id IS NULL OR m.group_id IN ($group_list)
)";
  }

  return $invocant->SUPER::_do_list_select(@_);
}
## REDHAT EXTENSION END 821293

###############################
####       Methods         ####
###############################

sub set_description { $_[0]->set('description',     $_[1]); return; }
sub set_url         { $_[0]->set('url',             $_[1]); return; }
sub set_full_url    { $_[0]->set('full_url',        $_[1]); return; }
sub set_type        { $_[0]->set('type',            $_[1]); return; }
sub set_can_get     { $_[0]->set('can_get',         $_[1]); return; }
sub set_can_send    { $_[0]->set('can_send',        $_[1]); return; }
sub set_must_send   { $_[0]->set('must_send',       $_[1]); return; }
sub set_send_once   { $_[0]->set('send_once',       $_[1]); return; }
sub set_regex       { $_[0]->set('ext_bz_regex_id', $_[1]); return; }
sub set_matchrx     { $_[0]->set('matchrx',         $_[1]); return; }
sub set_format      { $_[0]->set('format',          $_[1]); return; }

sub set_groups {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  # Get the new groups
  # IDs (or names) are passed in, but we want the object.
  my $group_ids  = shift;
  my $new_groups = [];
  foreach my $group_id (@$group_ids) {
    my $group = Bugzilla::Group->new($group_id);
    unless ($group) {
      $group = Bugzilla::Group->new({name => $group_id});
    }

    ThrowUserError('invalid_group_name', {name => $group_id}) unless ($group);

    push @$new_groups, $group;
  }

  my ($removed, $added) = diff_arrays($self->groups, $new_groups, 'id');

  foreach my $group (@$added) {
    my $query = "INSERT INTO ext_bz_group_map (ext_bz_id, group_id) VALUES (?, ?)";
    $dbh->do($query, undef, $self->id, $group->id);
  }

  foreach my $group (@$removed) {
    my $query = "DELETE FROM ext_bz_group_map WHERE ext_bz_id = ? AND group_id = ?";
    $dbh->do($query, undef, $self->id, $group->id);
  }

  return;
}

###############################
####      Accessors      ######
###############################

sub id          { return $_[0]->{'id'}; }
sub description { return $_[0]->{'description'}; }
sub url         { return $_[0]->{'url'}; }
sub full_url    { return $_[0]->{'full_url'}; }
sub type        { return $_[0]->{'type'}; }
sub can_get     { return $_[0]->{'can_get'}; }
sub can_send    { return $_[0]->{'can_send'}; }
sub must_send   { return $_[0]->{'must_send'}; }
sub send_once   { return $_[0]->{'send_once'}; }
sub regex       { return $_[0]->{'ext_bz_regex_id'}; }
sub matchrx     { return $_[0]->{'matchrx'}; }
sub format      { return $_[0]->{'format'}; }

sub ext_login {
  my $self = shift;

  unless (exists $self->{'ext_login'}) {
    my $logins = Bugzilla->localconfig->{ext_logins};

    foreach my $acct (@$logins) {
      if ($acct->{name} && $acct->{name} eq $self->description) {
        $self->{'ext_login'} = $acct->{user};
        last;
      }
    }
  }

  return $self->{'ext_login'};
}

sub ext_password {
  my $self = shift;

  unless (exists $self->{'ext_password'}) {
    my $logins = Bugzilla->localconfig->{ext_logins};

    foreach my $acct (@$logins) {
      if ($acct->{name} && $acct->{name} eq $self->description) {
        $self->{'ext_password'} = $acct->{password};
        last;
      }
    }
  }

  return $self->{'ext_password'};
}

sub groups {
  my $self = shift;
  return $self->{'groups'} if exists $self->{'groups'};
  my $group_ids
    = Bugzilla->dbh->selectcol_arrayref(
    'SELECT group_id FROM ext_bz_group_map WHERE ext_bz_id = ?',
    undef, $self->id);
  $self->{'groups'} = Bugzilla::Group->new_from_list($group_ids);
  return $self->{'groups'};
}

sub group_ids {
  my $self = shift;
  return $self->{'group_ids'} if exists $self->{'group_ids'};
  $self->{'group_ids'}
    = Bugzilla->dbh->selectcol_arrayref(
    'SELECT group_id FROM ext_bz_group_map WHERE ext_bz_id = ?',
    undef, $self->id);
  return $self->{'group_ids'};
}

sub regex_pattern {
  my $self  = shift;
  my $regex = $self->regex || return '';

  return $self->{'ext_bz_regexs'}{$regex}{regex}
    if exists $self->{'ext_bz_regexs'};
  $self->{'ext_bz_regexs'}
    = Bugzilla->dbh->selectall_hashref('SELECT * FROM ext_bz_regex', 'id');

  return $self->{'ext_bz_regexs'}{$regex}{regex};
}

sub regex_description {
  my $self  = shift;
  my $regex = $self->regex || return '';

  return $self->{'ext_bz_regexs'}{$regex}{description}
    if exists $self->{'ext_bz_regexs'};
  $self->{'ext_bz_regexs'}
    = Bugzilla->dbh->selectall_hashref('SELECT * FROM ext_bz_regex', 'id');

  return $self->{'ext_bz_regexs'}{$regex}{description};
}

###############################
####     Validators        ####
###############################

sub _check_url {
  my ($invocant, $url) = @_;
  if (!defined $url || $url eq '') {
    ThrowUserError("ext_bz_require_url");
  }
  return $url;
}

sub _check_description {
  my ($invocant, $description) = @_;
  if (!defined $description || $description eq '') {
    ThrowUserError("ext_bz_require_description");
  }
  return $description;
}

sub _check_full_url {
  my ($invocant, $full_url) = @_;
  if (!defined $full_url || $full_url eq '') {
    ThrowUserError("ext_bz_require_full_url");
  }
  return $full_url;
}

sub _check_type {
  my ($invocant, $type) = @_;

  $type = trim($type);
  $type =~ /^([A-Z][A-Za-z0-9_]+)$/i
    or ThrowUserError('ext_bz_unsafe_chars', {type => $type});

  # Detaint $value.
  $type = $1;

  # do the actual import if type is specified
  if ($type ne "None") {
    my $module = "Bugzilla::Extension::ExternalBugs::Type::" . $type;
    ## no critic (ProhibitStringyEval)
    # protected by regex above
    eval("require $module")
      || ThrowUserError("ext_bz_require_valid_type", {type => $type});
    ## use critic
  }

  return $type;
}

sub _check_regex {
  my ($invocant, $regex) = @_;
  $regex || return ();
  trick_taint($regex);

  my $rex = Bugzilla::Extension::ExternalBugs::Regex->new($regex);

  unless ($rex) {
    $rex = Bugzilla::Extension::ExternalBugs::Regex->new({name => $regex});
  }

  ThrowUserError("ext_bz_invalid_regex_id", {type => $regex}) unless $rex;

  return $rex->id;
}

sub _check_match {
  my ($invocant, $match) = @_;
  if ($match && $match ne '') {
    $match = trim($match);
    trick_taint($match);
    eval { qr{m/$match/}; }
      || ThrowUserError("ext_bz_invalid_match", {match => $match, message => $@});
  }

  return ($match);
}

sub _check_format {
  my ($invocant, $format) = @_;
  if ($format && $format ne '') {
    $format = trim($format);
    trick_taint($format);

    # hmm how do validate a sprintf format?
  }

  return ($format);
}
###############################
####     Constructors     #####
###############################

sub create {
  my $class = shift;
  my $dbh   = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  $class->check_required_create_fields(@_);

  my $params = $class->run_create_validators(@_);
  my $groups = delete $params->{groups};

  my $ext_bz = $class->insert_create_data($params);

  # Add the groups
  $ext_bz->set_groups($groups);

  $dbh->bz_commit_transaction();
  return $ext_bz;
}

###############################
####      Subroutines    ######
###############################

sub bug_count {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{'bug_count'}) {
    $self->{'bug_count'} = $dbh->selectrow_array(
      qq{
            SELECT COUNT(DISTINCT bug_id) FROM ext_bz_bug_map
             WHERE ext_bz_id = ?}, undef, $self->id
    );
  }

  return $self->{'bug_count'};
}

sub bug_ids {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{'bug_ids'}) {
    $self->{'bug_ids'} = $dbh->selectcol_arrayref(
      q{
            SELECT bug_id FROM ext_bz_bug_map
             WHERE ext_bz_id = ?}, undef, $self->id
    );
  }

  return $self->{'bug_ids'};
}

sub legal_types {
  my $type;

  if (@_) {
    my $invocant = shift;
    if ($invocant->isa(__PACKAGE__)) {
      $type = shift;
    }
    else {
      $type = $invocant;
    }
  }

  my @types           = ('None');
  my $extensions_path = bz_locations()->{'extensionsdir'};

  $type = "*" if !$type;

  foreach my $type_file (<$extensions_path/ExternalBugs/lib/Type/$type.pm>) {
    my ($type_name) = fileparse($type_file, '.pm');
    my $type_module = __PACKAGE__ . "::" . $type_name;
    trick_taint($type_module);
    ## no critic (ProhibitStringyEval)
    # you have to hijack invocant to control $type
    eval("require $type_module");
    ## use critic
    if ($@) {
      Bugzilla->logger->logdie("Error loading type module: $@");
    }
    push(@types, $type_module->TYPE);
  }

  return \@types;
}

# The next six subroutines do nothing, but can be overwritten by inheritance

sub get_data {
  my ($self, $ext_bug) = @_;
  return 0;
}

sub activiate {
  my ($self, $ext_bug) = @_;
  return 0;
}

sub deactivate {
  my ($self, $ext_bug) = @_;
  return 0;
}

sub send_changes {
  my ($self, $params) = @_;
  return 0;
}

sub sfdc_hostname {

  # For SFDC trackers, this value will be overwritten
  return;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 0;
}


sub do_rest_call {
  my ($self, $url, $method, $clean) = @_;

  # Send the request
  my $uri = URI->new($url);
  my $ua  = LWP::UserAgent->new;
  $ua->agent("RedHatBugzilla/" . BUGZILLA_VERSION);    # :)

  if ( Bugzilla->params->{external_tracker_proxy}
    && Bugzilla->params->{external_tracker_proxy} ne '')
  {
    $ua->proxy(['http', 'https'], Bugzilla->params->{external_tracker_proxy});
  }

  my $result  = undef;
  my $content = '';      ##defined $callobj ? encode_json($callobj) : '';

  eval {
    # Create a request
    my @headers = ('content-type' => 'application/json;charset=UTF-8');
##    push(@headers, ('authorization' => "Basic $auth")) if ($auth);

    my $req = HTTP::Request->new($method, $uri, \@headers, $content,);

    # Pass request to the user agent and get a response back
    my $res = $ua->request($req);

    Bugzilla->logger->debug("REST Calling: " . $uri->as_string);

    unless ($res->is_success) {
      die $res->status_line;
    }

    # Some calls return no content
    $result = '';
    if ($res->content) {
      my $txt = $res->content;
      if ($clean) {
        $txt =~ s/$clean//;
      }
      $result = decode_json($txt);
    }
  };

  if ($@) {
    my $err = $@;
    ThrowUserError('ext_bz_rest_error',
      {errtext => "$err", url => $url, et_type => TYPE});
  }

  if (ref $result and $result->{errors}) {
    my $error = Dumper($result->{errors});
    ThrowUserError('ext_bz_rest_error',
      {errtext => "$error", url => $url, et_type => TYPE});
  }

  # Log a comment, except if we were trying to log in
  Bugzilla->logger->debug("REST Success: $url");

  return $result;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs::Type - External Bug Type class.

=head1 SYNOPSIS

    use Bugzilla::Extension::ExternalBugs::Type;

    my $tracker = Bugzilla::Extension::ExternalBugs::Type->new(1);
    my $tracker = Bugzilla::Extension::ExternalBugs::Type->new({ name => 'AcmeProduct' });

    my $id          = $tracker->id;
    my $url         = $tracker->url;
    my $description = $tracker->description;
    my $full_url    = $tracker->full_url;
    my $type        = $tracker->type;

=head1 DESCRIPTION

Type represents a bug tracker object. It is an implementation
of L<Bugzilla::Object>, and thus provides all methods that
L<Bugzilla::Object> provides.

The methods that are specific to C<Bugzilla::Extension::ExternalBugs::Type> are listed
below.

=head1 ATTRIBUTES

can_get:
  Can this type pull data from remote sites? False, must be overridden by
  sub-classes.

can_send:
  Can this type send data to remotes sites? False, must be overridden by
  sub-classes.

description:
  Short text describing this Tracker. Used as name.

ext_bz_regex_id:
  Do IDs for this type need to match a regular expression? If so this is it.

format:
  sprintf format to use when displaying links as anchors.

full_url:
  The full path to get REST or RPC results from.

matchrx:
  A regex used to split up the ID for use in formatting.

must_send:
  Force changes to be sent to a Tracker that supports push even if there is a
  reason not to send it.

send_once:
  Only sends the first update to this tracker if there are multiple instances of
  this Tracker on the bug being changed.

type:
  Base type or sub-class?

url:
  The root URL for this site.

=head1 METHODS

=over

=item TODO

=back

=head1 SEE ALSO

L<Bugzilla::Object>

=cut
