# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ListOfBugs;

use 5.10.1;
use strict;
use warnings;

use parent qw(Bugzilla::Extension);

# This code for this is in ../extensions/ListOfBugs/lib/Util.pm
use Bugzilla;
use Bugzilla::Error;
use Bugzilla::Search;
use Bugzilla::Constants;

use Bugzilla::Extension::ListOfBugs::Util;
use Bugzilla::Extension::ListOfBugs::BugList;

our $VERSION = '0.01';

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  $schema->{lob} = {
    FIELDS => [
      id          => {TYPE => 'SERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      name        => {TYPE => 'TEXT',   NOTNULL => 1},
      description => {TYPE => 'TEXT'},
      owner       => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid'}
      },
      lastused => {TYPE => 'DATETIME', DEFAULT => 'NOW()', NOTNULL => 1},
      updated  => {TYPE => 'DATETIME', DEFAULT => 'NOW()', NOTNULL => 1},
    ],
    INDEXES => [lob_name_idx => {FIELDS => ['name'], TYPE => 'UNIQUE'},]
  };

  $schema->{lob_bugs} = {
    FIELDS => [
      lob_id => {
        TYPE       => 'INTEGER',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'lob', COLUMN => 'id'}
      },
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id'}
      },
    ],
    INDEXES => [lob_list_idx => {FIELDS => ['lob_id', 'bug_id'], TYPE => 'UNIQUE'}]
  };

  $schema->{lob_group_map} = {
    FIELDS => [
      lob_id => {
        TYPE       => 'INT4',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'lob', COLUMN => 'id'}
      },
      group_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'groups', COLUMN => 'id'}
      },
      is_view_group => {TYPE => 'BOOLEAN', NOTNULL => 1},
    ],
    INDEXES =>
      [lob_group_map_uniq => {FIELDS => [qw(lob_id group_id is_view_group)]},]
  };

  return;
}

sub buglist_filter_params {
  my ($self, $args) = @_;

  my $query = $args->{params};

  if (defined $query->{lob}) {
    my $buglist = delete $query->{lob};

    my $list = Bugzilla::Extension::ListOfBugs::BugList->new({name => $buglist});
    ThrowUserError('buglist_no_view_permission', {name => $buglist})
      unless (defined $list);
    ThrowUserError('buglist_no_view_permission', {name => $buglist})
      unless $list->can_view;

    $list->update_used;    # force update of last used timestamp

    # Empty out the query hash so no other arguments can pollute the search results.
    foreach my $key (keys %{$query}) {
      delete $query->{$key};
    }

    $query->{query_format} = 'advanced';
    $query->{f1}           = 'bug_id';
    $query->{o1}           = 'anyexact';
    $query->{v1}           = $list->bug_ids;
    $query->{limit}        = 20000;
  }

  return;
}

sub user_preferences {
  my ($self, $args) = @_;
  my $tab = $args->{'current_tab'};
  return unless $tab eq 'user_bug_lists';

  my $vars    = $args->{'vars'};
  my $handled = $args->{'handled'};

  my $user = Bugzilla->user(LOGIN_REQUIRED);

  my @owned_lists = grep { $_->owner_id == $user->id }
    Bugzilla::Extension::ListOfBugs::BugList->get_all;

  $vars->{lists} = \@owned_lists;

  $$handled = 1;

  return;
}

sub search_operator_field_override {
  my ($self, $args) = @_;
  my $operators = $args->{'operators'};
  my $search    = $args->{'search'};

  $operators->{bug_id}->{listofbugs} = sub {
    return _listofbugs(@_);
  };

  return;
}

sub _listofbugs {
  my ($self, $args) = @_;

  my $field    = $args->{field};
  my $operator = $args->{operator};

  ThrowUserError("search_field_operator_invalid",
    {field => $field, operator => $operator})
    unless ($field eq "bug_id");

  my $list = Bugzilla::Extension::ListOfBugs::BugList->new(
    {name => $args->{condition}->value});

  # Find out the list name.  Check if it's viewable.
  if (defined $list && $list->can_view) {
    my $full_field = $args->{full_field};
    my $list_name  = Bugzilla->dbh->quote($args->{condition}->value);
    $args->{term}
      = "$full_field IN (SELECT bug_id FROM lob_bugs WHERE lob_id = (SELECT id FROM lob WHERE name = $list_name))";
  }
  else {
    # Either the list does not exist, or is not viewable.
    ThrowUserError('buglist_no_view_permission',
      {name => $args->{condition}->value});
  }

  return;
}


sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};

  return unless substr($page, 0, 10) eq 'listofbugs';

  my $vars = $args->{vars};

  if ($page eq 'listofbugs/create.html') {
    Bugzilla->login(LOGIN_REQUIRED);
    $self->_create_bug_list($args);
  }
  elsif ($page eq 'listofbugs/save.html') {
    Bugzilla->login(LOGIN_REQUIRED);
    $self->_save_bug_list($args);
  }
  elsif ($page eq 'listofbugs/manage.html') {
    Bugzilla->login(LOGIN_REQUIRED);
    $self->_admin_bug_list($args);
  }
  elsif ($page eq 'listofbugs/show.html') {
    Bugzilla->login(LOGIN_REQUIRED);
    $self->_admin_bug_list($args);
  }
  elsif ($page eq 'listofbugs/delete.html') {
    Bugzilla->login(LOGIN_REQUIRED);
    $self->_delete_bug_list($args);
  }

  return;
}

sub _delete_bug_list {
  my ($self, $args) = @_;

  my $vars = $args->{vars};

  my $id = int scalar Bugzilla->cgi->param('list_id');

  my $list = Bugzilla::Extension::ListOfBugs::BugList->new($id);
  my $user = Bugzilla->user;

  if (!defined($list) || !$list->can_edit) {
    ThrowUserError('buglist_no_edit_permission', {name => $id});
  }

  my $confirm = int scalar Bugzilla->cgi->param('confirm');
  if ($confirm) {
    $list->remove_from_db;
    $vars->{confirm} = 1;
  }
  else {
    $vars->{list} = $list;
    $vars->{user} = $user;
  }

  return;
}

sub _admin_bug_list {
  my ($self, $args) = @_;

  my $user = Bugzilla->user(LOGIN_REQUIRED);
  unless ($user->in_group('admin')) {
    ThrowUserError('auth_failure',
      {group => 'admin', action => 'access', object => 'administrative_pages'});
  }

  my $vars = $args->{vars};

  my @all_lists = Bugzilla::Extension::ListOfBugs::BugList->get_all;

  $vars->{lists} = \@all_lists;

## ... why is this not being used?
##  my @owned_lists;
##  if ($user->id) {
##    @owned_lists = grep { $_->owner->id == $user->id } @all_lists;
##  }
##  my @visible_lists
##    = grep { $_->can_view && $_->owner->id != $user->id } @all_lists;
##
##  $vars->{user_lists}  = \@owned_lists;
##  $vars->{other_lists} = \@visible_lists;

  return;
}

sub _save_bug_list {
  my ($self, $args) = @_;

  my $vars = $args->{vars};

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $cgi = Bugzilla->cgi;

  my $name    = scalar $cgi->param('buglist_name');
  my $desc    = scalar $cgi->param('description');
  my $confirm = scalar $cgi->param('overwrite') // 0;
  my $list_id = scalar $cgi->param('list_id');

  if (!defined($name) || length($name) == 0) {
    ThrowUserError('buglist_no_name');
  }

  # Check to see if there's a list with this name.
  # If there is, ask for confirmation before we overwrite.
  my $existing = Bugzilla::Extension::ListOfBugs::BugList->new({name => $name});
  if (!$confirm && $existing) {
    if ($existing->owner->id == $user->id) {
      $vars->{overwrite}   = 1;
      $vars->{list_id}     = $list_id;
      $vars->{name}        = $name;
      $vars->{description} = $desc;
      $vars->{oldlist}     = $existing;
      return;
    }
    ThrowUserError('buglist_not_owner');
  }

  my $last_search = Bugzilla::Search::Recent->check({id => $list_id});

  Bugzilla->dbh->bz_start_transaction;

  if ($confirm && $existing) {

    # Delete the old search so it can be re-created.
    $existing->remove_from_db;
  }

  my $bl_args = {name => $name, description => $desc};

  my @view_groups = $cgi->param('view_group');
  my @edit_groups = $cgi->param('edit_group');

  my @view_group_obj = map { Bugzilla::Group->check({name => $_}) } @view_groups;
  my @edit_group_obj = map { Bugzilla::Group->check({name => $_}) } @edit_groups;

  my $list = Bugzilla::Extension::ListOfBugs::BugList->create($bl_args);
  if (scalar @view_group_obj) {
    $list->set_view_groups(\@view_group_obj);
  }
  if (scalar @edit_group_obj) {
    $list->set_edit_groups(\@edit_group_obj);
  }
  $list->set_bug_ids($last_search->bug_list);
  $list->update;

  Bugzilla->dbh->bz_commit_transaction;

  $vars->{name} = $name;

  return;
}

sub _get_groups {
  my $user = Bugzilla->user(LOGIN_REQUIRED);

  return [sort { $a->name cmp $b->name }
    grep { $user->in_group($_) }
    grep { $_->is_bug_group } Bugzilla::Group->get_all];
}

sub _create_bug_list {
  my ($self, $args) = @_;

  my $vars = $args->{vars};
  my $cgi  = Bugzilla->cgi;

  my $list_id = scalar $cgi->param('list_id');

  my $last_search = Bugzilla::Search::Recent->check({id => $list_id});

  my $bug_objs = Bugzilla::Bug->new_from_list($last_search->bug_list);

  my $groups = $self->_get_groups;

  $vars->{bugs}        = $bug_objs;
  $vars->{list_id}     = $list_id;
  $vars->{view_groups} = $groups;
  $vars->{edit_groups} = $groups;

  return;
}

sub webservice {
  my ($self, $args) = @_;

  $args->{dispatch}->{ListOfBugs} = 'Bugzilla::Extension::ListOfBugs::WebService';

  return;
}

__PACKAGE__->NAME;

__END__

=head1 NAME

Bugzilla::Extension::ListOfBugs

=head1 DESCRIPTION

A list of bugs is a static list of bug numbers that can be given out to users.
Unlike a bug search, the contents of a list can only be changed manually.
Also unlike a search, the URL will always be compact.

This provides a convenient way to give someone a list of bugs.  Normally the only
way to do this is to craft a custom search that lists bug numbers.  This is clumsy
and can quickly run into URL length issues.

=head2 Creation

A list can be created by running a normal bug search and then saving the results by
clicking the 'Save as List' button.  Alternatively a list can be created by using the
RPC interface.  This requires that individual bug numbers be given. See the RPC documentation
for instructions on how to do this. L<Bugzilla::Extension::ListOfBugs::WebService/create>

=head2 Updating

You can update a list of bugs by running a new search and saving the list with
the same name as the old search.  This will overwrite the old list with the new
list of bugs.

Alternatively the list can be manipulated via the RPC interface where bugs can
be added or removed. See the RPC documentation L<Bugzilla::Extension::ListOfBugs::WebService/update>

=head2 Access & Rights

=over

=item View Access

A list can be made private by restricting view access to a list of groups.  Only members
of the listed groups will be albe to use the bug list.

If you select no groups then the list will be publicly visible.

=item Update Access

A list can be updated by either the owner of the list, or a member of a group authorized
to update that list.

If no groups are selected, then only the creator of the list may edit it.
If groups are selected then any member of the selected groups may update the bug list.

=back

=head2 Owned Lists

You can find a list of bug lists that you have created in your user preference page.

=head2 Searching

You can search the list of bugs for a subset of this list by crafting your search
terms as normal, and then select the Custom Search option 'Bug ID' 'Is in the list of bugs' 'list name'.

If you wish to find bugs that are not it the list, but match your search terms,
you can tick the 'Not' box for that custom search row.

=head1 Author

Matt Tyson <mtyson@redhat.com>

