package Bugzilla::Extension::MultipleValues;
use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::Extension);

our $VERSION = '0.01';

use Bugzilla;
use Bugzilla::Util;
use Bugzilla::Bug;
use Bugzilla::Error;
use List::MoreUtils qw(uniq);

# Monkeypatch the symbol table to add these methods to
# Bugzilla::Bug
BEGIN {
  *Bugzilla::Bug::all_components       = \&_all_components;
  *Bugzilla::Bug::all_component_objs   = \&_all_component_objs;
  *Bugzilla::Bug::set_extra_components = \&_set_extra_components;
  *Bugzilla::Bug::extra_components     = \&_extra_components;

  *Bugzilla::Bug::all_versions       = \&_all_versions;
  *Bugzilla::Bug::set_extra_versions = \&_set_extra_versions;
  *Bugzilla::Bug::extra_versions     = \&_extra_versions;
}

# The tables that we are using to hold the multiple values.
use constant {
  COMPONENTS_TABLE => 'rh_multiple_bug_components',
  VERSIONS_TABLE   => 'rh_multiple_bug_versions',
};

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  # Components tabe

  $schema->{+COMPONENTS_TABLE} = {
    FIELDS => [
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'cascade'}
      },
      component_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'components', COLUMN => 'id', DELETE => 'cascade'}
      }
    ],
    INDEXES => [
      component_bugs_unique_idx =>
        {FIELDS => ['bug_id', 'component_id'], TYPE => 'UNIQUE'}
    ]
  };

  $schema->{+VERSIONS_TABLE} = {
    FIELDS => [
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'cascade'}
      },
      version_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'versions', COLUMN => 'id', DELETE => 'cascade'}
      }
    ],
    INDEXES => [
      versions_bugs_unique_idx =>
        {FIELDS => ['bug_id', 'version_id'], TYPE => 'UNIQUE'}
    ]
  };

  return;
}

# Any updates to the schema should go in here.
sub install_update_db {
  my ($self, $args) = @_;

  return;
}

# If we are coming from the webservice, check to see if we
# have any multiple components and versions.
#
# If we do, sort them into the component and extra_component
# fields, same for versions.
#
# It is an error to put a bug in more than one component or
# version if the product does not support this.
sub bug_before_create {
  return unless i_am_webservice();

  my ($self, $args) = @_;

  my $params = $args->{params};

  my $product;
  if (ref($params->{component}) eq 'ARRAY' || ref($params->{version}) eq 'ARRAY')
  {
    $product = Bugzilla::Product->new({name => $params->{product}});
  }

  if (ref($params->{component}) eq 'ARRAY') {
    if (scalar(@{$params->{component}}) > 1 && !$product->multiple_components) {
      ThrowUserError('no_multiple_components', {product_name => $product->name});
    }

    my $ary = $params->{component};

    $params->{component}        = shift @$ary;
    $params->{extra_components} = $ary if scalar(@$ary);
  }

  if (ref($params->{version}) eq 'ARRAY') {
    if (scalar(@{$params->{version}}) > 1 && !$product->multiple_versions) {
      ThrowUserError('no_multiple_versions', {product_name => $product->name});
    }
    my $ary = $params->{version};

    $params->{version}        = shift @$ary;
    $params->{extra_versions} = $ary if scalar(@$ary);
  }

  return;
}


# Take the data from the form and load it into 'component' and 'extra_components'
# Bugzilla::Object::set_all will call 'set_component' and 'set_extra_components'
# to apply the data to the object.
sub bug_before_set_all {
  my ($self,   $args) = @_;
  my ($params, $bug)  = @$args{qw(params bug)};

  my $product
    = $bug->product_obj || ThrowCodeError('product_not_defined', $params);

# When changing product we need to validate against the new product
  if ($params->{product} && ($params->{product} ne $product->name)) {
    $product = new Bugzilla::Product({name => $params->{product}});
  }

  if (ref $params->{'component'}) {

    # Check that this product can do multiple components
    if (!$product->multiple_components && scalar(@{$params->{component}}) > 1) {
      ThrowUserError('no_multiple_components', {product_name => $product->name});
    }

    my @components
      = sort { extraValuesSort($bug->component) } uniq @{$params->{'component'}};
    $params->{'component'}        = shift @components;
    $params->{'extra_components'} = \@components;
  }
  elsif (defined $params->{'component'}) {

    # Only a single component is defined, not an ARRAYREF, remove the extras.
    $params->{'extra_components'} = [];
  }

  if (ref $params->{'version'}) {
    if (!$product->multiple_versions && scalar(@{$params->{version}}) > 1) {
      ThrowUserError('no_multiple_versions', {product_name => $product->name});
    }

    my @versions
      = sort { extraValuesSort($bug->version) } uniq @{$params->{version}};
    $params->{version}        = shift @versions;
    $params->{extra_versions} = \@versions;
  }
  elsif (defined $params->{'version'}) {

    # Only a single version is defined, not an ARRAYREF, remove the extras
    $params->{extra_versions} = [];
  }

  return;
}


# Save the new bug state to the database, delete or add versions & components as required.
# This function is called from inside a transaction, so no need to set one here.
sub bug_start_of_update {
  my ($self, $args) = @_;

  my $new_bug = $args->{bug};
  my $old_bug = $args->{old_bug};

  # Changes is used to store the history information.
  my $changes = $args->{changes};

  my $dbh = Bugzilla->dbh;

  if ($new_bug->product_obj->multiple_components) {

    my ($removed, $added) = diff_arrays(
      $old_bug->extra_components // [],
      $new_bug->extra_components // [],
    );

    if (scalar(@$removed)) {
      my $query
        = 'DELETE FROM '
        . COMPONENTS_TABLE
        . ' WHERE bug_id = ? AND component_id IN ('
        . join(',', split(//, '?' x scalar(@$removed))) . ')';

      my @component_ids = map { $_->id } map {
        Bugzilla::Component->check({name => $_, product => $new_bug->product_obj})
      } @$removed;

      $dbh->do($query, undef, $new_bug->id, @component_ids);
    }
    if (scalar(@$added)) {
      my $query
        = 'INSERT INTO '
        . COMPONENTS_TABLE
        . ' (bug_id, component_id) VALUES (?, '
        . '(SELECT id FROM components WHERE name = ? AND product_id = ?))';

      my $sth = $dbh->prepare($query);
      $sth->bind_param(1, $new_bug->id);
      $sth->bind_param(3, $new_bug->product_id);

      foreach my $component (@$added) {
        $sth->bind_param(2, $component);
        $sth->execute();
      }
    }

    # Set the extra component history.
    $changes->{extra_components} = [join(',', @$removed), join(',', @$added),];
  }

  if ($new_bug->product_obj->multiple_versions) {
    my ($removed, $added)
      = diff_arrays($old_bug->extra_versions // [], $new_bug->extra_versions // []);

    if (scalar(@$removed)) {
      my $query
        = 'DELETE FROM '
        . VERSIONS_TABLE
        . ' WHERE bug_id = ? AND version_id IN ('
        . join(',', split(//, '?' x scalar(@$removed))) . ')';

      my @version_ids = map { $_->id } map {
        Bugzilla::Version->check({name => $_, product => $new_bug->product_obj})
      } @$removed;

      $dbh->do($query, undef, $new_bug->id, @version_ids);
    }

    if (scalar(@$added)) {
      my $query
        = 'INSERT INTO '
        . VERSIONS_TABLE
        . ' (bug_id, version_id) VALUES (?,'
        . '(SELECT id FROM versions WHERE value = ? AND product_id = ?))';

      my $sth = $dbh->prepare($query);
      $sth->bind_param(1, $new_bug->id);
      $sth->bind_param(3, $new_bug->product_id);

      foreach my $version (@$added) {
        $sth->bind_param(2, $version);
        $sth->execute;
      }
    }

    # Set the extra version history.
    $changes->{extra_versions} = [join(',', @$removed), join(',', @$added),];
  }

  return;
}


# the '$bug->version' or '$bug->component' method call normally returns a single
# value.  In this method we cheat by preloading multiple values into the bug object
# so that when the version or component method is called, it will return ALL components
# and versions for that bug.
#
# (A bug object will cache its version and component, in this method we cheat by setting
#  the cached value inside the bug object to be a list of all versions and components)
sub template_before_process {
  my ($self, $args) = @_;
  my ($vars, $file) = @$args{qw(vars file)};

  if ($file =~ m{bug/(create/create|edit|show).html.tmpl}) {
    my $product
      = defined $vars->{'bug'} ? $vars->{'bug'}->product_obj : $vars->{'product'};

    # We only do the below if we are modifying a bug
    if (defined $product and defined $vars->{bug}) {

      # Overriding legal_values for the extra fields and also select values
      # in the version and component fields.

      if ($product->multiple_components) {
        $vars->{bug}->{component} = $vars->{bug}->all_components;
      }

      if ($product->multiple_versions) {
        $vars->{bug}->{version} = $vars->{bug}->all_versions;
      }
    }
  }

  return;
}

sub object_end_of_set_all {
  my ($self, $args) = @_;

  my $object = @$args{object};

  ## REDHAT EXTENSION START 1135908
  if ($object->isa('Bugzilla::Bug')) {
    delete $object->{all_components};
    delete $object->{all_versions};
  }
  ## REDHAT EXTENSION END 1135908

  return;
}

sub search_operator_field_override {
  my ($self, $args) = @_;

  my $operators = $args->{'operators'};

  $operators->{component} = {
    _non_changed => sub { _component_nonchanged(@_) }
  };
  $operators->{version} = {
    _non_changed => sub { _version_nonchanged(@_) }
  };

  return;
}

sub _component_nonchanged {
  my ($invocant, $args) = @_;

  $args->{full_field} = 'components.name';
  $invocant->_do_operator_function($args);
  my $query = '('
    . Bugzilla::Search::build_subselect('bugs.component_id', 'components.id',
    'components', $args->{term})
    . ' OR ';

  $args->{full_field} = '_rh_c2.name';
  $invocant->_do_operator_function($args);
  $query .= Bugzilla::Search::build_subselect(
    'bugs.bug_id',
    '_rh_mbc.bug_id',
    'rh_multiple_bug_components _rh_mbc JOIN components _rh_c2 ON (_rh_mbc.component_id = _rh_c2.id)',
    $args->{term}
  ) . ')';

  $args->{term} = $query;

  return;
}

sub _version_nonchanged {
  my ($invocant, $args) = @_;

  $invocant->_do_operator_function($args);
  my $query = '(' . $args->{term} . ' OR ';

  $args->{full_field} = '_rh_v2.value';
  $invocant->_do_operator_function($args);
  $query .= Bugzilla::Search::build_subselect(
    'bugs.bug_id',
    '_rh_mbv.bug_id',
    'rh_multiple_bug_versions _rh_mbv JOIN versions _rh_v2 ON (_rh_mbv.version_id = _rh_v2.id)',
    $args->{term}
  ) . ')';

  $args->{term} = $query;

  return;
}


# Given an array of hashes (e.g. from DBI::selectall_arrayref) will
#  add the additional components and versions for bug records.
# If you are using a Bugzilla::Bug object, you should use
# $obj->all_components and $obj->all_versions instead.

sub add_extra_values {
  my $self = shift;
  my $args = shift;
  my $dbh  = Bugzilla->dbh;

  my $bugs         = $args->{bugs}         || return;
  my $component    = $args->{component}    || 0;
  my $version      = $args->{version}      || 0;
  my $bug_id_field = $args->{bug_id_field} || 'bug_id';

  foreach my $bug (@$bugs) {
    my $bug_id = $bug->{$bug_id_field};
    next if not $bug_id;

    # Should we update components?
    if ($component) {
      my $extra_components = $dbh->selectcol_arrayref("
                SELECT c.name FROM rh_multiple_bug_components mbc JOIN components c ON (mbc.component_id = c.id)  WHERE bug_id = ?",
        undef, $bug_id);
      $bug->{component} = [$bug->{component}, @$extra_components];
    }

    # Should we update versions?
    if ($version) {
      my $extra_versions = $dbh->selectcol_arrayref("
                SELECT v.value FROM rh_multiple_bug_versions mbv JOIN versions v ON (mbv.version_id = v.id) WHERE bug_id = ?",
        undef, $bug_id);
      $bug->{version} = [$bug->{version}, @$extra_versions];
    }
  }

  return;
}


sub buglist_columns {
  my ($self, $args) = @_;

  my $columns = $args->{columns};

  state $pg_suffix
    = lc(Bugzilla->localconfig->{db_driver}) eq 'pg' ? '::text' : '';

  $columns->{extra_components} = {
    name  => Bugzilla->dbh->sql_group_concat('DISTINCT rh_mbc_c.name' . $pg_suffix),
    title => 'Extra Components'
  };

  $columns->{extra_versions} = {
    name => Bugzilla->dbh->sql_group_concat('DISTINCT rh_mbv_v.value' . $pg_suffix),
    title => 'Extra Versions',
  };

  return;
}

sub buglist_column_joins {
  my ($self, $args) = @_;

  my $joins = $args->{column_joins};

  $joins->{extra_components} = {
    table   => 'rh_multiple_bug_components',
    as      => 'rh_mbc',
    then_to => {
      table => 'components',
      as    => 'rh_mbc_c',
      from  => 'rh_mbc.component_id',
      to    => 'id',
    }
  };

  $joins->{extra_versions} = {
    table   => 'rh_multiple_bug_versions',
    as      => 'rh_mbv',
    then_to => {
      table => 'versions',
      as    => 'rh_mbv_v',
      from  => 'rh_mbv.version_id',
      to    => 'id',
    }
  };

  return;
}

# Setter for Bugzilla::Bug
sub _set_extra_components {
  my ($self, $names, undef) = @_;

  # Ensure that the new names are valid
  my @new_extra_components = map { $self->_check_component($_) } @$names;

  $self->{extra_components} = [map { $_->name } @new_extra_components];

  return;
}

# Getter function
sub _extra_components {
  my ($self) = @_;

  unless (defined($self->{extra_components})) {
    my $query
      = 'SELECT name FROM '
      . COMPONENTS_TABLE
      . ' ec JOIN components c ON (ec.component_id = c.id) WHERE bug_id = ?';
    my $result = Bugzilla->dbh->selectcol_arrayref($query, undef, $self->id);

    $self->{extra_components} = $result // [];
  }

  return $self->{extra_components};
}

sub _all_components {
  my $self = shift;

  unless (defined($self->{all_components})) {
    $self->{all_components} = [$self->component, @{$self->extra_components}];
  }

  return $self->{all_components};
}

sub _all_component_objs {
  my ($self) = @_;
  return $self->{all_component_obj} if defined $self->{all_component_obj};
  return {} if $self->{error};

  $self->{all_component_obj}
    = [
    map { Bugzilla::Component->new({product => $self->product_obj, name => $_}) }
      @{$self->all_components}
    ];

  return $self->{all_component_obj};
}

sub _set_extra_versions {
  my ($self, $values) = @_;

  $self->{extra_versions} = [map { $self->_check_version($_) } @$values];

  return;
}

sub _extra_versions {
  my $self = shift;

  unless (defined($self->{extra_versions})) {
    my $query
      = 'SELECT v.value FROM '
      . VERSIONS_TABLE
      . ' ev JOIN versions v ON (ev.version_id = v.id) WHERE bug_id = ?';

    my $result = Bugzilla->dbh->selectcol_arrayref($query, undef, $self->id);

    $self->{extra_versions} = $result // [];
  }

  return $self->{extra_versions};
}

sub _all_versions {
  my $self = shift;

  unless (defined($self->{all_versions})) {
    $self->{all_versions} = [$self->version, @{$self->extra_versions}];
  }

  return $self->{all_versions};
}

# This will skew the sort function so that a component with the name of whatever is stored
# in $value will appear at the beginning of the sorted results.  This is done to prevent
# excessive component churn in the bug history.
sub extraValuesSort {
  my $value = shift;
  if (
    not defined $value    # Prevent undef warnings
    or $a eq $b
    )
  {    # Although $a and $b should never be the same, we need to handle it
    return $a cmp $b;
  }
  elsif ($a eq $value) {

    # $a should be sorted first
    return -1;
  }
  elsif ($b eq $value) {

    # $b should be sorted first
    return 1;
  }
  return $a cmp $b;
}


__PACKAGE__->NAME;
