# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::PlotlyReports;

use 5.10.1;
use strict;
use warnings;

use parent qw(Bugzilla::Extension);

use Bugzilla::Constants;

# This code for this is in ../extensions/PlotlyReports/lib/Util.pm
use Bugzilla::Extension::PlotlyReports::Util;
use Data::Dumper;

our $VERSION = '0.01';

BEGIN {
  *Bugzilla::Product::report_group         = \&_product_report_group;
  *Bugzilla::Product::set_report_group     = \&_product_set_report_group;
  *Bugzilla::_check_group                  = \&_check_group;
  *Bugzilla::User::can_report_on_product   = \&_can_report_on_product;
  *Bugzilla::User::get_reportable_products = \&_get_reportable_products;
}

# See the documentation of Bugzilla::Hook ("perldoc Bugzilla::Hook"
# in the bugzilla directory) for a list of all available hooks.
sub install_update_db {
  my ($self, $args) = @_;
  my $dbh = Bugzilla->dbh;

  if (!$dbh->bz_column_info('products', 'report_group')) {
    $dbh->bz_add_column(
      'products',
      'report_group',
      {
        TYPE       => 'INT3',
        DEFAULT    => 'NULL',
        REFERENCES => {TABLE => 'groups', COLUMN => 'id', DELETE => 'SET NULL'}
      }
    );

    $dbh->do(
      'update products set report_group = 161 where classification_id IN (1, 5, 7, 8, 10)'
    );
  }

  return;
}


sub page_before_template {
  my ($self, $args) = @_;
  my $page = $args->{'page_id'};
  my $vars = $args->{'vars'};

  if ($page eq 'bug_status.html') {
    _bug_status_page($vars);
  }
  elsif ($page eq 'bug_severity.html') {
    _bug_severity_page($vars);
  }
  elsif ($page eq 'non_defaults.html') {
    _non_defaults_page($vars);
  }

  return;
}

sub _bug_status_page {
  my ($vars) = @_;

  Bugzilla->login(LOGIN_REQUIRED);
  my $user  = Bugzilla->user;
  my $input = Bugzilla->input_params;

  # If the user cannot enter bugs in any product, stop here.
  scalar @{$user->get_reportable_products()} || ThrowUserError('no_products');

  # Create data structures representing each classification
  my @classifications = ();
  foreach my $c (@{$user->get_selectable_classifications}) {

    # Create hash to hold attributes for each classification.
    my %classification = (
      'name'     => $c->name,
      'products' => [@{$user->get_reportable_products($c->id)}]
    );

    # Assign hash back to classification array.
    push @classifications, \%classification;
  }

  $vars->{'classifications'} = \@classifications;
  my @prod_names;

  if ($input->{'product'}) {
    my @prods;
    if (ref($input->{'product'}) eq 'ARRAY') {
      @prods = @{$input->{'product'}};
    }
    else {
      push(@prods, $input->{'product'});
    }

    foreach my $p_name (@prods) {
      my $prod = Bugzilla::Product->check($p_name);

      push(@prod_names, $prod->name) if ($user->can_report_on_product($prod));
    }
  }
  $vars->{'prod_names'}      = \@prod_names;
  $vars->{'auto'}            = $input->{'auto'} // 0;
  $vars->{'hide_empty'}      = $input->{'hide_empty'} // 0;
  $vars->{'customer'}        = $input->{'customer'} // 0;
  $vars->{'date_from'}       = $input->{'date_from'} // '';
  $vars->{'date_to'}         = $input->{'date_to'} // '';
  $vars->{'interval_period'} = $input->{'interval_period'} // '';
  $vars->{'group'}           = $input->{'group'} // '';
  $vars->{'by_group'}        = $input->{'by_group'} // '';
  $vars->{'show'}            = $input->{'show'} // '';

  return;
}

sub _bug_severity_page {
  my ($vars) = @_;

  Bugzilla->login(LOGIN_REQUIRED);

  my $user  = Bugzilla->user;
  my $input = Bugzilla->input_params;

  # If the user cannot enter bugs in any product, stop here.
  scalar @{$user->get_reportable_products()} || ThrowUserError('no_products');

  # Create data structures representing each classification
  my @classifications = ();
  foreach my $c (@{$user->get_selectable_classifications}) {

    # Create hash to hold attributes for each classification.
    my %classification = (
      'name'     => $c->name,
      'products' => [@{$user->get_reportable_products($c->id)}]
    );

    # Assign hash back to classification array.
    push @classifications, \%classification;
  }

  $vars->{'classifications'} = \@classifications;

  my @prod_names;

  if ($input->{'product'}) {
    my @prods;
    if (ref($input->{'product'}) eq 'ARRAY') {
      @prods = @{$input->{'product'}};
    }
    else {
      push(@prods, $input->{'product'});
    }

    foreach my $p_name (@prods) {
      my $prod = Bugzilla::Product->check($p_name);

      push(@prod_names, $prod->name) if ($user->can_report_on_product($prod));
    }
  }
  $vars->{'prod_names'}        = \@prod_names;
  $vars->{'hide_empty'}        = $input->{'hide_empty'} // 0;
  $vars->{'auto'}              = $input->{'auto'} // 0;
  $vars->{'deadline'}          = $input->{'deadline'} // '';
  $vars->{'internal_deadline'} = $input->{'internal_deadline'} // 0;
  $vars->{'all_severities'}    = $input->{'all_severities'} // 0;
  $vars->{'customer'}          = $input->{'customer'} // 0;

  return;
}

sub _non_defaults_page {
  my ($vars) = @_;

  Bugzilla->login(LOGIN_REQUIRED);

  my $user  = Bugzilla->user;
  my $input = Bugzilla->input_params;

  # If the user cannot enter bugs in any product, stop here.
  scalar @{$user->get_reportable_products()} || ThrowUserError('no_products');

  # Create data structures representing each classification
  my @classifications = ();
  foreach my $c (@{$user->get_selectable_classifications}) {

    # Create hash to hold attributes for each classification.
    my %classification = (
      'name'     => $c->name,
      'products' => [@{$user->get_reportable_products($c->id)}]
    );

    # Assign hash back to classification array.
    push @classifications, \%classification;
  }

  $vars->{'classifications'} = \@classifications;

  my @prod_names;

  if ($input->{'product'}) {
    my @prods;
    if (ref($input->{'product'}) eq 'ARRAY') {
      @prods = @{$input->{'product'}};
    }
    else {
      push(@prods, $input->{'product'});
    }

    foreach my $p_name (@prods) {
      my $prod = Bugzilla::Product->check($p_name);

      push(@prod_names, $prod->name) if ($user->can_report_on_product($prod));
    }
  }

  $vars->{'prod_names'} = \@prod_names;
  $vars->{'new'}        = $input->{'new'} // 0;
  $vars->{'auto'}       = $input->{'auto'} // 0;
  $vars->{'closed'}     = $input->{'closed'} // '';
  $vars->{'title'}      = $input->{'title'} // undef;
  $vars->{'by_pool'}    = $input->{'by_pool'} // 0;

  return;
}

sub _product_report_group     { return $_[0]->{'report_group'}; }
sub _product_set_report_group { $_[0]->set('report_group', $_[1]); return; }

sub update_product_add_vars {
  my ($self, $args) = @_;

  my $product = $args->{product};
  my $changes = $args->{changes};
  my $vars    = $args->{vars};
  my $dbh     = Bugzilla->dbh;
  $dbh->bz_start_transaction();

  my $curr_report_group = $product->report_group();
  my $report_group;
  $report_group = Bugzilla::Group->new($curr_report_group)
    if ($curr_report_group);

  if (($vars->{report_group} ? $vars->{report_group} : '') ne
    ($report_group ? $report_group->name : ''))
  {
    my $new_group = Bugzilla::Group->new({name => $vars->{report_group}});

    $changes->{report_group} = [
      ($report_group ? $report_group->name : 'All logged in users'),
      ($new_group    ? $new_group->name    : 'All logged in users')
    ];

    $product->set_report_group($new_group ? $new_group->id : '');
    $product->update();
  }
  $dbh->bz_commit_transaction();

  return;
}

sub _check_group {
  my ($invocant, $id) = @_;

  return unless ($id);

  my $group;

  if ($id =~ m/^\d+$/) {
    $group = Bugzilla::Group->new($id);
  }
  else {
    $group = Bugzilla::Group->new({name => $id});
  }

  return ($group->id);
}

sub _can_report_on_product {
  my ($user, $product) = @_;

  my $can = 1;

  if ($product->report_group && !$user->in_group_id($product->report_group)) {
    $can = 0;
  }

  return $can;
}

sub _get_reportable_products {
  my $user             = shift;
  my $class_id         = shift;
  my $class_restricted = Bugzilla->params->{'useclassification'} && $class_id;

  if (!defined $user->{reportable_products}) {
    my @prods = ();
    foreach my $product (@{$user->get_selectable_products()}) {
      push(@prods, $product) if ($user->can_report_on_product($product));
    }

    $user->{reportable_products} = \@prods;
  }

  if ($class_restricted) {
    return [grep { $_->classification_id == $class_id }
        @{$user->{reportable_products}}];
  }

  return ($user->{reportable_products});
}

sub webservice {
  my ($self, $args) = @_;

  $args->{dispatch}->{PlotlyReports}
    = 'Bugzilla::Extension::PlotlyReports::WebService';

  return;
}

__PACKAGE__->NAME;


__END__

=head1 NAME

Bugzilla::Extension::PlotlyReports

=head1 DESCRIPTION

This module contains hand crafted reports displayed with the L<Plotly.js|https://plot.ly/javascript/> library.

=head1 ACLs

A user must be logged in to generate these reports.

A user must be able to access a product to be able to report on it.

A new product level field named I<report_group> has been added to allow additional
control. When set this will limit report generation for the product to users who
are in the selected group.

Reports are B<not> filtered for user bug access; a user can see statistics
including bugs they do not have access to. This is considered acceptable as no
data is leaked about individual bugs. The most a user can know is how many
bugs exist that they cannot access. If this knowledge must be protected then the
I<report_group> must be set to a group that enforces this restriction.

=head2 Current Reports

=over

=item L<Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity>

=item L<Bugzilla::Extension::PlotlyReports::Graphs::BugStatus>

=back

=head2 TODO

Some interesting links about Plotly

=over

=item L<bar-chart-with-hover-text|https://plot.ly/javascript/bar-charts/#bar-chart-with-hover-text>

Extra info where required?

=back

=head2 group_check

testing ...

=cut
