# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::PlotlyReports::Graphs::NonDefaults;

use strict;
use warnings;
use 5.10.1;

use Bugzilla;
use Data::Dumper;
local $Data::Dumper::Sortkeys = 1;
use List::MoreUtils qw(uniq);

###############################################################################
# Variables
###############################################################################


sub new {
  my $class = shift;
  my $opt   = shift;
  my $self  = bless {
    'products' => $opt->{products} // 'Bugzilla',
    'data'     => {},
    'title'    => $opt->{title} // 'Non-Defaults Report',
    'closed'   => defined($opt->{closed}) ? $opt->{closed} : 0,
    'new'      => defined($opt->{'new'}) ? $opt->{'new'} : 0,
    'by_pool'  => defined($opt->{by_pool}) ? $opt->{by_pool} : 0,
  }, $class;

  return $self;
}


sub run {
  my $self = shift;

  my $dbh = Bugzilla->dbh;

  my $status = '';

  if ($self->{'new'}) {
    $status = qq{b.bug_status = 'NEW'\nAND};
  }
  elsif (!$self->{'closed'}) {
    $status = qq{b.bug_status <> 'CLOSED'\nAND};
  }

  my $products = '';

  # restrict to products, if given
  $products
    = "p.name IN ("
    . join(', ', map { $dbh->quote($_) } split(',', $self->{products}))
    . ")\nAND";

  my $start_sql = <<EOS;
SELECT p.name as prod, count(*) as num, string_agg(b.bug_id::text, ',') as bugs
FROM bugs b
LEFT JOIN products p on p.id = b.product_id
LEFT JOIN components c on c.id = b.component_id
LEFT JOIN bug_rh_sub_components bsc on bsc.bug_id = b.bug_id
LEFT JOIN rh_sub_components sc on sc.id = bsc.rh_sub_component_id
LEFT JOIN bug_agile_pool bap on bap.bug_id = b.bug_id
WHERE
    $status
    $products
EOS

  my $def_assignee_sql = <<EOS;
    (
        b.assigned_to <>  c.initialowner
        AND (bsc.rh_sub_component_id IS NULL OR b.assigned_to <> sc.initialowner)
    )
EOS

  my $def_qa_sql = <<EOS;
(
    (
            c.initialqacontact IS NOT NULL
        AND b.qa_contact <>  c.initialqacontact
    )
    AND
    (
        bsc.rh_sub_component_id IS NULL
        OR (
                c.initialqacontact IS NOT NULL
            AND b.qa_contact <> sc.initialqacontact
        )
    )
)
EOS

  my $def_docs_sql = <<EOS;
(
    (       c.initialdocscontact IS NOT NULL
        AND b.docs_contact <> c.initialdocscontact
    )
    AND
    (
        bsc.rh_sub_component_id IS NULL
        OR (
                sc.initialdocscontact IS NOT NULL 
            AND b.docs_contact <> sc.initialdocscontact
        )
    )
)
EOS

  my $def_pool_sql = <<EOS;
(
    (
            c.default_pool IS NOT NULL
        AND bap.pool_id <> c.default_pool
    )
    AND
    (
        bsc.rh_sub_component_id IS NULL
        OR (
                sc.default_pool IS NOT NULL
            AND bap.pool_id <> sc.default_pool
        )
    )
)
EOS

  my $end_sql = <<EOS;
GROUP BY p.name
EOS

  my %results;

  foreach my $prod (split(',', $self->{products})) {
    $results{$prod}->{'counts'} = [0,          0,    0,      0];
    $results{$prod}->{'labels'} = ['Assignee', 'QA', 'Docs', 'Pool'];
    $results{$prod}->{'bugs'}   = ['',         '',   '',     ''];
  }

  my $sql = $start_sql . $def_assignee_sql;
  $sql .= "AND $def_pool_sql" if ($self->{by_pool});
  $sql .= $end_sql;

  my $sth = $dbh->prepare($sql);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref) {
    $results{$row->{'prod'}}->{'counts'}->[0] = $row->{'num'};
    $results{$row->{'prod'}}->{'bugs'}->[0]   = $row->{'bugs'};
  }
  $sth->finish();

  $sql = $start_sql . $def_qa_sql;
  $sql .= "AND $def_pool_sql" if ($self->{by_pool});
  $sql .= $end_sql;

  $sth = $dbh->prepare($sql);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref) {
    $results{$row->{'prod'}}->{'counts'}->[1] = $row->{'num'};
    $results{$row->{'prod'}}->{'bugs'}->[1]   = $row->{'bugs'};
  }
  $sth->finish();

  $sql = $start_sql . $def_docs_sql;
  $sql .= "AND $def_pool_sql" if ($self->{by_pool});
  $sql .= $end_sql;

  $sth = $dbh->prepare($sql);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref) {
    $results{$row->{'prod'}}->{'counts'}->[2] = $row->{'num'};
    $results{$row->{'prod'}}->{'bugs'}->[2]   = $row->{'bugs'};
  }
  $sth->finish();

  $sql = $start_sql . $def_pool_sql . $end_sql;

  $sth = $dbh->prepare($sql);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref) {
    $results{$row->{'prod'}}->{'counts'}->[3] = $row->{'num'};
    $results{$row->{'prod'}}->{'bugs'}->[3]   = $row->{'bugs'};
  }
  $sth->finish();

  $self->{data} = \%results;

  return;
}

1;

__END__


=head1 Name

Bugzilla::Extension::PlotlyReports::Graphs::NonDefaults

=head1 Description

This module generates reports listing bugs that have non-default users in the
assignee, qa, or docs roles, or with a non-default pool.

=head1 METHODS

=head2 new

=over

=item B<Description>

Creates an object to generate a graph with.

  my %opt = (new => 1, by_pool => 1);
  my $gf  =
    Bugzilla::Extension::PlotlyReports::Graphs::NonDefaults->new({%opt});

  $gt->run();
  $gt->generate();

=item B<Params>

=over

=item C<title>

C<string> The start string for the graph title. Defaults to
'Non-Defaults Report'. It will get modified by other parameters.

=item C<new>

C<bool> limit report to new bugs.

=item C<closed>

C<bool> report on all bugs including closed bugs.

=item C<by_pool>

C<bool> limit report to bugs that have a non-default pool.

=item C<products>

C<string> a comma separated list of product names to restrict the results to.


=back

=item B<Returns>

A C<Bugzilla::Extension::PlotlyReports::Graphs::NonDefaults> object.

=back

=head2 run

=over

=item B<Description>

Run the queries to gather the bug data.

=item B<Params>

None.

=item B<Returns>

Nothing.

=back

=head2 generate

=over

=item B<Description>

Process the bug data and generate the results.

=item B<Params>

None.

=item B<Returns>

Nothing.

=back

=cut
