package Bugzilla::Extension::Push::Connector::STOMP_SSL;

use strict;
use warnings;
use 5.10.1;

use base 'Bugzilla::Extension::Push::Connector::Base';

use Bugzilla::Constants;
use Bugzilla::Extension::Push::Constants;
use Bugzilla::Extension::Push::Util;
use Net::STOMP::Client;
use IO::Socket::SSL qw(SSL_VERIFY_PEER);
use Authen::Credential;
use Data::Dumper;

sub options {
  return (
    {
      name     => 'destination',
      label    => 'Destination',
      type     => 'string',
      required => 1,
    },
    {name => 'cert', label => 'Certificate', type => 'string', required => 1,},
    {name => 'key',  label => 'Keyfile',     type => 'string', required => 1,},
    {name => 'pass', label => 'Pass Phrase', type => 'string',},
    {
      name     => 'max_retries',
      label    => 'Connection Retry Count',
      type     => 'string',
      default  => '3',
      required => 1,
    },
    {
      name     => 'receipt_timeout',
      label    => 'Broker receipt timeout',
      type     => 'string',
      default  => '5',
      required => 1,
    },
    {
      name     => 'ca_store',
      label    => 'SSL CA Store',
      type     => 'string',
      default  => '/etc/pki/tls/certs/ca-bundle.crt',
      required => 1,
    },
    {
      name     => 'host',
      label    => 'Primary host',
      type     => 'string',
      default  => 'localhost',
      required => 1,
    },
    {
      name     => 'port',
      label    => 'Primary port',
      type     => 'string',
      default  => '61617',
      required => 1,
    },

    # Backup hosts to use in case the primary has failed.
    {name => 'backup_host_1', label => 'Backup host 1', type => 'string',},
    {name => 'backup_port_1', label => 'Backup port 1', type => 'string',},
    {name => 'backup_host_2', label => 'Backup host 2', type => 'string',},
    {name => 'backup_port_2', label => 'Backup port 2', type => 'string',},
  );
}

sub stop {
  my ($self) = @_;

  Bugzilla->push_ext->logger->debug('STOMP: Disconnecting');

  if (defined($self->{mq})) {

 # Net::STOMP::Client will attempt to negotiate a disconnect with the remote server.
 # If the remote is no longer connected to the socket, the disconnect() call will
 # fail with a broken pipe or a similar error.  To guard against this call failing
 # we trap the error and ignore it.
    eval { $self->{mq}->disconnect; };
    if ($@) {
      chomp $@;
      Bugzilla->push_ext->logger->debug(
        "Net::STOMP::Client->disconnnect() failed with error: '$@'");
    }
    delete $self->{mq};
  }

  return;
}

sub _connect {
  my ($self) = @_;

  $self->stop if $self->_is_connected;

  $self->{mq} = $self->_get_connection();

  return defined $self->{mq};
}


# Attempt to connect to a STOMP message queue server.
# Fall back to the backup hosts if the primary hosts are not functional.
sub _get_connection {
  my $self = shift;

  my $config = $self->config;
  my $mq;

  # Try the main server.
  $mq = $self->_test_connection($config->{host}, $config->{port});
  return $mq if defined $mq;

  # First backup server.
  $mq
    = $self->_test_connection($config->{backup_host_1}, $config->{backup_port_1});
  return $mq if defined $mq;

  # Second backup server.
  $mq
    = $self->_test_connection($config->{backup_host_2}, $config->{backup_port_2});
  return $mq if defined $mq;

  # If we get here, then the primary and all backups have failed.
  Bugzilla->logger->error('All message queue connections have failed!');

  return;
}

# This is just a wrapper to hide the ugly if(defined) logic.
# Check that the parameters have been set
# Also, make sure that the host:port has been defined.  (backups may be undef)
# The output of this function will be undef if there was a connection error.
sub _test_connection {
  my $self = shift;
  my $host = shift;
  my $port = shift;

  my $mq = undef;

  if (defined($host) && defined($port)) {
    $mq = $self->_try_connect($host, $port);
  }

  # Could be valid or undef, depending on what happened inside
  # _try_connect.
  return $mq;
}

# Do the actual connecting.
# Returns undef if a connection attempt failed.
# Net::STOMP::Client will die with an error if a connection fails,
# so we need to trap that and allow the caller to try the next
# message queue.
sub _try_connect {
  my $self = shift;
  my $host = shift;
  my $port = shift;
  my $mq;

  my $credential = $self->_credential;
  my ($login, $pass) = split(':', $self->config->{pass});
  $credential = $self->_credential unless ($login);
  eval {
    $mq = Net::STOMP::Client->new(
      sockopts => {
        SSL_verify_mode => SSL_VERIFY_PEER,
        SSL_ca_file     => $self->config->{ca_store},
      },
      host => $host,
      port => $port,
      auth => $credential,
    );
    my %args;
    if ($login) {
      %args = (login => $login, passcode => $pass,);
    }
    $mq->connect(%args);
  };

  if ($@) {
    Bugzilla->logger->info(
      "Failed to connect to broker at $host:$port. Error: '$@'");
    return;
  }

  return $mq;
}

sub _is_connected {
  my ($self) = @_;

  return defined $self->{mq};
}

sub should_send {
  return 1;
}


# This function will do the actual sending.  It contains some logic to retry
# if the ->send call fails.  EG the remote side has closed the connection,
# possibly due to a broker failover.

sub _do_send {
  my ($self, $message) = @_;

# If the other side has closed the connection, this call can fail with a broken pipe
# or another error.  If this happens we'll just close the connection and retry.
  my $max_retries    = int $self->config->{max_retries};
  my $broker_timeout = int $self->config->{receipt_timeout};
  my $i;
  my $err_msg = "";

  for ($i = 0; $i < $max_retries; $i++) {
    my $res = undef;
    eval {
      my $stomp = $self->{mq};

      die 'Connection not valid' unless defined $stomp;

      ## REDHAT EXTENSION START 1461344
      Bugzilla->logger->debug(Dumper($message));

      $res = $stomp->send(
        destination => $self->config->{destination} . '.' . $message->routing_key,
        ## REDHAT EXTENSION END 1461344
        body            => $message->payload,
        receipt         => $stomp->uuid,
        esbSourceSystem => 'bugzilla',
        esbMessageType  => 'bugzillaNotification',
      );

      $stomp->wait_for_receipts(timeout => $broker_timeout);

      if ($stomp->receipts) {
        my $peer    = $stomp->peer;
        my $peerstr = $peer ? $peer->uri : '-- peer not found --';
        die("Timeout waiting for receipt from broker :$peerstr");
      }
    };
    if ($@) {
      $err_msg = $@;
      chomp $err_msg;

      Bugzilla->push_ext->logger->debug(
        "Got error '$err_msg', retrying with new connection");

      # Log errors, except for Timeouts
      Bugzilla->logger->error($err_msg) unless ($err_msg =~ m/timeout/ig);
      $self->stop;
      $self->_connect;
      next;
    }

    # if ->send() dies, $res will be undef.
    last if defined $res;
  }

  # We've hit the retry limit.  Stop for now and they will be retried later.
  if ($i == $max_retries) {
    my $message = qq{STOMP: $err_msg. Retry limit reached. Stopping for now};
    Bugzilla->logger->error($message);
    die $message;
  }

  return;
}

sub send {
  my ($self, $message) = @_;

  my $config = $self->config;

  unless ($self->_is_connected) {
    $self->_connect() or die 'Connection could not be established';
  }

  my $message_id = $message->message_id;

  ## REDHAT EXTENSION START 1461344
  my $destination = $config->{destination} . '.' . $message->routing_key;
  ## REDHAT EXTENSION END 1461344

  Bugzilla->push_ext->logger->debug(
    "STOMP: Publishing message $message_id to $destination");

  $self->_do_send($message);

  return PUSH_RESULT_OK;
}

sub options_validate {
  my ($class, $config) = @_;

  # Check that the cert files exist on disk.

  my $key         = $config->{key};
  my $cert        = $config->{cert};
  my $max_retries = int $config->{max_retries};

  die "Key file '$key' does not exist!"         unless -e $config->{key};
  die "Certificate file '$cert' does not exist" unless -e $config->{cert};
  die "Connection Retry Count must be greater than zero" if $max_retries <= 0;

  return;
}

sub _credential {
  my ($self) = @_;

  my @args = (
    scheme => 'x509',
    cert   => $self->config->{cert},
    key    => $self->config->{key},
  );

  # Add the pass phrase if it exists.
  my $phrase = $self->config->{phrase};
  if (defined($phrase) && length($phrase)) {
    push @args, (pass => $phrase);
  }

  return Authen::Credential->new(@args);
}

1;
