# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::Schema;

use strict;
use warnings;
use 5.10.1;

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::User::Setting;

use base qw(Exporter);
our @EXPORT = qw(
  redhat_db_schema_abstract_schema
  redhat_install_before_final_checks
  redhat_install_update_db
);

## REDHAT EXTENSION BEGIN 917935
use YAML::Syck qw(LoadFile);
## REDHAT EXTENSION END 917935

use JSON qw(from_json);
use List::MoreUtils qw(none);

sub redhat_db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  ## REDHAT EXTENSION START 420461
  $schema->{'longdescs_activity'} = {
    FIELDS => [
      comment_id => {
        TYPE    => 'INT4',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'longdescs', COLUMN => 'comment_id', DELETE => 'CASCADE'}
      },
      who => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'}
      },
      change_when => {TYPE => 'DATETIME', NOTNULL => 1},
      old_comment => {TYPE => 'LONGTEXT', NOTNULL => 1},
      security    => {TYPE => 'BOOLEAN',  NOTNULL => 1, DEFAULT => 0},
    ],
    INDEXES => [
      longdescs_activity_comment_id_idx             => ['comment_id'],
      longdescs_activity_change_when_idx            => ['change_when'],
      longdescs_activity_comment_id_change_when_idx => [qw(comment_id change_when)],
    ],
  };
  ## REDHAT EXTENSION END 420461

  $schema->{bugs_release} = {
    FIELDS => [
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'CASCADE'}
      },
      value => {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'---'"},
    ],
    INDEXES => [
      bugs_release_bug_id_idx       => ['bug_id'],
      bugs_release_target_value_idx => ['value'],
    ],
  };

  $schema->{releases} = {
    FIELDS => [
      id         => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      product_id => {
        TYPE       => 'INT2',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'products', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      value    => {TYPE => 'varchar(64)', NOTNULL => 1},
      sortkey  => {TYPE => 'INT2',        NOTNULL => 1, DEFAULT => 0},
      isactive => {TYPE => 'BOOLEAN',     NOTNULL => 1, DEFAULT => 'TRUE'},
    ],
    INDEXES => [
      releases_product_id_idx => {FIELDS => [qw(product_id value)], TYPE => 'UNIQUE'},
    ],
  };

  $schema->{doc_type_text} = {
    FIELDS => [
      id             => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      cf_doc_type_id => {
        TYPE       => 'INT2',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'cf_doc_type', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      value => {TYPE => 'LONGTEXT', NOTNULL => 1},
    ],
    INDEXES => [
      doc_type_text_cf_doc_type_id_idx =>
        {FIELDS => [qw(cf_doc_type_id)], TYPE => 'UNIQUE'},
    ],
  };

  return;
}

sub redhat_install_update_db {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  _update_custom_fields(['cf_doc_type']);

  ## REDHAT EXTENSION 406171
  $dbh->bz_add_column('bug_status', 'description', {TYPE => 'MEDIUMTEXT'});

  ## REDHAT - Removing password expiry related columns from profiles
  #          table , bug #509100
  $dbh->bz_drop_column('profiles', 'password_expire_days');
  $dbh->bz_drop_column('profiles', 'password_last_changed');

  # 2008-01-31 dkl@redhat.com - Bug 174037
  $dbh->bz_add_column('flagtypes', 'view_group_id', {TYPE => 'INT3'});

  _update_status_and_resolution();

  ## REDHAT EXTENSION START 831328
  $dbh->bz_add_column('bugs_activity', 'object_id', {TYPE => 'INT3'});
  ## REDHAT EXTENSION END 831328

  _populate_releases_table();

  # Made the size of the value field in the releases table match
  # the size of the value column in the bugs_release table.
  $dbh->bz_alter_column('releases', 'value',
    {TYPE => 'varchar(64)', NOTNULL => 1});

  _add_products_defaultrelease();


  if (!exists $dbh->bz_column_info('releases', 'sortkey')->{DEFAULT}) {
    $dbh->bz_alter_column('releases', 'sortkey',
      {TYPE => 'INT2', NOTNULL => 1, DEFAULT => 0});
  }


  if ($dbh->bz_column_info('products', 'releaseurl')) {
    $dbh->bz_alter_column('products', 'releaseurl',
      {TYPE => 'TINYTEXT', NOTNULL => 1, DEFAULT => "''"});
  }

  $dbh->bz_add_column('releases', 'id',
    {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1});

  $dbh->bz_drop_column('products', 'releaseurl');

  # 2011-02-16 Drop the column (bugs_release table replaces it)
  if ($dbh->bz_column_info('bugs', 'target_release')) {
    $dbh->bz_drop_column('bugs', 'target_release');
  }

  ## REDHAT EXTENSION BEGIN 706784
  if (!$dbh->bz_column_info('products', 'multiple_components')) {
    $dbh->bz_add_column('products', 'multiple_components',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  if (!$dbh->bz_column_info('products', 'multiple_target_releases')) {
    $dbh->bz_add_column('products', 'multiple_target_releases',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  if (!$dbh->bz_column_info('products', 'multiple_versions')) {
    $dbh->bz_add_column('products', 'multiple_versions',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }
  ## REDHAT EXTENSION END 706784

  ## REDHAT EXTENSION BEGIN 653915
  if (!$dbh->bz_column_info('setting', 'required_groups')) {
    $dbh->bz_add_column('setting', 'required_groups', {TYPE => 'VARCHAR(32)'},);
  }
  ## REDHAT EXTENSION END 653915

  ## REDHAT EXTENSION BEGIN 711031
  if (!$dbh->bz_column_info('flagtypes', 'category')) {
    $dbh->bz_add_column('flagtypes', 'category', {TYPE => 'VARCHAR(50)'});
  }
  ## REDHAT EXTENSION END 711031

  ## REDHAT EXTENSION BEGIN 751352
  if (!$dbh->bz_column_info('groups', 'category')) {
    $dbh->bz_add_column('groups', 'category', {TYPE => 'VARCHAR(50)'});
  }
  ## REDHAT EXTENSION END 751352

  # As recommended by Monty Widenius for GNOME's upgrade.
  if (Bugzilla->localconfig->{db_driver} eq 'mysql') {
    $dbh->do('SET SESSION myisam_sort_buffer_size = 3221225472');
  }

  ## REDHAT EXTENSION BEGIN
  if (!$dbh->bz_column_info('releases', 'isactive')) {
    $dbh->bz_add_column('releases', 'isactive',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'TRUE'});
  }
  ## REDHAT EXTENSION END

  ## REDHAT EXTENSION START 829154
  $dbh->bz_add_column('products', 'always_private',
    {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'FALSE'});
  ## REDHAT EXTENSION END 829154

  # This column is no longer used, so should be dropped
  $dbh->bz_drop_column('fielddefs', 'public');

  ## REDHAT EXTENSION START 523845
  # This will monitor when a saved search was lased used, so it can be
  # removed after an expiry period.
  $dbh->bz_add_column('namedquery_group_map', 'last_used',
    {TYPE => 'DATETIME', 'DEFAULT' => 'NOW()', NOTNULL => 1});
  ## REDHAT EXTENSION END 523845

  ## REDHAT EXTENSION START 876015
  $dbh->bz_add_column('bugs', 'docs_contact', {TYPE => 'INT3'});
  $dbh->bz_add_fk("bugs", "docs_contact",
    {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'});
  $dbh->bz_add_column('components', 'initialdocscontact', {TYPE => 'INT3'});
  $dbh->bz_add_fk('components', 'initialdocscontact',
    {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'SET NULL'});
  ## REDHAT EXTENSION END 876015

  # BUGZILLA 4.4 UPGRADE
  $dbh->bz_alter_column('longdescs_activity', 'comment_id',
    {TYPE => 'INT4', NOTNULL => 1});

  if (!$dbh->bz_column_info('longdescs_activity', 'security')) {

    # Default new field off
    $dbh->bz_add_column('longdescs_activity', 'security',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );

    # But set it on for all existing changes
    $dbh->do('UPDATE longdescs_activity SET security = 1');
  }

  ## REDHAT EXTENSION BEGIN 903031
  $dbh->bz_alter_column('bugs_aliases', 'alias', {TYPE => 'varchar(40)'});
  ## REDHAT EXTENSION END 903031

  ## REDHAT EXTENSION 993863: While these are not strictly related to this
  # bug, they were uncovered when implementing it
  $dbh->bz_alter_column('namedqueries', 'name',
    {TYPE => 'varchar(255)', NOTNULL => 1});

  $dbh->bz_alter_column('profiles_activity', 'oldvalue',
    {TYPE => 'varchar(2000)'});
  $dbh->bz_alter_column('profiles_activity', 'newvalue',
    {TYPE => 'varchar(2000)'});

  ## REDHAT EXTENSION START 1067210
  $dbh->bz_add_index('component_cc', 'component_cc_comp_id_idx',
    ['component_id']);
  ## REDHAT EXTENSION END 1067210

  ## REDHAT EXTENSION START 1326559
  # was rh_on_dev_products
  if (!$dbh->bz_column_info('products', 'allows_on_dev')) {
    $dbh->bz_add_column('products', 'allows_on_dev',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  # was rh_no_clear_acks
  if (!$dbh->bz_column_info('products', 'allows_no_clear_acks')) {
    $dbh->bz_add_column('products', 'allows_no_clear_acks',
      {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 0},
    );
  }

  # was rh_always_cc
  if (!$dbh->bz_column_info('components', 'always_cc')) {
    $dbh->bz_add_column('components', 'always_cc',
      {TYPE => 'varchar(1024)', NOTNULL => 1, DEFAULT => "''"},
    );
  }
  ## REDHAT EXTENSION END 1326559

  ## REDHAT EXTENSION BEGIN 1651832
  if (Bugzilla->params->{'group_categories'} =~ m/Admin/) {
    my $sql
      = q{select count (*) from group_group_map where member_id = 6 and grantor_id not in (select id from groups where category = 'Admin')};
    my $res   = $dbh->selectall_arrayref($sql);
    my $count = $res->[0]->[0];
    if ($count ne '0') {
      $sql
        = q{delete from group_group_map where member_id = 6 and grantor_id not in (select id from groups where category = 'Admin')};
      $dbh->do($sql);
    }
  }
  ## REDHAT EXTENSION END 1651832

  ## REDHAT EXTENSION BEGIN 1737448
  if (!$dbh->bz_column_info('logincookies', 'id')
    && lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite')
  {

    # Primary key changes need to be committed before adding new pkey.
    $dbh->do(
      'ALTER TABLE logincookies DROP CONSTRAINT IF EXISTS logincookies_pkey');

    $dbh->_bz_real_schema->set_column('logincookies', 'cookie',
      {'TYPE' => 'varchar(16)', 'NOTNULL' => 1});
    $dbh->_bz_store_real_schema;

    $dbh->bz_alter_column('logincookies', 'cookie',
      {TYPE => 'varchar(22)', NOTNULL => 1});
    $dbh->bz_add_index('logincookies', 'logincookies_cookie_idx',
      {FIELDS => ['cookie'], TYPE => 'UNIQUE'});

    $dbh->bz_add_column('logincookies', 'id',
      {TYPE => 'INTSERIAL', NOTNULL => 1, PRIMARYKEY => 1});
    $dbh->bz_add_index('logincookies', 'logincookies_cookie_idx',
      {FIELDS => ['cookie'], TYPE => 'UNIQUE'});
  }
  ## REDHAT EXTENSION END 1737448

  return;
}

sub redhat_install_before_final_checks {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  ## REDHAT EXTENSION START 465941
  add_setting('private_comment_default', ['on', 'off'], 'off');
  ## REDHAT EXTENSION END 465941

  ## REDHAT EXTENSION START 212492
  add_setting('warn_duplicate', ['on', 'off'], 'on');
  ## REDHAT EXTENSION END 212492

  ## REDHAT EXTENSION START 554957
  add_setting('merge_bug_data', ['Always', 'Never', 'Prompt'], 'Never');
  ## REDHAT EXTENSION END 554957

  ## REDHAT EXTENSION START 904164
  add_setting('product_in_subject', ['on', 'off'], 'off');
  ## REDHAT EXTENSION END 904164

  add_setting('view_customer_cases_on_sfdc',
    ['as_sfdc', 'as_portal', 'as_ascension'], 'as_portal');

  if (!$dbh->bz_column_info('bugs', 'cf_deadline')) {
    my $field = Bugzilla::Field->new({name => 'cf_deadline'});
    if ($field) {
      $dbh->do("DELETE FROM fielddefs WHERE name = 'cf_deadline'");
    }
  }

  ## REDHAT EXTENSION START 1267048 1434222
  add_setting('headers_in_body', ['on', 'off'], 'off');
  add_setting('status_colours', ['off', 'pastel', 'colorbrewer'], 'off');
  add_setting('def_table_size', [25, 50, 100, 500, -1], 25);
  ## REDHAT EXTENSION END 1267048 1434222

  # Has to happen after extensions modify schema.
  _update_custom_fields();

  ## REDHAT EXTENSION START 1241769
  $dbh->bz_add_index('bugs', 'bugs_cf_clone_of_idx', ['cf_clone_of']);
  ## REDHAT EXTENSION END 1241769

  ## REDHAT EXTENSION START 1371333
  _add_doc_types();
  ## REDHAT EXTENSION END 1371333

  my $count
    = $dbh->selectrow_array(
    q{SELECT count(*) FROM bugs_activity WHERE id = 28889734 AND removed = '2018-09-21 20:42:33'}
    );

  if ($count) {
    say "Adjusting bugs_activity entries";
    $count = 0;
    my $SQL
      = q{select bug_id, added, removed, id from bugs_activity where fieldid = 6965 and removed <> '' order by bug_id, bug_when};

    my $sth = $dbh->prepare($SQL);

    $SQL = q{UPDATE bugs_activity SET removed = ? WHERE id = ?};
    my $update_sth = $dbh->prepare($SQL);

    $sth->execute();

    my $prev_bug_id = 0;
    my $prev_added  = '';
    while (my ($bug_id, $added, $removed, $id) = $sth->fetchrow_array()) {
      if ($prev_bug_id && ($bug_id == $prev_bug_id) && ($removed ne $added)) {
        say "Bug: $bug_id, record: $id changing '$removed' to '$prev_added'";
        $update_sth->execute($prev_added, $id);
      }
      $prev_bug_id = $bug_id;
      $prev_added  = $added;
      $count++;
    }

    say "Adjusted $count entries";
  }

## REDHAT EXTENSION START 1708003
  my $eb_group = Bugzilla::Group->new({name => 'editbugs'});

  if ($eb_group->category && $eb_group->category ne 'Admin') {

    my $com_group    = Bugzilla::Group->new({name => 'community'});
    my $bz_group     = Bugzilla::Group->new({name => 'team Bugzilla'});
    my $nobody_group = Bugzilla::Group->new({name => 'nobody'});

    unless ($com_group) {
      $com_group = Bugzilla::Group->create({
        name        => "community",
        description => "Anyone associated with any community in Red Hat Bugzilla",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Public',
      });
    }

    unless ($nobody_group) {
      $nobody_group = Bugzilla::Group->create({
        name => "nobody",
        description =>
          "A group to stop users accessing things. This group should have no direct or indirect members.",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Public',
      });
    }

    # move current groups on edit bugs to community
    eval {
      $dbh->do(q{UPDATE group_group_map SET grantor_id = ? WHERE grantor_id = 7},
        undef, $com_group->id);
      $dbh->do(q{UPDATE group_group_map SET member_id = ? WHERE member_id = 7},
        undef, $com_group->id);
    };

    # add admin/bot groups to editbugs
    $eb_group->set_category('Admin');
    $eb_group->update();

    my $sth_insert = $dbh->prepare(
      'INSERT INTO group_group_map
      (member_id, grantor_id, grant_type) VALUES (?, 7, 0)'
    );

    # admin, expunge, edit_all_bugs
    my @grp_ids = qw/6 502 319/;
    eval {
      foreach my $id (@grp_ids) {
        $sth_insert->execute($id);
      }
    };

    # Setup editbugs groups on Red Hat products
    my $rh_group = Bugzilla::Group->new({name => 'redhat'});

    foreach my $prod (Bugzilla::Product->get_all) {

      # Red Hat
      if ($prod->classification->name =~ m/(?:Red Hat|JBoss|Internal)/) {
        $prod->set_group_controls($rh_group, {editbugs => 1,});
      }
      elsif ($prod->classification->name =~ m/(?:Retired)/) {
        $prod->set_group_controls($nobody_group, {editbugs => 1,});
      }
      elsif ($prod->name eq 'Bugzilla') {
        $prod->set_group_controls($bz_group, {editbugs => 1,});
      }
      elsif ($prod->classification->name =~ m/(?:Workflows)/) {
        ## nop
      }
      else {
        $prod->set_group_controls($com_group, {editbugs => 1,});
      }

      $prod->update;
    }
  }
## REDHAT EXTENSION END 1708003

## REDHAT EXTENSION START 1729338
  my $ap_group  = Bugzilla::Group->new({name => 'all_partners'});
  my $class_obj = Bugzilla::Classification->new({name => 'Red Hat'});

  if ($ap_group && $class_obj) {
    $count = $dbh->selectrow_array(
      q{select * from group_control_map where product_id = 370 and editbugs = 1 and group_id = ?},
      undef, $ap_group->id
    );

    if ($ap_group && $class_obj && !$count) {
      $ap_group->set_is_active(1);
      $ap_group->update();

      foreach my $prod (@{$class_obj->products}) {

        # Red Hat
        if ($prod->is_active) {
          $prod->set_group_controls($ap_group, {editbugs => 1,});
          $prod->update;
        }
      }
    }
  }
## REDHAT EXTENSION END 1729338

## REDHAT EXTENSION START 1708041
  $dbh->do(
    q{DELETE FROM team_edit_groups WHERE group_id = 161 AND team_id IN (select id from agile_team where name like 'sst_%')}
  );

  $dbh->do(
    q{UPDATE agile_role SET can_edit_team = 1 WHERE name in ('Product Manager', 'Team Lead')}
  );
## REDHAT EXTENSION END 1708041

## REDHAT EXTENSION START 1257476
  my $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'bug_agile_pool.pool_id',
    'bug_agile_pool.pool_order',
    'docs_contact',
    'percentage_complete',
    'reporter_accessible',
    'rh_rule',
    'votes'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_INTEGER);

  $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'bug_id'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_BUG_ID);

  $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'creation_ts',
    'delta_ts'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_DATETIME);
## REDHAT EXTENSION END 1257476

  $sql = <<EOS;
  UPDATE fielddefs
  SET type = ?
  WHERE name IN (
    'ext_bz_bug_map.ext_bz_bug_id'
  )
EOS

  $dbh->do($sql, undef, FIELD_TYPE_FREETEXT);

  my $iu_group = Bugzilla::Group->new({name => 'invalid_account'});
  my $at_group = Bugzilla::Group->new({name => 'automaton'});
  unless ($iu_group && $at_group) {
    unless ($iu_group) {
      $iu_group = Bugzilla::Group->create({
        name        => "invalid_account",
        description => "Red Hat accounts that fail validation",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Red Hat',
      });
    }

    unless ($at_group) {
      $at_group = Bugzilla::Group->create({
        name        => "automaton",
        description => "Red Hat List and Bot accounts",
        isbuggroup  => 1,
        isactive    => 1,
        secure_mail => 0,
        category    => 'Red Hat',
      });
    }
  }

  return;
}

sub _update_status_and_resolution {
  my $dbh = Bugzilla->dbh;

  # Create the redhat specific statuses and resolutions and remove unused
  my $field = Bugzilla::Field->new({name => 'bug_status'});
  foreach my $new_value (
    qw/NEW ASSIGNED POST MODIFIED ON_DEV ON_QA VERIFIED RELEASE_PENDING CLOSED/)
  {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $new_value});
    next if $value;
    print "Creating value $new_value for field bug_status\n";
    Bugzilla::Field::Choice->type($field)->create({value => $new_value});
  }
  $dbh->do("DELETE FROM bug_status WHERE value = 'UNCONFIRMED'");
  foreach my $old_value (qw/REOPENED/) {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $old_value});
    next if !$value;
    print "Removing value $old_value for field bug_status\n";
    $value->remove_from_db();
  }

  $field = Bugzilla::Field->new({name => 'resolution'});
  foreach my $new_value (
    qw/NOTABUG DEFERRED CURRENTRELEASE NEXTRELEASE RAWHIDE
    ERRATA UPSTREAM CANTFIX INSUFFICIENT_DATA/
    )
  {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $new_value});
    next if $value;
    print "Creating value $new_value for field resolution\n";
    Bugzilla::Field::Choice->type($field)->create({value => $new_value});
  }
  foreach my $old_value (qw/INVALID/) {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $old_value});
    next if !$value;
    print "Removing value $old_value for field resolution\n";
    $value->remove_from_db();
  }

  # Update sort key for statuses
  my $count = 10;
  $field = Bugzilla::Field->new({name => 'bug_status'});
  foreach my $value (
    qw/NEW ASSIGNED POST MODIFIED ON_DEV ON_QA VERIFIED RELEASE_PENDING CLOSED/)
  {
    my $value = Bugzilla::Field::Choice->type($field)->new({name => $value});
    die("Nooooooo") if !$value;
    $value->set_sortkey($count);
    $value->update;
    $count += 10;
  }

  # Initialize standard status workflow
  my @workflow = (
    [undef,             'NEW'],
    [undef,             'ASSIGNED'],
    ['ASSIGNED',        'NEW'],
    ['ASSIGNED',        'VERIFIED'],
    ['ASSIGNED',        'CLOSED'],
    ['ASSIGNED',        'MODIFIED'],
    ['ASSIGNED',        'ON_DEV'],
    ['ASSIGNED',        'ON_QA'],
    ['ASSIGNED',        'RELEASE_PENDING'],
    ['ASSIGNED',        'POST'],
    ['CLOSED',          'NEW'],
    ['CLOSED',          'ASSIGNED'],
    ['CLOSED',          'POST'],
    ['CLOSED',          'MODIFIED'],
    ['CLOSED',          'ON_QA'],
    ['MODIFIED',        'NEW'],
    ['MODIFIED',        'VERIFIED'],
    ['MODIFIED',        'ASSIGNED'],
    ['MODIFIED',        'CLOSED'],
    ['MODIFIED',        'ON_DEV'],
    ['MODIFIED',        'ON_QA'],
    ['MODIFIED',        'RELEASE_PENDING'],
    ['MODIFIED',        'POST'],
    ['NEW',             'VERIFIED'],
    ['NEW',             'ASSIGNED'],
    ['NEW',             'CLOSED'],
    ['NEW',             'MODIFIED'],
    ['NEW',             'ON_DEV'],
    ['NEW',             'ON_QA'],
    ['NEW',             'RELEASE_PENDING'],
    ['NEW',             'POST'],
    ['ON_DEV',          'NEW'],
    ['ON_DEV',          'VERIFIED'],
    ['ON_DEV',          'ASSIGNED'],
    ['ON_DEV',          'CLOSED'],
    ['ON_DEV',          'MODIFIED'],
    ['ON_DEV',          'ON_QA'],
    ['ON_DEV',          'RELEASE_PENDING'],
    ['ON_DEV',          'POST'],
    ['ON_QA',           'NEW'],
    ['ON_QA',           'VERIFIED'],
    ['ON_QA',           'ASSIGNED'],
    ['ON_QA',           'CLOSED'],
    ['ON_QA',           'MODIFIED'],
    ['ON_QA',           'ON_DEV'],
    ['ON_QA',           'RELEASE_PENDING'],
    ['ON_QA',           'POST'],
    ['POST',            'NEW'],
    ['POST',            'VERIFIED'],
    ['POST',            'ASSIGNED'],
    ['POST',            'CLOSED'],
    ['POST',            'MODIFIED'],
    ['POST',            'ON_DEV'],
    ['POST',            'ON_QA'],
    ['POST',            'RELEASE_PENDING'],
    ['RELEASE_PENDING', 'NEW'],
    ['RELEASE_PENDING', 'VERIFIED'],
    ['RELEASE_PENDING', 'ASSIGNED'],
    ['RELEASE_PENDING', 'CLOSED'],
    ['RELEASE_PENDING', 'MODIFIED'],
    ['RELEASE_PENDING', 'ON_DEV'],
    ['RELEASE_PENDING', 'ON_QA'],
    ['RELEASE_PENDING', 'POST'],
    ['VERIFIED',        'NEW'],
    ['VERIFIED',        'ASSIGNED'],
    ['VERIFIED',        'CLOSED'],
    ['VERIFIED',        'MODIFIED'],
    ['VERIFIED',        'ON_DEV'],
    ['VERIFIED',        'ON_QA'],
    ['VERIFIED',        'RELEASE_PENDING'],
    ['VERIFIED',        'POST'],
  );

  print "Removing old bug status transitions...\n";
  $dbh->do("DELETE FROM status_workflow");

  print
    "Now filling the 'status_workflow' table with valid bug status transitions for Red Hat...\n";
  my $sth_select = $dbh->prepare('SELECT id FROM bug_status WHERE value = ?');
  my $sth
    = $dbh->prepare(
    'INSERT INTO status_workflow (old_status, new_status, require_comment) VALUES (?, ?, 0)'
    );

  foreach my $transition (@workflow) {
    my ($from, $to);

    # If it's an initial state, there is no "old" value.
    $from = $dbh->selectrow_array($sth_select, undef, $transition->[0])
      if $transition->[0];
    $to = $dbh->selectrow_array($sth_select, undef, $transition->[1]);

    # If one of the bug statuses doesn't exist, the transition is invalid.
    next if (($transition->[0] && !$from) || !$to);
    $sth->execute($from, $to);
  }

  return;
}

sub _update_custom_fields {
  my $permitted = shift;

  my $dbh = Bugzilla->dbh;

  ## REDHAT EXTENSION START 917935
  # Add Red Hat specific custom fields
  my $file = bz_locations()->{extensionsdir} . '/RedHat/custom_fields.yaml';
  if (!-e $file) {
    die "Cannot open custom field file: $file";
  }

  my $fields = LoadFile($file);

  foreach my $field (@$fields) {
    next
      if ($permitted && @$permitted && none { $field->{name} eq $_ } @$permitted);

    # We need to use SQL to see if the field exists. Bugzilla::Fields
    # hides fields that are not visible to a non privileged user
    ## BUGBUG why is the upgrade not priveleged?
    my $old_field = Bugzilla::Field->new({name => $field->{name}});
    next if ($old_field);

    my $vgrps = delete $field->{view_groups}  // [];
    my $egrps = delete $field->{edit_groups}  // [];
    my $agrps = delete $field->{admin_groups} // [];

    ## REDHAT EXTENSION START 1672226
    ## BUGBUG this breaks the tests because it's based on live data
    ## and the Ids change for the fields and products.
    if (lc(Bugzilla->localconfig->{db_driver}) eq 'sqlite') {
      delete $field->{visibility_field_id};
      delete $field->{value_field_id};
      delete $field->{visibility_values};
    }
    ## REDHAT EXTENSION END 1672226

    my $new_field = Bugzilla::Field->create($field);

    if (lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite') {
      $new_field->set_view_groups($vgrps);
      $new_field->set_edit_groups($egrps);
      $new_field->set_admin_groups($agrps);
    }

    ## REDHAT EXTENSION START 1672226
    if ($field->{name} eq 'cf_zstream_target_release') {
      my @vals = qw{
        8.0.0 8.1.0 8.2.0 8.3.0 8.4.0 8.5.0
        8.0.1 8.0.2 8.0.3 8.0.4 8.0.5
      };
      foreach my $val (@vals) {
        Bugzilla::Field::Choice->type($new_field)->create({value => $val});
      }
    }
    ## REDHAT EXTENSION END 1672226

    ## REDHAT EXTENSION START 1672734
    if ($field->{name} eq 'cf_epm_pri') {
      my @vals = (
        'Low', 'Medium', 'High', 'URGENT', 'VIP', 'BD', 'MVP', 'RUD (Only Highs)',
        'FastTrack'
      );
      my $count = 1;
      foreach my $val (@vals) {
        Bugzilla::Field::Choice->type($new_field)
          ->create({value => $val, sortkey => $count++});
      }
    }

    if ($field->{name} eq 'cf_epm_prf_state') {
      my @vals = (
        'PFR Submitted',
        'PFR Reviewed',
        'PFR Validated',
        'PFR RPL Ready',
        'PFR Release Assigned'
      );
      my $count = 1;
      foreach my $val (@vals) {
        Bugzilla::Field::Choice->type($new_field)
          ->create({value => $val, sortkey => $count++});
      }
    }
    ## REDHAT EXTENSION END 1672734
  }
  ## REDHAT EXTENSION START 917935

  return;
}

sub _populate_releases_table {
  my $dbh = Bugzilla->dbh;

  # Adding a table for target releases to the database
  # If the releases table is empty, and we're still back in a Bugzilla
  # that has a bugs.product field, that means that we just created
  # the releases table and it needs to be populated.
  my $releases_exist = $dbh->selectrow_array("SELECT DISTINCT 1 FROM releases");
  if (!$releases_exist && $dbh->bz_column_info('bugs', 'product')) {
    print "Replacing blank releases...\n";

    # If we are upgrading from 2.8 or earlier, we will have *created*
    # the releases table with a product_id field, but Bugzilla expects
    # it to have a "product" field. So we change the field backward so
    # other code can run. The change will be reversed later in checksetup.
    if ($dbh->bz_column_info('releases', 'product_id')) {

      # Dropping the column leaves us with a releases_product_id_idx
      # index that is only on the "value" column. We need to drop the
      # whole index so that it can be correctly re-created later.
      $dbh->bz_drop_index('releases', 'releases_product_id_idx');
      $dbh->bz_drop_column('releases', 'product_id');
      $dbh->bz_add_column('releases', 'product',
        {TYPE => 'varchar(64)', NOTNULL => 1}, '');
    }

    # Populate the release table with all existing values in the database
    my $sth = $dbh->prepare(
      "SELECT DISTINCT r.value, b.product_id
                                   FROM bugs b, bugs_release r
                                   WHERE b.bug_id = r.bug_id"
    );
    $sth->execute();

    print "Populating releases table...\n";

    while (my ($value, $product) = $sth->fetchrow_array()) {

      # check if the value already exists
      my $sortkey = substr($value, 1);
      if ($sortkey !~ /^\d+$/) {
        $sortkey = 0;
      }
      else {
        $sortkey *= 10;
      }
      my $ms_exists = $dbh->selectrow_array(
        "SELECT value FROM releases
                  WHERE value = ? AND product = ?", undef, $value, $product
      );

      if (!$ms_exists) {
        $dbh->do(
          "INSERT INTO releases(value, product, sortkey)
                          VALUES (?,?,?)", undef, $value, $product, $sortkey
        );
      }
    }
  }

  return;
}

sub _add_products_defaultrelease {
  my $dbh = Bugzilla->dbh;

  # Added a defaultrelease field to the products table, so that
  # we know which release to initially assign bugs to.
  if (!$dbh->bz_column_info('products', 'defaultrelease')) {
    $dbh->bz_add_column('products', 'defaultrelease',
      {TYPE => 'varchar(64)', NOTNULL => 1, DEFAULT => "'---'"});
    my $sth = $dbh->prepare("SELECT id, defaultrelease FROM products");
    $sth->execute();
    while (my ($product_id, $default_ms) = $sth->fetchrow_array()) {
      my $exists = $dbh->selectrow_array(
        "SELECT value FROM releases
                  WHERE value = ? AND product_id = ?", undef, $default_ms, $product_id
      );
      if (!$exists) {
        $dbh->do("INSERT INTO releases (value, product_id) " . "VALUES (?, ?)",
          undef, $default_ms, $product_id);
      }
    }
  }

  return;
}

sub _add_doc_types {
  my $dbh = Bugzilla->dbh;

  my %types = (
    'If docs needed, set a value'        => 5,
    'Bug Fix'                            => 10,
    'Enhancement'                        => 20,
    'Rebase: Bug Fixes and Enhancements' => 30,
    'Rebase: Bug Fixes Only'             => 40,
    'Rebase: Enhancements Only'          => 50,
    'Known Issue'                        => 60,
    'Release Note'                       => 70,
    'Technology Preview'                 => 80,
    'Deprecated Functionality'           => 90,
    'Removed functionality '             => 95,
    'No Doc Update'                      => 100,
  );

  my %templates = (
    "Known Issue" =>
      "Cause: \n\nConsequence: \n\nWorkaround (if any): \n\nResult: ",
    "Rebase: Bug Fixes and Enhancements" =>
      "Important: if this rebase instead contains *only bug fixes,* or *only enhancements*, select the correct option from the Doc Type drop-down list.\n\nRebase package(s) to version: \n\nHighlights, important fixes, or notable enhancements: ",
    "Enhancement" => "Feature: \n\nReason: \n\nResult: ",
    "Bug Fix"     => "Cause: \n\nConsequence: \n\nFix: \n\nResult: ",
    "Rebase: Bug Fixes Only" =>
      "Important: if this rebase also contains *enhancements* (or contains only enhancements), select the correct option from the Doc Type drop-down list.\n\nRebase package(s) to version: \n\nHighlights and important bug fixes: ",
    "Deprecated Functionality" =>
      "Deprecated functionality should describe removed or no longer supported features.",
    "Release Note" =>
      "A release note should describe in detail a new feature being introduced into the product.\n\nA release note is NOT a bug, errata enhancement, rebase, known issue, Technology Preview or deprecated functionality description; for these, select the corresponding Doc Type option.",
    "Technology Preview" =>
      "Package(s) providing the Technology Preview: \n\nDescription of the Technology Preview: ",
    "Rebase: Enhancements Only" =>
      "Important: if this rebase also contains *bug fixes* (or contains only bug fixes), select the correct option from the Doc Type drop-down list.\n\nRebase package(s) to version: \n\nHighlights and notable enhancements: ",
    "If docs needed, set a value" =>
      "If this bug requires documentation, please select an appropriate Doc Type value.",
    "No Doc Update"         => "",
    "Removed functionality" => "",

  );

  my $sth = $dbh->prepare(
    q{INSERT INTO cf_doc_type (value, sortkey,isactive) values (?, ?, 1)});

  my $sth1
    = $dbh->prepare(
    q{INSERT INTO doc_type_text (cf_doc_type_id, value) select cf_doc_type.id, ? FROM cf_doc_type where cf_doc_type.value = ?}
    );

  foreach my $key (keys(%templates)) {
    eval { $sth->execute($key, $types{$key}); };
    eval { $sth1->execute($templates{$key}, $key); };
  }

  return;
}

1;
