# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::WebService::Bugzilla;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Bug;
use Bugzilla::BugMail;
use Bugzilla::Constants;
use Bugzilla::Group;
use Bugzilla::Util;
use Bugzilla::WebService::Util qw(validate);
use Bugzilla::Error;

use JSON::XS;
use Data::Dumper;
use URI::Escape;
use List::MoreUtils qw(any);
use Bugzilla::Token qw(issue_hash_token);

use constant PUBLIC_METHODS => qw(
  product_get_objects
  product_update_objects
  product_create_objects
  product_delete_objects
  search_as
  can_delete_user
  delete_user
  terminate_sessions
  admin_apikeys
  edit_comment
);

=pod

=head1 NAME

Bugzilla::Extension::RedHat::WebService::Bugzilla - Red Hat specific XMLRPC methods

=head1 DESCRIPTION

This module provides methods that are Red Hat customisations.

Method calls are normally in the form of namespace.method(arguments).
Where namespace in this case is "RedHat". So for example
RedHat.getBug(bugid, username, password) would return the results
of getBug() for bug id $bugid.

=head1 METHODS

=head2 C<product_create_objects>

## REDHAT EXTENSION 1354240 BEGIN

Used to create a milestone, release or version object in a product or products

Pass in a single hash named 'data'.  This hash must contain an array of hashes
for each product that will have milestones, releases or versions created against
it

Below is the keys to be included in the array of hashes.

=over

=item B<name>

B<string> The name of the product that will be modified. Either the name or ID of the product must be specified

=item B<id>

B<integer> The ID of the product that will be modified. Either the name or ID of the product must be specified

=item B<versions>

B<array-of-hashes> I<optional> An array of hashes describing each milestone to be created.
The arguments for the versions hash is detailed below

=over

=item B<name>

B<string> The name of the version

=back

=item B<milestones>

B<array-of-hashes> I<optional> An array of hashes describing each milestone to be created.
The arguments for the versions hash is detailed below

=over

=item B<name>

B<string> The name of the milestone

=item B<sortkey>

B<integer> I<optional> The sort order of the milestone

=item B<default>

B<boolean> I<optional> True if you wish this to be the new default milestone

=back

=item B<releases>

B<array-of-hashes> I<optional> An array of hashes describing each release to be created.
The arguments for the release hash is detailed below

=over

=item B<name>

B<string> The name of the release

=item B<sortkey>

B<integer> I<optional> The sort order of this release

=item B<default>

B<boolean> I<optional> True if you want this to be the new default

=back

=back

=cut

sub product_create_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $changes;

  my $data = $params->{data}
    // ThrowCodeError('param_required', {param => 'data'});

  foreach my $product_hash (@$data) {
    my $product;
    if (defined($product_hash->{name})) {
      $product = Bugzilla::Product->check({name => $product_hash->{name}});
    }
    elsif (defined($product_hash->{id})) {
      $product = Bugzilla::Product->new($product_hash->{id});
    }
    else {
      ThrowCodeError('params_required', {params => ['name', 'id']});
    }

    my $product_changes = {name => $product->name, id => $product->id,};

    $user->check_can_admin_product($product->name);

    foreach my $milestone_hash (@{$product_hash->{milestones}}) {
      my $name = $milestone_hash->{name}
        // ThrowCodeError('param_required', {param => 'name'});
      my $sortkey = $milestone_hash->{sortkey} // 0;

      my $milestone
        = Bugzilla::Milestone->create({
        value => $name, sortkey => $sortkey, product => $product
        });

      if (defined($milestone_hash->{default}) && $milestone_hash->{default}) {
        $product->set_default_milestone($milestone->name);
        $product->update;
      }

      push @{$product_changes->{milestones}},
        {
        id      => $milestone->id,
        name    => $milestone->name,
        sortkey => $milestone->sortkey,
        };
    }

    foreach my $release_hash (@{$product_hash->{releases}}) {
      my $name = $release_hash->{name}
        // ThrowCodeError('param_required', {param => 'name'});
      my $sortkey = $release_hash->{sortkey} // 0;

      my $release
        = Bugzilla::Release->create({
        value => $name, sortkey => $sortkey, product => $product
        });

      if (defined($release_hash->{default}) && $release_hash->{default}) {
        $product->set_default_release($release->name);
        $product->update;
      }

      push @{$product_changes->{releases}},
        {id => $release->id, name => $release->name, sortkey => $release->sortkey};
    }

    foreach my $version_hash (@{$product_hash->{versions}}) {
      my $name = $version_hash->{name}
        // ThrowCodeError('param_required', {param => 'name'});

      my $version = Bugzilla::Version->create({value => $name, product => $product,});

      push @{$product_changes->{versions}},
        {id => $version->id, name => $version->name,};
    }

    push @{$changes->{data}}, $product_changes;
  }

  $dbh->bz_commit_transaction;

  return $changes;
}


# RedHat.product_update
# Update information about a products target release, target milestone, target_version

=head2 C<product_update_objects>

Update a set of product milestones, releases and version

=over

=item C<versions>

B<array> I<optional> A hash named versions contaning an array of hashes that describes each version to be updated.
The arguments for each version hash are detailed below

=over

=item B<id>

B<integer> The ID of the version to be updated.

=item B<name>

B<string> I<optional> The new name of the version.

=item B<is_active>

B<boolean> I<optional> True to enable the version, False to disable.

=back

=item C<milestones>

B<array> I<optional> An array of hashes describing each milestone to be updated.  One hash per milestone.
The arguments for each version hash are detailed below

=over

=item B<id>

B<integer> The ID of the milestone to be updated

=item B<name>

B<string> I<optional> The new name of the milestone

=item B<sortkey>

B<integer> I<optional> The sort order of the milestone

=item B<default>

B<boolean> I<optional> set to true to make this the new default milestone.

=item B<is_active>

B<boolean> I<optional> True to enable the milestone, False to disable.

=back

=item C<releases>

B<array> I<optional> An array of hashes describing each release to be updated.  One hash per release
The arguments for each hash are detailed below

=over

=item B<id>

B<integer> The ID of the release to be updated

=item B<name>

B<string> I<optional> The new name of the release

=item B<sortkey>

B<integer> I<optional> The sort order of this release

=item B<default>

B<boolean> I<optional> Set to true to make this the default release.

=item B<is_active>

B<boolean> I<optional> True to enable the release, False to disable.

=back

=back

=cut

sub _name_remap {
  my ($field) = @_;

  if ($field eq 'name') {
    return 'value';
  }
  elsif ($field eq 'is_active') {
    return 'isactive';
  }
  else {
    return $field;
  }
}

sub product_update_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $changes       = {};
  my $product_cache = {};

  foreach my $version_data (@{$params->{versions}}) {
    my $version_id = int $version_data->{id}
      // ThrowCodeError('param_required', {param => 'id'});
    my $version = Bugzilla::Version->check({id => $version_id});

    _check_product($user, $version->product_id, $product_cache);

    foreach my $field (qw(name is_active)) {
      next unless defined $version_data->{$field};

      # Bugzilla 5.0 has changed the setter name on the Bugzilla::Version object.
      # In 4.4 it was set_name, in 5.0 it is set_value.  We need to remap 'name'
      # to 'value' so this API will continue to work. Same for is_active -> isactive
      my $method = 'set_' . _name_remap($field);
      $version->$method($version_data->{$field});
    }

    my $vs_changes = $version->update;

    push @{$changes->{versions}->{$version_id}}, $vs_changes;
  }
  foreach my $milestone_data (@{$params->{milestones}}) {
    my $milestone_id = int $milestone_data->{id}
      // ThrowCodeError('param_required', {param => 'id'});
    my $milestone = Bugzilla::Milestone->check({id => $milestone_id});

    _check_product($user, $milestone->product_id, $product_cache);

    foreach my $field (qw(name is_active sortkey)) {
      next unless defined $milestone_data->{$field};
      my $method = "set_$field";
      $milestone->$method($milestone_data->{$field});
    }

    my $ms_changes = $milestone->update;

    if (defined($milestone_data->{default}) && $milestone_data->{default}) {
      my $product = Bugzilla::Product->check({id => $milestone->product_id});
      $product->set_default_milestone($milestone->name);
      $product->update;
      $ms_changes->{default} = 1;
    }


    push @{$changes->{milestones}->{$milestone_id}}, $ms_changes;
  }
  foreach my $release_data (@{$params->{releases}}) {
    my $release_id = int $release_data->{id}
      // ThrowCodeError('param_required', {param => 'id'});
    my $release = Bugzilla::Release->check({id => $release_id});

    _check_product($user, $release->product_id, $product_cache);

    foreach my $field (qw(name is_active sortkey)) {
      next unless defined $release_data->{$field};
      my $method = "set_$field";
      $release->$method($release_data->{$field});
    }

    if (defined($release_data->{default}) && $release_data->{default}) {
      my $product = Bugzilla::Release->check({id => $release->product_id});
      $product->set_default_release($release->name);
      $product->update;
    }

    my $rs_changes = $release->update;

    push @{$changes->{releases}->{$release_id}}, $rs_changes;
  }


  $dbh->bz_commit_transaction;

  return $changes;
}


# Fetch information about target release, target milestone, versions for a list
# of products or product ids.
# { names => ['Foo', 'Bar', 'Baz'], ids => [1, 2, 3, 4, 5] }

=head2 C<product_get_objects>

Returns all the versions, milestones and releases for the specified products

Arguments to this function are a list of product names and/or product ids

=head3 C<Params>

=over

=item B<names>

B<array-of-strings> I<optional> An array of strings of the product names.

=item B<ids>

B<array-of-integers> I<optional> An array of product ids

=back

=head3 C<returns>

Returns an array of hashes, one per product describing each version, milestone and component
in the product

=over

=item C<name>

B<string> The product name

=item C<id>

B<integer> The product id

=item C<versions>

A hash named versions that contains an array of hashes describing each
version in the product

=over

=item B<name>

B<string> The version name

=item B<id>

B<integer> The version ID

=item B<is_active>

B<integer> 1 if the version is active, 0 if inactive

=back

=item C<milestones>

A hash named milestones that contains an array of hashes describing each
milestone in the product

=over

=item B<name>

B<string> The milestone name

=item B<id>

B<integer> The milestone ID

=item B<is_active>

B<integer> 1 if the milestone is active, 0 if inactive

=item B<default>

B<integer> 1 if this milestone is the default.  Will not be present if
this is not the default

=back

=item C<releases>

=over

=item B<name>

B<string> The release name

=item B<id>

B<integer> The release ID

=item B<is_active>

B<integer> 1 if the release is active, 0 if inactive

=item B<default>

B<integer> 1 if this release is the default.  Will not be present if
this is not the default

=back

=back

=cut

sub product_get_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  # Convert product ids/names to product objects
  my @product_objs;
  if (defined($params->{names})) {
    push @product_objs,
      map { Bugzilla::Product->check({name => $_}) } @{$params->{names}};
  }
  if (defined($params->{ids})) {
    push @product_objs, @{Bugzilla::Product->new_from_list($params->{ids})};
  }

# For each product, get information about target_release, target_milestone, versions
  my $product_set = [];
  foreach my $product (@product_objs) {

    # Check that they can admin the products.
    $user->check_can_admin_product($product->name);

    my $product_data = {};

    # Target Milestone
    my $milestones = [];
    foreach my $milestone_obj (@{$product->milestones}) {
      my $milestone = {
        id        => $milestone_obj->id,
        name      => $milestone_obj->name,
        sortkey   => $milestone_obj->sortkey,
        is_active => $milestone_obj->is_active,
      };
      if ($product->default_milestone eq $milestone_obj->name) {
        $milestone->{default} = 1;
      }
      push @$milestones, $milestone;
    }

    # Target Release
    my $releases = [];
    foreach my $release_obj (@{$product->releases}) {
      my $release = {
        id        => $release_obj->id,
        name      => $release_obj->name,
        sortkey   => $release_obj->sortkey,
        is_active => $release_obj->is_active,
      };
      if ($product->default_release eq $release_obj->name) {
        $release->{default} = 1;
      }
      push @$releases, $release;
    }

    # Versions
    my $versions = [];
    foreach my $version (@{$product->versions}) {
      push @$versions,
        {
        id        => $version->id,
        name      => $version->name,
        is_active => $version->is_active
        };
    }

    $product_data->{milestones} = $milestones if scalar(@$milestones);
    $product_data->{releases}   = $releases   if scalar(@$releases);
    $product_data->{versions}   = $versions   if scalar(@$versions);

    $product_data->{name} = $product->name;
    $product_data->{id}   = $product->id;

    push @$product_set, $product_data;
  }

  return $product_set;
}

=head2 C<product_delete_objects>

Delete a version, milestone or release.

=head3 C<params>

=over

=item B<versions>

B<array> I<optional> An array of version ids that should be deleted.

=item B<milestones>

B<array> I<optional> An array of milestone ids that should be deleted.

=item B<releases>

B<array> I<optional> An array of release ids that should be deleted.

=back

=cut

sub product_delete_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  my $changes;
  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $product_cache = {};

  foreach my $version_id (@{$params->{versions}}) {
    my $version = Bugzilla::Version->check({id => $version_id});

    _check_product($user, $version->product_id, $product_cache);

    push @{$changes->{versions}}, $version->id;

    $version->remove_from_db;
  }
  foreach my $milestone_id (@{$params->{milestones}}) {
    my $milestone = Bugzilla::Milestone->check({id => $milestone_id});

    _check_product($user, $milestone->product_id, $product_cache);

    push @{$changes->{milestones}}, $milestone->id;

    $milestone->remove_from_db;
  }
  foreach my $release_id (@{$params->{releases}}) {
    my $release = Bugzilla::Release->check({id => $release_id});

    _check_product($user, $release->product_id, $product_cache);

    push @{$changes->{releases}}, $release->id;

    $release->remove_from_db;
  }

  $dbh->bz_commit_transaction;

  return $changes;
}

sub _check_product {
  my ($user, $product_id, $product_cache) = @_;

  unless (defined($product_cache->{$product_id})) {
    my $product = Bugzilla::Product->check({id => $product_id});
    $user->check_can_admin_product($product->name);
    $product_cache->{$product->id} = 1;
  }

  return 1;
}


## REDHAT EXTENSION 1354240 END

## REDHAT EXTENSION START 1219299

=head2 C<search_as>

Convert a URI query string in to an RPC call.

=head3 C<params>

=over

=item B<format>

B<string> The format required, perl or json.

=item B<search>

B<string> The search string to convert.

=back

=cut

sub search_as {
  my ($self, $params) = @_;

  my $format = $params->{format} || die "no format";
  my $search = $params->{search} || die "no search";

  my @split = split(/&/, $search);

  my $rule = _rule(_sort_list(_squash_bug_status(_filter_junk(\@split))));
  my $str  = '';

  $str = _as_perl($rule) if ($format eq 'perl');
  $str = _as_json($rule) if ($format eq 'json');

  return ($str);
}

sub _squash_bug_status {
  my $list = shift;

  unless (any { $_ =~ /bug_status/ } @$list) {
    return $list;
  }

  my @combined_status;
  foreach my $status (@$list) {
    if ($status =~ m/^bug_status/) {
      my (undef, $st) = split(/=/, $status);
      push @combined_status, $st;
    }
  }

  my @list   = grep { $_ !~ /^bug_status/ } @$list;
  my $states = join(',', @combined_status);
  push @list, "bug_status=$states";
  return \@list;
}

sub _filter_junk {
  my $list = shift;
  my @junk = qw(list_id known_name query_based_on);

  my @newlist;

  foreach my $item (@$list) {
    next if $item =~ /=$/;
    push @newlist, $item unless any { $item =~ /^$_=/ } @junk;
  }

  return \@newlist;
}

sub _rule {
  my $list = shift;

  my %rule;
  my %fmap;

  foreach my $ws_field (keys %{Bugzilla::Bug->FIELD_MAP}) {
    my $int_field = Bugzilla::Bug->FIELD_MAP->{$ws_field};
    $fmap{$int_field} = $ws_field;
  }

  foreach my $item (@$list) {
    my ($key, $val) = split /=/, $item;
    $val = uri_unescape($val);
    if ($key eq 'columnlist') {
      my @cols;
      foreach my $col (split ',', $val) {
        $col = $fmap{$col} if (defined($fmap{$col}));
        push(@cols, $col);
      }
      $rule{include_fields} = \@cols;
    }
    else {
      $rule{$key} = $val;
    }
  }

  return (\%rule);
}

sub _as_perl {
  my $rule = shift;

  my $dumper = Data::Dumper->new([$rule]);
  $dumper->Sortkeys(1);
  $dumper->Terse(1);

  return ($dumper->Dump());
}

sub _as_json {
  my $rule = shift;

  my $json = JSON::XS->new->ascii->pretty->allow_nonref;
  $json->canonical(1);

  return ($json->encode($rule));
}

sub _sort_list {
  my $list = shift;

  my (@f, @o, @v, @other);

  foreach my $item (@$list) {
    if ($item =~ /^f(?:[0-9]+)=/) {
      push @f, $item;
    }
    elsif ($item =~ /^o(?:[0-9]+)=/) {
      push @o, $item;
    }
    elsif ($item =~ /^v(?:[0-9]+)=/) {
      push @v, $item;
    }
    else {
      push @other, $item;
    }
  }

  @f = sort @f;
  @o = sort @o;
  @v = sort @v;

  my @output;
  my $count = scalar(@f);
  for (my $i = 0; $i < $count; $i++) {
    push @output, $f[$i] if defined $f[$i];
    push @output, $o[$i] if defined $o[$i];
    push @output, $v[$i] if defined $v[$i];
  }
  push @other, @output;

  return \@other;
}
## REDHAT EXTENSION END 1219299

=head2 C<can_delete_user>

Report if a user can be deleted and if not, why not.

=head3 C<params>

=over

=item B<names>

B<array> of user names to check.

=back

=head3 C<returns>

Returns an array of hashes, one per user reported on. Each hash contains:

=over

=item C<name>

B<string> The user name

=item C<id>

B<int> The user name

=item C<can>

B<bool> Can the user be deleted?

=item C<not_ok_to_delete>

B<hash> A hash of counts of things that can prevent deletion.

=item C<ok_to_delete>

B<hash> A hash of counts of things that cannot prevent deletion.

=back

=cut

sub can_delete_user {
  my $self = shift;
  my ($params) = @_;


  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "users"});

  my $email = trim($params->{names})
    || ThrowCodeError('param_required', {param => 'email'});

  my @user_objects = map { Bugzilla::User->check($_) } @{$params->{names}};

  my @users;

  foreach my $user (@user_objects) {
    push(@users, $user->can_delete());
  }

  return {users => \@users};
}


=head2 C<delete_user>

Delete some users.

Users will only be deleted if the pass L<can_delete_user>.

=head3 C<params>

=over

=item B<names>

B<array> The user names to delete.

=back

=head3 C<returns>

Returns an array of hashes, one per user deleted on.

=cut

sub delete_user {
  my $self = shift;
  my ($params) = @_;


  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "users"});

  my $email = trim($params->{names})
    || ThrowCodeError('param_required', {param => 'email'});

  my @user_objects = map { Bugzilla::User->check($_) } @{$params->{names}};

  my @users;

  foreach my $user (@user_objects) {
    push(@users, $user->delete_me());
  }

  return {users => \@users};
}

=head2 C<terminate_sessions>

Terminate sessions for users.

You must be in the I<admin> group to use this API.

=head3 C<params>

=over

=item B<sessions>

B<array> of hashes with I<name> and I<ids> of sessions to terminate.

  {
    'sessions' =>
      [
        {
          'name' =>'fu@example.com',
          'ids'  => [11,56,78]
        },
        {
          'name' => 'bar@example.com',
          'ids'  => [34, 42]
        },
      ]
  }

=back

=head3 C<returns>

Returns a hash, with the user name as the key, which contains an array of sessions
terminated.

  {
    'terminated' =>
      {
        'fu@example.com'  => [11,56,78],
        'bar@example.com' => [34, 42],
      }
  }

=cut

sub terminate_sessions {
  my $self = shift;
  my ($params) = @_;

  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "terminate_sessions"});

  my $user_sessions = $params->{sessions}
    || ThrowCodeError('param_required', {param => 'sessions'});

  my %results;
  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  foreach my $user_session (@$user_sessions) {
    my @terminated;

    my $user_name = $user_session->{name}
      || ThrowCodeError('param_required', {param => 'name'});
    my $session_ids
      = $user_session->{ids} || ThrowCodeError('param_required', {param => 'ids'});

    my $user_object = Bugzilla::User->new({name => $user_name});

    my $sessions = Bugzilla::User::Session->new_from_list($session_ids);

    foreach my $session (@$sessions) {
      if ($session->userid == $user_object->id) {
        $session->remove_from_db;
        push(@terminated, $session->id);
      }
    }

    if (@terminated) {
      $results{'terminated'}->{$user_name} = \@terminated;
    }
  }
  $dbh->bz_commit_transaction;

  return (\%results);
}

=head2 C<admin_apikeys>

Ban and revoke APIKEYs for users.

You must be in the I<admin> group to use this API.

=head3 C<params>

=over

=item B<apikeys>

B<array> of hashes with I<name> and I<ids> of APIKEYs to ban or revoke.

  {
    'apikeys' =>
      [
        {
          'name'        => 'fu@example.com',
          'banned_ids'  => [11,56],
          'revoked_ids' => [78]
        },
        {
          'name'        => 'bar@example.com',
          'banned_ids'  => [34, 42]
        },
      ]
  }

=back

=head3 C<returns>

Returns an array of hashes for APIKEYs that have been banned or revoked.

  [
    {
      'name'        => 'fu@example.com',
      'banned_ids'  => [11,56],
      'revoked_ids' => [78]
    },
    {
      'name'       => 'bar@example.com',
      'banned_ids' => [34, 42],
    }
  ]

=cut

sub admin_apikeys {
  my $self = shift;
  my ($params) = @_;

  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "ban_apikeys"});

  my $user_apikeys = $params->{apikeys}
    || ThrowCodeError('param_required', {param => 'apikeys'});

  my @results;
  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  foreach my $user_apikey (@$user_apikeys) {
    my @banned;
    my @revoked;


    my $user_name
      = $user_apikey->{name} || ThrowCodeError('param_required', {param => 'name'});
    my $banned_ids  = $user_apikey->{banned_ids};
    my $revoked_ids = $user_apikey->{revoked_ids};

    ThrowCodeError('params_required', {params => ['banned_ids', 'revoked_ids']})
      unless (@$banned_ids || @$revoked_ids);

    my %changes = ('name' => $user_name,);

    my $user_object = Bugzilla::User->new({name => $user_name});

    my $apikeys = Bugzilla::User::APIKey->new_from_list($banned_ids);

    foreach my $apikey (@$apikeys) {
      if ($apikey->user_id == $user_object->id) {
        $apikey->set_banned(1);
        $apikey->update();
        push(@banned, $apikey->id);
      }
      else {
        ThrowUserError('api_key_ban_failed',
          {user_name => $user_name, apikey_id => $apikey->id});
      }
    }

    if (@banned) {
      $changes{'banned'} = \@banned;
    }

    $apikeys = Bugzilla::User::APIKey->new_from_list($revoked_ids);

    foreach my $apikey (@$apikeys) {
      if ($apikey->user_id == $user_object->id) {
        $apikey->set_revoked(1);
        $apikey->update();
        push(@revoked, $apikey->id);
      }
      else {
        ThrowUserError('api_key_ban_failed',
          {user_name => $user_name, apikey_id => $apikey->id});
      }
    }

    if (@revoked) {
      $changes{'revoked'} = \@revoked;
    }

    if (@banned || @revoked) {
      push(@results, \%changes);
    }
  }

  $dbh->bz_commit_transaction;

  return (\@results);
}

## REDHAT EXTENSION START 1404885

=head2 C<edit_comment>

Edit a comment on a bug.

The rules for this are

=over 4

=item A security or admin user can edit any comment.

=item A security or admin user can hide previous comment revisions when editing a
comment.

=item A reporter may edit the description, comment 0, on their bugs.

=item A user with I<editbugs> privileges on a product may edit the description comment
of bugs in that product.

=back

When an edit is made an entry is added to the bugs_activity table indicating who
made the change and what comment number was updated. The text of the change is not
included in this entry. When a user views the bug history, either on the bug
history page or when using the inline history feature, they will be able to see
who edited a comment and when, but not the change made.

The texts are stored in a separate table and when entries from this table are
viewed the ACLs are applied. If you lack the authorization to see a change then
the text will be replaced with the string I<REDACTED>.

The security flag is retroactive, it hides all previous revisions of a comment.

Ths feature is accessed in the UI on the bug edit page by clicking the edit icon
in the comment header. If you cannot see this icon on a comment then you do not
have permission to edit that comment.

The security flag is set in the UI by selecting the I<Hide history> checkbox on
the edit form. If you cannot see this checkbox then you do not have access to lock the
comment history.

If you try and edit a comment or set the security flag and you do not have access
to do so the API will throw an error and reject the entire request.

=head3 C<params>

comment:
  B<int> I<optional> The comment ID. This or bug are required.

bug:
  B<int> I<optional> The bug ID. This or comment are required.

number:
  B<int> I<optional> The comment number on a bug to update.
  This is only used if bug is supplied.
  Defaults to zero.

text:
  B<string> The text to change the comment to.

security:
  B<boolean> I<optional> Hide previous versions of this comment due to security concerns.
  B<Note:> You need to be in the admin or security groups to set this parameter

render:
  B<boolean> I<optional> Return the rendered HTML for this comment.

new_token:
  B<boolean> I<optional> Return a new bug edit token.

=head3 C<returns>

Returns a hash, containing:

activity_count:
  B<int> The number of times the comment has been edited.

delta_ts:
  B<bool> The new last changed time stamp for the bug.

longdesc:
  B<array> A array of 2 things. The first is the comment number on the bug. The
  second is what happened to the comment. Currently only 'updated'.
  This text will appear in the bug history and inline history if enabled.

html:
  B<string> If requested the HTML formatted version of the comment.
  This is what a user will see when loading the updated comment in the web UI.

new_token:
  B<string> If requested, a token that can be used for API calls to edit the
  current bug.

=head3 C<Examples>

=head4 JSONRPC call

=over 4

=item By comment id

  {
      "jsonrpc": "2.0",
      "method": "RedHat.edit_comment",
      "id": 2,
      "params": [
        {
          "comment": 13178344,
          "new_token": 1
          "render": 1,
          "security": 0,
          "text": "Description of problem:\nThe old comment is just no good!\nBug 11\n",
        }
      ]
  }

=item By bug id & comment number

  {
      "jsonrpc": "2.0",
      "method": "RedHat.edit_comment",
      "id": 3,
      "params": [
        {
          "bug": 42,
          "comment": 3,
          "text": "Description of problem:\nExamples are hard\n",
        }
      ]
  }

=back

=head4 JSONRPC reply

=over 4

=item By comment id

  {
      "id": 2,
      "error": null,
      "result": {
          "activity_count": 2,
          "delta_ts": "2020-03-06 03:38:50",
          "html": "Description of problem:\nThe old comment is just no good!\n<a href=\"https://bugzilla.redhat.com/show_bug.cgi?id=11\">https://bugzilla.redhat.com/show_bug.cgi?id=11</a>\n"
          "longdesc": [0, "updated"],
          "new_token": "111-222-333-444-555",
      }
  }

=item By bug id & comment number

  {
      "id": 3,
      "error": null,
      "result": {
          "activity_count": 4,
          "delta_ts": "2020-03-06 03:38:50",
          "longdesc": [3, "updated"],
      }
  }

=back

=cut

sub edit_comment {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $text      = $params->{text} || die "no text";
  my $security  = $params->{security} // 0;
  my $new_token = $params->{new_token} // 0;
  my $number    = $params->{number} // 0;
  my $render    = $params->{render} // 0;

  if ($security && !$user->can_edit_comments()) {
    ThrowUserError('only_security_can_hide_comment_history');
  }

  $security = $security == 1 ? 1 : 0;

  my $bug_id     = $params->{bug};
  my $comment_id = $params->{comment};

  if (!$bug_id && !$comment_id) {
    ThrowCodeError('params_required', {params => ['bug', 'comment']});
  }

  my ($comment, $bug);

  if ($comment_id) {
    $comment = Bugzilla::Comment->check({id => $comment_id});
    $bug     = $comment->bug;
  }
  else {
    $bug     = Bugzilla::Bug->check($bug_id);
    $comment = $bug->comments({order => 'oldest_to_newest'})->[$number];

    unless ($comment) {
      ThrowUserError('no_such_comment_number',
        {number => $number, bug_id => $bug->id});
    }

    if ($comment->count() != $number) {
      ThrowUserError('comment_number_mismatch',
        {requested => $number, actual => $comment->count()});
    }
  }

  $bug->check_is_visible();
  $bug->edit_comment($comment->id, $text, $security, $comment->count());

  my ($changes, $delta_ts, $old_bug) = $bug->update();
  $changes->{delta_ts} = $self->type('dateTime', $bug->delta_ts);
  delete $comment->{'activity'};

  # skip mail for security changes
  if ($security) {
    Bugzilla->dbh->do('UPDATE bugs SET lastdiffed = ? WHERE bug_id = ?',
      undef, ($delta_ts, $bug->id));
  }
  else {
    $bug->send_changes($changes, {edited_comment => $comment});
  }

  $changes->{activity_count} = scalar @{$comment->activity} - 1;

  if ($render) {
    my $html = Bugzilla::Template::quoteUrls($text, $bug);
    $changes->{html} = $html;
  }

  if ($new_token) {
    $changes->{new_token} = issue_hash_token([$bug->id, $bug->delta_ts]);
  }

  return ($changes);
}
## REDHAT EXTENSION END 1404885

1;

__END__

=head1 SEE ALSO

L<Bugzilla|Bugzilla> L<Bugzilla::WebService> L<Bugzilla::Extension::RedHat>
