# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Releases;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Token;
use Bugzilla::Util qw(trim);
use Bugzilla::Error;

use Bugzilla::Extension::Releases::Release;
use Bugzilla::Extension::Releases::Components;
use Bugzilla::Extension::Releases::Schedule;
use Bugzilla::Extension::Releases::Util;
use Bugzilla::Extension::Releases::ACKView;
use Bugzilla::Extension::Releases::ACKList;

################
# Installation #
################

our $VERSION = 0.3;

BEGIN {
  *Bugzilla::User::can_edit_releases = \&_user_can_edit_releases;
}

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  ## REDHAT EXTENSION START 463810
  $schema->{rh_release_components} = {
    FIELDS => [
      id         => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      release_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'rh_release', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      component_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'components', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      component_type => {TYPE => 'varchar(255)', NOTNULL => 1},
    ],
    INDEXES => [
      rh_release_components_unique_idx =>
        {FIELDS => [qw(release_id component_id)], TYPE => 'UNIQUE'},
    ],
  };
  ## REDHAT EXTENSION END 463810

  ## REDHAT EXTENSION START
  $schema->{rh_release_component_tags} = {
    FIELDS => [
      id                   => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      release_component_id => {
        TYPE    => 'INT3',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'rh_release_components', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      tag => {TYPE => 'varchar(255)', NOTNULL => 1},
    ],
    INDEXES => [
      rh_release_component_tags_unique_idx =>
        {FIELDS => [qw(release_component_id tag)], TYPE => 'UNIQUE'},
    ],
  };
  ## REDHAT EXTENSION END

  ## REDHAT EXTENSION START 465338
  $schema->{rh_release} = {
    FIELDS => [
      id      => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      type_id => {
        TYPE       => 'INT2',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'flagtypes', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      description => {TYPE => 'varchar(255)', NOTNULL => 1},
      sortkey     => {TYPE => 'INT2',         NOTNULL => 1, DEFAULT => 0},
      is_active   => {TYPE => 'BOOLEAN',      NOTNULL => 1, DEFAULT => 'TRUE'}
    ],
    INDEXES =>
      [rh_release_unique_idx => {FIELDS => [qw(id type_id)], TYPE => 'UNIQUE'},],
  };
  ## REDHAT EXTENSION END 465338

  ## REDHAT EXTENSION START 465357
  $schema->{rh_release_schedule} = {
    FIELDS => [
      id         => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      release_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'rh_release', COLUMN => 'id', DELETE => 'CASCADE'}
      },
    ],
    INDEXES => [
      rh_release_schedule_unique_idx =>
        {FIELDS => [qw(id release_id)], TYPE => 'UNIQUE'},
    ],
  };

  $schema->{rh_release_schedule_phase} = {
    FIELDS => [
      schedule_id => {
        TYPE    => 'INT3',
        NOTNULL => 1,
        REFERENCES =>
          {TABLE => 'rh_release_schedule', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      phase      => {TYPE => 'varchar(255)', NOTNULL => 1},
      date_start => {TYPE => 'DATETIME',     NOTNULL => 1},
      date_end   => {TYPE => 'DATETIME',     NOTNULL => 1},
    ],
    INDEXES => [
      rh_release_schedule_phase_unique_idx =>
        {FIELDS => [qw(schedule_id phase)], TYPE => 'UNIQUE'},
    ],
  };
  ## REDHAT EXTENSION END 465357

  ## REDHAT EXTENSION START 461973
  $schema->{rh_release_activity} = {
    FIELDS => [
      release_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'rh_release', COLUMN => 'id', DELETE => 'CASCADE'}
      },
      who => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'}
      },
      what     => {TYPE => 'varchar(255)',  NOTNULL => 1},
      act_when => {TYPE => 'DATETIME',      NOTNULL => 1},
      act_type => {TYPE => 'varchar(255)',  NOTNULL => 1},
      added    => {TYPE => 'varchar(4000)', NOTNULL => 1},
      removed  => {TYPE => 'varchar(4000)', NOTNULL => 1},
    ],
    INDEXES => [rh_release_activity_idx => ['release_id', 'act_when', 'act_type'],],
  };
  ## REDHAT EXTENSION END 461973

  ## REDHAT EXTENSION START 1211439
  $schema->{rh_release_ackview_settings} = {
    FIELDS => [
      userid => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE'}
      },
      value => {TYPE => 'TEXT', NOTNULL => 1},
    ],
  };
  ## REDHAT EXTENSION END 1211439

  return;
}


#########
# Pages #
#########

sub _user_can_edit_releases {
  my $self = shift;

  if (!defined $self->{'can_edit_releases'}) {
    $self->{'can_edit_releases'}
      = ($self->in_group('admin') || $self->in_group('program_management'));
  }

  return ($self->{'can_edit_releases'});
}

sub user_can_administer {
  my ($self, $args) = @_;

  my $can_see_admin = $args->{can_see_admin};

  if (Bugzilla->user->can_edit_releases()) {
    $$can_see_admin = 1;
  }

  return;
}

sub page_before_template {
  my ($self, $args) = @_;

  my $page = $args->{page_id};
  my $vars = $args->{vars};

  if ($page =~ m{^releases/releases\.}) {
    _page_releases_list($vars);
  }
  if ($page =~ m{^releases/activity\.}) {
    _page_activity($vars);
  }
  elsif ($page =~ m{^releases/components/list\.}) {
    _page_components_list($vars);
  }
  elsif ($page =~ m{^releases/components/edit\.}) {
    _page_components_edit($vars);
  }
  elsif ($page =~ m{^releases/schedule/list\.}) {
    _page_schedule_list($vars);
  }
  elsif ($page =~ m{^releases/schedule/edit\.}) {
    _page_schedule_edit($vars);
  }
  elsif ($page =~ m{^releases/ack_view\.}) {
    _page_ack_view($vars);
  }
  elsif ($page =~ m{^releases/ack_list\.}
    || $page =~ m{^releases/ack_edit_columns.*})
  {
    _page_ack_list($vars);
  }

  return;
}

sub _page_ack_view {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;
  $user->is_insider() || ThrowUserError(
    "auth_failure",
    {
      group  => Bugzilla->params->{'insidergroup'},
      action => "ack_view",
      object => "releases"
    }
  );

  my $av = Bugzilla::Extension::Releases::ACKView->new();
  $av->extended_view($vars);

  return;

}

sub _page_ack_list {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  my $page = $input->{page} // '';

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;
  $user->is_insider() || ThrowUserError(
    "auth_failure",
    {
      group  => Bugzilla->params->{'insidergroup'},
      action => "ack_list",
      object => "releases"
    }
  );

  my $av = Bugzilla::Extension::Releases::ACKList->new();
  if ($page eq 'flag-bugs-submit') {
    $av->flag_bugs_submit($vars);
  }
  elsif ($page eq '' || $page eq 'flag-bugs') {
    $av->show_flag_bugs($vars);
  }
  elsif ($page eq 'edit-columns') {
    $av->edit_columns($vars);
  }
  elsif ($page eq 'commit-columns') {
    $av->commit_columns($vars);
  }
  else {
    ThrowUserError("bad_ackview_page", {page => $page});
  }

  return;

}

sub _page_releases_list {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "releases"});

  # often used variables
  my $token = $input->{'token'};

  # Perform any updating needed from selected releases in the web form
  my %changes = (removed => [], added => [], updated => []);

  if ($input->{'set_releases'}) {
    check_token_data($token, 'edit_releases');

    my $timestamp = $dbh->selectrow_array('SELECT NOW()');

    # Add new flags as release
    foreach my $flag_id (grep(/^type_id-\d+$/, keys %$input)) {
      $flag_id =~ s/^type_id-//;

      my $flag = Bugzilla::FlagType->new($flag_id)
        || ThrowCodeError('flag_type_nonexistent', {id => $flag_id});

      my $release
        = Bugzilla::Extension::Releases::Release->new({release_id => $flag->id});

      if ($release) {
        my $updated = 0;
        if (!$release->is_active) {
          $release->set_is_active(1);
          push(@{$changes{'added'}}, $flag->name);
          $updated = 1;
        }
        if ($input->{"release_description-" . $flag->id} ne $release->description) {
          $release->set_description($input->{"release_description-" . $flag->id});
          $updated = 1;
        }
        if ($input->{"release_sortkey-" . $flag->id} ne $release->sortkey) {
          $release->set_sortkey($input->{"release_sortkey-" . $flag->id});
          $updated = 1;
        }
        if ($updated) {
          push(@{$changes{'updated'}}, $flag->name);
          $release->update();
        }
      }
      else {
        my $release = Bugzilla::Extension::Releases::Release->create({
          type_id     => $flag->id,
          description => $input->{"release_description-" . $flag->id},
          sortkey     => $input->{"release_sortkey-" . $flag->id}
        });
        push(@{$changes{'added'}}, $flag->name);
        log_activity({
          act_when   => $timestamp,
          act_type   => 'release',
          release_id => $release->id,
          who        => Bugzilla->user->id,
          what       => 'release',
          removed    => '',
          added      => $flag->name
        });

      }
    }

    my $current_releases
      = Bugzilla::Extension::Releases::Release->match({is_active => 1});

    # Remove release no longer selected
    foreach my $release (@{$current_releases}) {
      if (!$input->{"type_id-" . $release->type_id}) {
        my $flag = Bugzilla::FlagType->new($release->type_id);

        # If a user can't see a release, then they can't remove it.
        next unless Bugzilla->user->can_see_flag($flag);
        push(@{$changes{'removed'}}, $flag->name);
        $release->set_is_active(0);
        $release->update();
        log_activity({
          act_when   => $timestamp,
          act_type   => 'release',
          release_id => $release->id,
          who        => Bugzilla->user->id,
          what       => 'release',
          removed    => $flag->name,
          added      => ''
        });
      }
    }

    delete_token($token);
  }

  $vars->{'token'}       = issue_session_token('edit_releases');
  $vars->{'doc_section'} = 'extensions/Releases/index-admin.html';
  $vars->{'doc_title'}   = 'Releases';
  $vars->{'changes'}     = \%changes;

  # List of all flags
  my @flags;
  my $all_flags
    = Bugzilla::FlagType::match({target_type => 'bug', is_active => 1});

  foreach my $flag (@{$all_flags}) {
    next if($flag->name =~ m/^approvals_request_/);

    # If a user can't see a flag, don't show them the flag.
    next unless Bugzilla->user->can_see_flag($flag);
    $flag->{'release'}
      = Bugzilla::Extension::Releases::Release->new({release_id => $flag->id});
    push(@flags, $flag);
  }

  @flags = sort {lc($a->name) cmp lc($b->name) } @flags;

  $vars->{'flags'} = \@flags;

  return;
}

sub _page_activity {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "releases"});

  my $id   = $input->{'release_id'};
  my $type = $input->{'type'};

  my $release;
  if ($id) {
    $release = Bugzilla::Extension::Releases::Release->new($id);
    $release || ThrowCodeError('flag_type_nonexistent', {id => $id});
  }

  $vars->{'release'}    = $release;
  $vars->{'type'}       = $type;
  $vars->{'operations'} = get_activity({release_id => $id, act_type => $type});

  return;
}

sub _page_components_list {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "releases"});

  $vars->{'doc_section'} = 'extensions/Releases/Components.html';
  $vars->{'doc_title'}   = 'Release Components';

  $vars->{'releases'}
    = Bugzilla::Extension::Releases::Release->match({is_active => 1});

  return;
}

sub _page_components_edit {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "releases"});

  my $id    = $input->{'release_id'};
  my $token = $input->{'token'};

  $id || ThrowCodeError('param_required', {param => 'release_id'});

  my $release = Bugzilla::Extension::Releases::Release->new($id);
  $release || ThrowCodeError('flag_type_nonexistent', {id => $id});

  # Create a release components object
  my $release_components = Bugzilla::Extension::Releases::Components->new(
    {release => $release, preload => 1});

  # Perform any updating needed from select components in the web form
  my %changes;

  # Check for midair collision
  if ($input->{'update_approved'} || $input->{'update_capacity'}) {
    my $operations = get_activity({
      release_id => $id,
      act_type   => 'component',
      starttime  => $input->{'timestamp'}
    });
    if (@$operations && !$input->{'submit_anyway'}) {
      $vars->{'action'}      = 'midair';
      $vars->{'operations'}  = $operations;
      $vars->{'release'}     = $release;
      $vars->{'script_name'} = "page.cgi?id=releases/components/edit.html";
      return;
    }
  }

  if ($input->{'update_approved'}) {
    check_token_data($token, 'release_components');

    # 'select' box components
    my %approved;
    $input->{'non_approved_capacity'} ||= [];
    my @non_approved_capacity
      = ref $input->{'non_approved_capacity'}
      ? @{$input->{'non_approved_capacity'}}
      : ($input->{'non_approved_capacity'});
    foreach my $comp (@non_approved_capacity) {
      $approved{$comp} = 1 if $comp;
    }

    # 'textarea' components
    my %approved_tags;
    foreach my $comp (split(/\n/, $input->{'approved'})) {
      $comp = trim($comp);
      next if !$comp || $comp eq '';

      # Remove any optional "tags" that people can add to the right of a component
      # Format: component_name [tag1 tag2 ...]
      my @tags;
      if ($comp =~ s/\s+\[(.*)\]//) {
        my $all_tags = $1;
        @tags = split(/[\s,]+/, $all_tags);
      }
      $approved_tags{$comp} = \@tags if @tags;
      $approved{$comp}      = 1      if $comp;
    }

    my $data_ref = {
      release => $release,    # pass $release instead of $id to save db calls
      approved        => [sort keys %approved],
      approved_tags   => \%approved_tags,
      approved_action => 'makeexact',
      allow_empty     => $input->{'allow_empty'} ? 1 : 0,
    };

    my $changes_ref = $release_components->set_release_components($data_ref);
    $changes{'approved'} = $changes_ref->{'approved'};

    delete_token($token);
  }

  if ($input->{'update_capacity'}) {
    check_token_data($token, 'release_components');

    # 'select' box components
    my %capacity;
    $input->{'non_approved_capacity'} ||= [];
    my @non_approved_capacity
      = ref $input->{'non_approved_capacity'}
      ? @{$input->{'non_approved_capacity'}}
      : ($input->{'non_approved_capacity'});
    foreach my $comp (@non_approved_capacity) {
      $capacity{$comp} = 1 if $comp;
    }

    # 'textarea' components
    my %capacity_tags;
    foreach my $comp (split(/\n/, $input->{'capacity'})) {
      my $comp = trim($comp);
      next if !$comp || $comp eq '';

      # Remove any optional "tags" that people can add to the right of a component
      # Format: component_name [tag1 tag2 ...]
      my @tags;
      if ($comp =~ s/\s+\[(.*)\]//) {
        my $all_tags = $1;
        @tags = split(/[\s,]+/, $all_tags);
      }

      $capacity_tags{$comp} = \@tags if @tags;
      $capacity{$comp}      = 1      if $comp;
    }

    my $data_ref = {
      release => $release,    # pass $release instead of $id to save db calls
      capacity        => [sort keys %capacity],
      capacity_tags   => \%capacity_tags,
      capacity_action => 'makeexact',
      allow_empty     => $input->{'allow_empty'} ? 1 : 0,
    };

    my $changes_ref = $release_components->set_release_components($data_ref);
    $changes{'capacity'} = $changes_ref->{'capacity'};

    delete_token($token);
  }

  my %non_approved_capacity = %{$release_components->components};

  my %unique_approved
    = map { $_->{'name'} => 1 } @{$release_components->approved};
  foreach my $comp (keys %unique_approved) {
    delete $non_approved_capacity{$comp};
  }

  my %unique_capacity
    = map { $_->{'name'} => 1 } @{$release_components->capacity};
  foreach my $comp (keys %unique_capacity) {
    delete $non_approved_capacity{$comp};
  }

  $vars->{'timestamp'}             = $dbh->selectrow_array('SELECT NOW()');
  $vars->{'non_approved_capacity'} = [sort keys %non_approved_capacity];
  $vars->{'approved'}              = $release_components->approved;
  $vars->{'approved_count'}        = scalar(@{$release_components->approved});
  $vars->{'capacity'}              = $release_components->capacity;
  $vars->{'capacity_count'}        = scalar(@{$release_components->capacity});
  $vars->{'release'}               = $release;
  $vars->{'token'}                 = issue_session_token('release_components');
  $vars->{'doc_section'}           = 'extensions/Releases/Components.html';
  $vars->{'doc_title'}             = 'Release Components';
  $vars->{'changes'}               = \%changes;

  return;
}

sub _page_schedule_list {
  my ($vars) = @_;
  my $input = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "releases"});

  $vars->{'doc_section'} = 'extensions/Releases/Schedule.html';
  $vars->{'doc_title'}   = 'Release Schedules';

  $vars->{'releases'}
    = Bugzilla::Extension::Releases::Release->match({is_active => 1});

  return;
}

sub _page_schedule_edit {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;

  # Preliminary checks:
  Bugzilla->login(LOGIN_REQUIRED);

  my $user = Bugzilla->user;
  $user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "releases"});

  my $id    = $input->{'release_id'};
  my $token = $input->{'token'};

  $id || ThrowCodeError('param_required', {param => 'release_id'});

  my $release = Bugzilla::Extension::Releases::Release->new($id);
  $release || ThrowCodeError('flag_type_nonexistent', {id => $id});

  my $schedule = Bugzilla::Extension::Releases::Schedule->new(
    {release_id => $release->flag_type->id});

  my $changes = {};

  if ($input->{'update_schedule'}) {
    check_token_data($token, 'edit_schedule');

    # Check for midair collision
    my $operations = get_activity({
      release_id => $id,
      act_type   => 'schedule',
      starttime  => $input->{'timestamp'}
    });
    if (@$operations) {
      $vars->{'action'}      = 'midair';
      $vars->{'operations'}  = $operations;
      $vars->{'release'}     = $release;
      $vars->{'script_name'} = "page.cgi?id=releases/schedule/edit.html";
      return;
    }

    my $phases = {};
    foreach my $phase (Bugzilla::Extension::Releases::Schedule->get_legal_phases())
    {
      next if !$input->{"date_start-$phase"} && !$input->{"date_end-$phase"};
      $input->{"date_start-$phase"}
        || ThrowUserError('illegal_date',
        {date => 'date start', format => 'YYYY-MM-DD'});
      $input->{"date_end-$phase"}
        || ThrowUserError('illegal_date',
        {date => 'date end', format => 'YYYY-MM-DD'});
      $phases->{$phase} = {
        date_start => $input->{"date_start-$phase"},
        date_end   => $input->{"date_end-$phase"}
      };
    }

    if (!$schedule) {
      $schedule = Bugzilla::Extension::Releases::Schedule->create(
        {release_id => $release->id, phases => $phases});
      $changes->{phases} = {};
      foreach my $phase (keys %$phases) {
        $changes->{phases}{added}{"$phase start"} = ["", $phases->{$phase}{date_start}];
        $changes->{phases}{added}{"$phase end"}   = ["", $phases->{$phase}{date_end}];
      }
    }
    else {
      $schedule->set_phases($phases);
      $changes = $schedule->update();
    }

    delete_token($token);
  }

  $vars->{'timestamp'}   = $dbh->selectrow_array('SELECT NOW()');
  $vars->{'token'}       = issue_session_token('edit_schedule');
  $vars->{'doc_section'} = 'extensions/Releases/Schedule.html';
  $vars->{'doc_title'}   = 'Release Schedules';
  $vars->{'changes'}     = $changes;
  $vars->{'release'}     = $release;
  $vars->{'schedule'}    = $schedule;
  $vars->{'phase_names'}
    = [Bugzilla::Extension::Releases::Schedule::get_legal_phases()];

  return;
}

###############
# WebServices #
###############

sub webservice {
  my ($self, $args) = @_;

  my $dispatch = $args->{dispatch};
  $dispatch->{Releases} = "Bugzilla::Extension::Releases::WebService";

  return;
}

## REDHAT EXTENSION START 731285
# Add the ability to have a custom search based on the approved and capicity
# component list
sub search_operator_field_override {
  my ($self, $args) = @_;

  my $operators = $args->{'operators'};

  $operators->{component_a} = {
    _non_changed => sub { _component_list_nonchanged(@_, 'approved') }
  };
  $operators->{component_c} = {
    _non_changed => sub { _component_list_nonchanged(@_, 'capacity') }
  };

  return;
}

sub _component_list_nonchanged {
  my ($invocant, $args, $word) = @_;
  my $dbh = Bugzilla->dbh;

  # If you are not in the redhat group, you can't do this
  # This is also checked in Bugzilla::Search, but it doesn't hurt to
  # do it again
  if (not $invocant->_user->is_insider()) {
    $args->{term} = '1=2';
    return;
  }

  # First thing we do is get a list of components that match the specified
  # criteria
  my $new_args = {
    operator   => $args->{operator},
    value      => $args->{value},
    quoted     => $args->{quoted},
    field      => 'name',
    full_field => 'f.name',
  };
  $invocant->_do_operator_function($new_args);

  my $query = "
        SELECT DISTINCT c.name
        FROM components c
            JOIN rh_release_components rc ON (c.id = rc.component_id)
            JOIN rh_release r ON (rc.release_id = r.id)
            JOIN flagtypes f ON (r.type_id = f.id)
        WHERE r.is_active = 1
            AND rc.component_type = ?
            AND " . $new_args->{term};

  my $components = $dbh->selectcol_arrayref($query, undef, $word);

  # If there is no match, make sure nothing is shown.
  if (scalar(@$components) == 0) {
    $args->{term} = '1=2';
    return;
  }

  # Then we use this list to generate the necessary SQL that uses these
  # components
  my @quoted_components = map { $dbh->quote($_) } @$components;
  $query = '('
    . Bugzilla::Search::build_subselect("bugs.component_id", "components.id",
    "components", $dbh->sql_in('components.name', \@quoted_components))
    . ' OR '
    .


    Bugzilla::Search::build_subselect(
    "bugs.bug_id",
    "mbc.bug_id",
    "rh_multiple_bug_components mbc JOIN components c2 ON (mbc.component_id = c2.id)",
    $dbh->sql_in('c2.name', \@quoted_components)
    ) . ')';

  Bugzilla->logger->error($query);

  $args->{term} = $query;

  return;
}
## REDHAT EXTENSION END 731285

__PACKAGE__->NAME;
