# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Releases::ACKView;

use warnings;
use strict;
use 5.10.1;

use base qw(Bugzilla::Object);
our $VERSION = '0.2';

use Bugzilla;
use Bugzilla::Extension::Releases::Util;
use Bugzilla::Group;

sub new {
  my $invocant = shift;
  my $class    = ref($invocant) || $invocant;
  my $params   = shift;
  my $self     = {};
  bless($self, $class);

  $self->{config} = ack_config();

  return $self;
}

sub get_ack_releases {
  my ($self, $param) = @_;

  my $show_releases          = $param->{show_releases} || [];
  my $show_inactive_releases = untaint($param->{show_inactive_releases});

  my $releases_ref
    = get_releases({'show_inactive_releases' => $show_inactive_releases});

  if (!@$releases_ref) {
    die 'No releases specified.';
  }

  if (is_empty_array($show_releases)) {
    return @$releases_ref;
  }

  my @ack_releases = ();

  foreach my $release (@$releases_ref) {
    for (my $i = 0; $i < @$show_releases; $i++) {
      if ($release eq $show_releases->[$i]) {
        push @ack_releases, $release;
        last;
      }
    }
  }

  if (!@ack_releases) {
    die 'No valid releases specified.';
  }

  return @ack_releases;
}

sub generate_overview_sql {
  my ($self, $param) = @_;
  my $columns  = $param->{columns};
  my $releases = $param->{releases};
  my $resolved = $param->{resolved};
  my $groups   = $param->{groups};
  my $member   = $param->{member};
  my $customer = $param->{customer};
  my $partner  = $param->{partner};
  my $flags    = $param->{flags};
  my $values   = $param->{flag_values};
  my $extras   = $param->{extras};

  my $dbh = Bugzilla->dbh();

  # prepare subquery statements for flags

  my %flag_queries;
  foreach my $flag (@$flags) {
    my $quote_flag = $dbh->quote($flag);
    my $query      = <<SQL_QUERY;
        SELECT
          flags.bug_id AS bug_id,
          flags.status AS status
        FROM flagtypes
          LEFT OUTER JOIN flags ON flags.type_id=flagtypes.id
        WHERE
          flagtypes.name=$quote_flag
SQL_QUERY

    $query
      .= " AND ("
      . join(' OR ',
      map { "status=" . $dbh->quote($_) } split(//, $values->{$flag}))
      . ")"
      if $values && $values->{$flag};
    $flag_queries{$flag} = $query;
  }

  # construct sql query

  my $select   = "SELECT flagtypes.name, flags.status";
  my $from     = "FROM flagtypes, flags, bugs";
  my $where    = "WHERE flagtypes.id=flags.type_id AND flags.bug_id=bugs.bug_id";
  my $group_by = "GROUP BY flagtypes.name, flags.status";
  my $order_by = "ORDER BY flagtypes.name";

  # additional select columns

  if ($extras) {
    foreach my $extra (@$extras) {
      if ($extra =~ /^\w+$/) {
        $select   .= ", bugs.$extra";
        $group_by .= ", bugs.$extra";
      }
      elsif ($extra =~ / AS (\w+)$/) {
        $group_by .= ", " . $1;
        $select   .= ", $extra";
      }
      else {
        die 'Invalid select column: ' . $extra;
      }
    }
  }

  # consider flags

  foreach my $flag (@$flags) {
    $select .= ", tbl_$flag.status AS $flag";
    $from
      .= " LEFT OUTER JOIN ($flag_queries{$flag}) AS tbl_$flag ON bugs.bug_id=tbl_$flag.bug_id";
    $group_by .= ", $flag";
  }

  # consider resolved

  $select
    .= ", ("
    . join(' OR ', map { "bugs.bug_status=" . $dbh->quote($_) } @$resolved)
    . ") AS resolved";
  $group_by .= ", resolved";

  # consider columns
  foreach my $column (@$columns) {

    # retrieve column definitions, and extract keywords if exist
    my $id         = 'BLOCKER_COLUMN_' . $column;
    my $column_sql = $self->{config}->{$id};
    if (!defined $column_sql) {
      die 'Sql clause for column ' . $column . ' not defined in configuration file.';
    }
    if ($column_sql =~ /\s+OR\s+/) {
      die 'Sql clause for column ' . $column . ' is not supported: ' . $column_sql;
    }

    my $sql_column       = '';
    my @not_use_keywords = ();
    my @sql_conjunction  = split(/\s+AND\s+/, $column_sql);
    foreach my $item (@sql_conjunction) {
      if ($item !~ /^[\w.]+\s*(=|!=)\s*'?[^']+'?$/) {
        die 'Sql clause within column ' . $column . ' is not supported: ' . $item;
      }

      if ($item =~ /^keyworddefs.name\s*=\s*'([^']+)'$/) {
        my $keyword_query
          = "SELECT keywords.bug_id FROM keywords, keyworddefs WHERE keywords.keywordid=keyworddefs.id AND keyworddefs.name="
          . $dbh->quote($1);
        $sql_column .= " AND " if $sql_column;
        $sql_column .= "bugs.bug_id IN ($keyword_query)";
      }
      elsif ($item =~ /^keyworddefs.name\s*!=\s*'([^']+)'$/) {
        push @not_use_keywords, $1;
      }
      else {
        $sql_column .= " AND " if $sql_column;
        $sql_column .= $item;
      }
    }

    if (@not_use_keywords) {
      my $keyword_query
        = "SELECT keywords.bug_id FROM keywords, keyworddefs WHERE keywords.keywordid=keyworddefs.id AND (";
      $keyword_query .= join(' OR ',
        map { "keyworddefs.name=" . $dbh->quote($_) } @not_use_keywords);
      $keyword_query .= ")";
      $sql_column    .= " AND " if $sql_column;
      $sql_column    .= "bugs.bug_id NOT IN ($keyword_query)";
    }

    $select   .= ", ($sql_column) AS _$column";
    $group_by .= ", _$column";
  }

  # restrict by releases

  $where .= " AND ("
    . join(' OR ', map { "flagtypes.name=" . $dbh->quote($_) } @$releases) . ")";

  # restrict by users
  my @filter_members;

  if ($member) {
    push @filter_members, $member;
  }

  if (!is_empty_array($groups)) {
    my $emails = group_user_email($groups, $self->{config}->{CUMULATE});
    push @filter_members, @$emails;
  }

  if (@filter_members) {
    my $field
      = ($param->{user_type} =~ /^(assigned_to|qa_contact|reporter)$/)
      ? $param->{user_type}
      : 'assigned_to';

    $where
      .= " AND (bugs.$field IN (SELECT userid FROM profiles WHERE ("
      . join(' OR ', map { "login_name=" . $dbh->quote($_) } @filter_members)
      . ")))";
  }

  # customer issues

  if ($customer) {
    $where
      .= " AND (bugs.bug_id IN (SELECT ext_bz_bug_map.bug_id FROM ext_bz_bug_map WHERE ext_bz_bug_map.ext_bz_id=60))";
  }

  # partners

  if ($partner) {
    $from  .= ", bug_cf_partner";
    $where .= " AND bugs.bug_id=bug_cf_partner.bug_id";

    my $partners = $param->{partners};
    if (@$partners) {
      $where
        .= " AND ("
        . join(' OR ', map { "bug_cf_partner.value=" . $dbh->quote($_) } @$partners)
        . ")";
    }
  }

  $select .= ", COUNT(DISTINCT bugs.bug_id)";

  # BUG 1121484: Add bug visibility restrictions
  # Ideally, I'd like to pass this through RPC, but that would require
  # to much of a change
  my $user_id   = Bugzilla->user->id;
  my $group_ids = Bugzilla->user->groups_as_string();

  $from .= qq{
        LEFT JOIN bug_group_map AS security_map ON bugs.bug_id = security_map.bug_id
        LEFT JOIN cc AS security_cc
            ON bugs.bug_id = security_cc.bug_id AND security_cc.who = $user_id
    };
  $where .= qq{
        AND ( (security_map.group_id IS NULL OR security_map.group_id IN ($group_ids))
            OR (bugs.reporter_accessible = 1 AND bugs.reporter = $user_id)
            OR (bugs.cclist_accessible = 1 AND security_cc.who IS NOT NULL)
            OR bugs.assigned_to = $user_id
            OR bugs.qa_contact = $user_id
            OR bugs.docs_contact = $user_id
        )
    };

  return "$select $from $where $group_by $order_by";
}

################
# runmode subs
#################################################

sub extended_view {
  my $self = shift;
  my $vars = shift;

  my $page = 'extended-view';
  my $cgi  = Bugzilla->cgi();

  my @param_name  = ();
  my @param_value = ();
  push @param_name,  'page';
  push @param_value, $page;
  my $q        = $cgi;
  my $show_sql = untaint($q->param('show-sql'));
  my $combined = untaint($q->param('combine'));

  my $show_groups = [$q->param('group')];
  my $apply_group = untaint($q->param('apply_group'));

  if ($apply_group) {
    $show_groups = $self->orgchart_group_id($apply_group);
  }

  my $use_member
    = untaint($q->param('apply_member'))
    ? untaint($q->param('apply_member'))
    : untaint($q->param('member'));
  my $use_product = $q->param('product');

  if (defined $use_product) {
    push @param_name,  'product';
    push @param_value, $use_product;
  }
  my $use_version = untaint($q->param('version'));
  if (defined $use_version) {
    push @param_name,  'version';
    push @param_value, $use_version;
  }
  my $customer_mode
    = defined untaint($q->param('mode')) ? untaint($q->param('mode')) : 0;
  my @show_updates
    = $q->param('show')
    ? $q->param('show')
    : ($q->param('update') ? $q->param('update') : ());

  if (@show_updates) {

    # legacy support
    if (@show_updates == 1 && $show_updates[0] =~ /[,\s]/) {
      @show_updates = split(/[,\s]+/, $show_updates[0]);
    }
  }
  my $column_set = untaint($q->param('column-set'));
  if ($column_set) {
    push @param_name,  'column-set';
    push @param_value, $column_set;
  }
  my $hide_resolved_info = untaint($q->param('hide-resolved-info')) ? 1 : 0;
  if ($hide_resolved_info) {
    push @param_name,  'hide-resolved-info';
    push @param_value, $hide_resolved_info;
  }

  my $dbh                = Bugzilla->dbh();
  my @ACK_RESOLVED_TYPES = @{
    $dbh->selectcol_arrayref(
      "SELECT value FROM bug_status WHERE isactive=1 AND value != 'NEW' ORDER BY sortkey"
      )
      || []
  };

  my @ack_resolved
    = ($q->param('s-resolved-type')) ? $q->param('s-resolved-type') : ();
  my %sel_ack_resolved;
  my $resolved_type_url;

  #support pre-config value and backward compatible for use in url
  if (!@ack_resolved) {
    my $first_state
      = (defined untaint($q->param('resolved-type')))
      ? untaint($q->param('resolved-type'))
      : undef;

    $first_state = map_old_state($first_state)
      if (defined $first_state && $first_state =~ /^\d$/);

    if (!defined $first_state
      || (defined $first_state && !has_member($first_state, @ACK_RESOLVED_TYPES)))
    {
      $first_state = 'ON_QA';
    }

    @ack_resolved = get_later_types($first_state, @ACK_RESOLVED_TYPES);
  }

  map {
    $sel_ack_resolved{$_} = 1;
    $resolved_type_url .= "&s-resolved-type=$_";
  } @ack_resolved;

  my @hidden = ();
  for (my $i = 0; $i < @param_name; $i++) {
    my %param_data;
    $param_data{'name'}  = $param_name[$i];
    $param_data{'value'} = $param_value[$i];
    push(@hidden, \%param_data);
  }

  my $groups = groups_selection($show_groups);

  if ($use_member && !($use_member =~ /\@/)) {
    $use_member .= '@redhat.com';
  }

  my $show_inactive_releases = untaint($q->param('show_inactive'));

  if (!defined $show_inactive_releases || $show_inactive_releases eq '') {
## BUGBUG        if ( BugzillaMetrics::Config->instance()
##            ->get_value('USE_RELEASE_FLAGS') )
##        {
##            $show_inactive_releases = BugzillaMetrics::Config->instance()
##                ->get_value('RELEASE_FLAGS_SHOW_INACTIVE') || 0;
##
##        }
  }

  my @ack_releases = $self->get_ack_releases(
    {'show_inactive_releases' => $show_inactive_releases});

  my $show_all = (@show_updates == 1 && $show_updates[0] eq 'All') ? 1 : 0;

  my @releases_select
    = ({name => 'All', value => 'All', selected => ($show_all) ? 1 : 0});

  for (my $i = 0; $i < @ack_releases; $i++) {
    my %update_data;
    $update_data{'name'}  = $ack_releases[$i];
    $update_data{'value'} = $ack_releases[$i];
    foreach my $show (@show_updates) {
      if ($show eq $ack_releases[$i]) {
        $update_data{'selected'} = 1;
        last;
      }
    }
    push(@releases_select, \%update_data);
  }

  my $show_match = untaint($q->param('show_match')) || undef;

  if (!$show_all) {
    if (@show_updates) {
      @ack_releases = $self->get_ack_releases({
        'show_inactive_releases' => $show_inactive_releases,
        'show_releases'          => \@show_updates
      });
    }
    elsif ($show_match) {
      @ack_releases = @{
        get_releases({
          'show_inactive_releases' => $show_inactive_releases,
          'flag_name'              => $show_match
        })
      };
    }
  }

  my $use_columns = 'BLOCKER_COLUMNS';
  my $use_colors  = 'BLOCKER_COLORS';
  if ($column_set) {
    $use_columns = 'BLOCKER_' . $column_set . '_COLUMNS';
    $use_colors  = 'BLOCKER_' . $column_set . '_COLORS';
  }

  my @ack_columns = @{$self->{config}->{$use_columns}};
  my @ack_colors  = @{$self->{config}->{$use_colors}};
  my $bugs_scale  = $self->{config}->{'BLOCKER_GRAPH_SCALE'} || 100;

  $vars->{contact}       = $self->{config}->{CONTACT};
  $vars->{is_flag_view}  = 1;
  $vars->{mode}          = $customer_mode;
  $vars->{updates}       = \@releases_select;
  $vars->{show_match}    = $show_match;
  $vars->{show_inactive} = $show_inactive_releases;

  my $y_params = '&view-version=3';
  my $x_params = '';
  foreach my $group (@$show_groups) {
    my $clean_group = $group;
    $clean_group =~ s/&/%26/g;
    $x_params .= "&group=$clean_group";
  }

  # Get the value of user_type (assigned_to, qacontact, reporter)
  $vars->{user_types} = [qw(assigned_to qa_contact reporter)];
  my $user_type = $vars->{user_type} = untaint($q->param('user_type'))
    // 'assigned_to';
  $x_params .= "&user_type=$user_type";

  if ($use_member) {
    $x_params .= "&member=$use_member";
  }
  if ($resolved_type_url) {
    $y_params .= $resolved_type_url;
  }
  if (defined $combined) {
    $y_params .= "&combine=$combined";
  }

  my $ref = "page.cgi?id=releases/$page$x_params$y_params";
  if (@show_updates) {
    $ref .= '&' . join('&', map {"show=$_"} @show_updates);
  }
  $ref .= "&mode";

  if ($customer_mode) {
    $vars->{all_bugs_url}  = "$ref=0";
    $vars->{show_customer} = 1;
  }
  else {
    $vars->{show_all}          = 1;
    $vars->{customer_bugs_url} = "$ref=1";
  }

  # purge flag list of composite blockers
  my @flag_list = @{$self->{config}->{FLAG_LIST}};
  @flag_list = grep(!/^BLOCKER-/, @flag_list);

  # extract true flag names
  my @flags = ();
  foreach my $flag (@flag_list) {
    my $flag_specifier = lc($flag);
    push @flags, $flag_specifier;
  }

  # only consider values + and ? for each of the flags
  my %flag_values;
  foreach my $flag (@flags) {
    $flag_values{$flag} = '+?';
  }

  # require additional bugs field
  my @extras = ('target_milestone');

  my @IGNORE_STATES  = $self->{config}->{'IGNORE_STATES'};
  my @resolved_types = ();

  foreach my $type (@ACK_RESOLVED_TYPES) {
    my %param_data;
    $param_data{'name'} = $type;
    if (defined $sel_ack_resolved{$type}) {
      $param_data{'resolved_types_selected'} = 1;
    }
    else {
      $param_data{'resolved_types_selected'} = 0;
    }
    push(@resolved_types, \%param_data);
  }

  $vars->{hidden_params} = \@hidden;
  $vars->{groups}        = $groups;

  if ($use_member) {
    $vars->{member} = $use_member;
  }
  if (@resolved_types) {
    $vars->{resolved_types} = \@resolved_types;
  }

  # if no updates are selected, only show option header

  if ((!@show_updates && !$show_match) || !@ack_releases) {
    $vars->{hide_resolved_info} = 1;
    return;
  }

  my $sql_query = $self->generate_overview_sql({
    'flags'       => \@flags,
    'flag_values' => \%flag_values,
    'columns'     => \@ack_columns,
    'releases'    => \@ack_releases,
    'resolved'    => \@ack_resolved,
    'extras'      => \@extras,
    'groups'      => $show_groups,
    'member'      => $use_member,
    'user_type'   => $user_type,
    'customer'    => $customer_mode
  });

  if ($show_sql) {
    return $sql_query;
  }

  # get all bug states
  my $sth = $dbh->prepare("SELECT value FROM bug_status ORDER BY sortkey");
  $sth->execute();

  my $unresolved_list = '';
  my $resolved_list   = '';
  for (my $j = 0; $j < @ack_resolved; $j++) {
    if ($resolved_list) {
      $resolved_list .= ',';
    }
    $resolved_list .= $ack_resolved[$j];
  }

  while (my $row = $sth->fetchrow_arrayref()) {
    my $j = 0;
    while ($j < @ack_resolved) {
      if ($ack_resolved[$j] eq $$row[0]) {
        last;
      }
      $j++;
    }
    if ($j < @ack_resolved) {
      next;
    }
    $j = 0;
    while ($j < @IGNORE_STATES) {
      if ($IGNORE_STATES[$j] eq $$row[0]) {
        last;
      }
      $j++;
    }
    if ($j < @IGNORE_STATES) {
      next;
    }
    if ($unresolved_list) {
      $unresolved_list .= ',';
    }
    $unresolved_list .= $$row[0];
  }
  $sth->finish();

  # perform query
  $sth = $dbh->prepare($sql_query);
  $sth->execute();

  my $fixed_cols  = 4;
  my @milestones  = ("no-target");
  my %values      = ();
  my @result_list = ('MILESTONE', 'RELEASE', 'STATUS');
  push @result_list, @flag_list;

  while (my $row = $sth->fetchrow_arrayref()) {
    my $rown        = $fixed_cols + @flag_list + @ack_columns;
    my $milestone   = ($$row[2] && ($$row[2] ne "---")) ? $$row[2] : $milestones[0];
    my $release     = $$row[0];
    my $status      = $$row[1];
    my $is_extended = 0;
    for (my $flag_id = 3; $flag_id < @result_list; $flag_id++) {
      if ($$row[$flag_id]) {
        my $base_id = $release . '_' . $result_list[$flag_id];

        # do we have a blocker flag?
        if ($result_list[$flag_id] eq 'BLOCKER') {

          # if yes, we need to consider the milestone
          my $i = 0;
          while ($i < @milestones) {
            if ($milestone eq $milestones[$i]) {
              last;
            }
            $i++;
          }
          if ($i == @milestones) {

            # a new milestone to add to our list
            push @milestones, $milestone;
          }
          $base_id = $release . '_BLOCKER-' . $milestone;
        }
        $values{$base_id} = 1;
        my $count_this_bug = 1;
        if ($$row[$flag_id] eq '+') {
          $base_id .= '_' . 'APPROVED';
          $is_extended = 1;
        }
        elsif ($$row[$flag_id] eq '?') {
          $base_id .= '_' . 'PROPOSED';
          $is_extended = 1;
        }
        else {
          $base_id .= '_' . 'REJECTED';
          $count_this_bug-- if !$self->{config}->{ADD_DENIED_TO_TOTAL};
        }
        if ($combined) {
          if ($status eq '+') {
            $base_id .= '_' . 'APPROVED';
          }
          elsif ($status eq '?') {
            $base_id .= '_' . 'PROPOSED';
          }
          else {
            $base_id .= '_' . 'REJECTED';
            $count_this_bug-- if !$self->{config}->{ADD_DENIED_TO_TOTAL};
          }
        }
        elsif (!($status eq '+' || $status eq '?')) {

          # do not include bugs denied for a release
          next;
        }
        my $release_id = $release;
        if ($$row[$#result_list + 1]) {
          $base_id    .= '_in';
          $release_id .= '_in';
        }
        else {
          $base_id    .= '_out';
          $release_id .= '_out';
        }
        my $ri = $fixed_cols + @flag_list;
        foreach my $column (@ack_columns) {
          if ($$row[$ri++]) {
            my $id = $base_id . '_' . $column;
            if (defined $values{$id}) {
              $values{$id} += $$row[$rown];
            }
            else {
              $values{$id} = $$row[$rown];
            }
            my $raw_id = $base_id . '_raw';
            if (defined $values{$raw_id}) {
              $values{$raw_id} += $$row[$rown];
            }
            else {
              $values{$raw_id} = $$row[$rown];
            }
            if ($result_list[$flag_id] ne 'FAST') {
              my $ack_id = $id . '_ack';
              if (defined $values{$ack_id}) {
                $values{$ack_id} += $$row[$rown];
              }
              else {
                $values{$ack_id} = $$row[$rown];
              }
            }
            if ($count_this_bug == 1) {
              $id = $release_id . '_' . $column;
              if (defined $values{$id}) {
                $values{$id} += $$row[$rown];
              }
              else {
                $values{$id} = $$row[$rown];
              }
              $raw_id = $release_id . '_raw';
              if (defined $values{$raw_id}) {
                $values{$raw_id} += $$row[$rown];
              }
              else {
                $values{$raw_id} = $$row[$rown];
              }
              if ($result_list[$flag_id] ne 'FAST') {
                my $ack_id = $id . '_ack';
                if (defined $values{$ack_id}) {
                  $values{$ack_id} += $$row[$rown];
                }
                else {
                  $values{$ack_id} = $$row[$rown];
                }
              }
            }
          }
        }
        my $id = $base_id . '_all';
        if (defined $values{$id}) {
          $values{$id} += $$row[$rown];
        }
        else {
          $values{$id} = $$row[$rown];
        }
        if ($result_list[$flag_id] ne 'FAST') {
          my $ack_id = $id . '_ack';
          if (defined $values{$ack_id}) {
            $values{$ack_id} += $$row[$rown];
          }
          else {
            $values{$ack_id} = $$row[$rown];
          }
        }
      }
    }
    if (!$is_extended) {
      my $base_id = $release . '_' . 'ACK';
      $values{$base_id} = 1;
      my $count_this_bug = 1;
      if ($status eq '+') {
        $base_id .= '_' . 'APPROVED';
      }
      elsif ($status eq '?') {
        $base_id .= '_' . 'PROPOSED';
      }
      else {
        $base_id .= '_' . 'REJECTED';
        $count_this_bug-- if !$self->{config}->{ADD_NAK_TO_TOTAL};
      }
      if ($combined) {
        if ($status eq '+') {
          $base_id .= '_' . 'APPROVED';
        }
        elsif ($status eq '?') {
          $base_id .= '_' . 'PROPOSED';
        }
        else {
          $base_id .= '_' . 'REJECTED';
        }
      }
      my $release_id = $release;
      if ($$row[$#result_list + 1]) {
        $base_id    .= '_in';
        $release_id .= '_in';
      }
      else {
        $base_id    .= '_out';
        $release_id .= '_out';
      }
      my $ri = $fixed_cols + @flag_list;
      foreach my $column (@ack_columns) {
        if ($$row[$ri++]) {
          my $id = $base_id . '_' . $column;
          if (defined $values{$id}) {
            $values{$id} += $$row[$rown];
          }
          else {
            $values{$id} = $$row[$rown];
          }
          my $raw_id = $base_id . '_raw';
          if (defined $values{$raw_id}) {
            $values{$raw_id} += $$row[$rown];
          }
          else {
            $values{$raw_id} = $$row[$rown];
          }
          my $ack_id = $id . '_ack';
          if (defined $values{$ack_id}) {
            $values{$ack_id} += $$row[$rown];
          }
          else {
            $values{$ack_id} = $$row[$rown];
          }
          if ($count_this_bug == 1) {
            $id = $release_id . '_' . $column;
            if (defined $values{$id}) {
              $values{$id} += $$row[$rown];
            }
            else {
              $values{$id} = $$row[$rown];
            }
            $raw_id = $release_id . '_raw';
            if (defined $values{$raw_id}) {
              $values{$raw_id} += $$row[$rown];
            }
            else {
              $values{$raw_id} = $$row[$rown];
            }
            $ack_id = $id . '_ack';
            if (defined $values{$ack_id}) {
              $values{$ack_id} += $$row[$rown];
            }
            else {
              $values{$ack_id} = $$row[$rown];
            }
          }
        }
      }
      my $id = $base_id . '_all';
      if (defined $values{$id}) {
        $values{$id} += $$row[$rown];
      }
      else {
        $values{$id} = $$row[$rown];
      }
      my $ack_id = $id . '_ack';
      if (defined $values{$ack_id}) {
        $values{$ack_id} += $$row[$rown];
      }
      else {
        $values{$ack_id} = $$row[$rown];
      }
    }
  }

  $sth->finish();

  # sort milestones
  my @milestones_sort = grep(!/^no-target$/, @milestones);
  @milestones_sort = sort(@milestones_sort);
  @milestones      = ("no-target");
  push @milestones, @milestones_sort;

  my @rows         = ();
  my @section_list = ('ACK');
  my @type_list    = ('PROPOSED', 'APPROVED');
  if ($combined) {
    @type_list = (
      'PROPOSED_PROPOSED', 'PROPOSED_APPROVED',
      'APPROVED_PROPOSED', 'APPROVED_APPROVED'
    );
  }

  for (my $i = 0; $i < @flag_list; $i++) {
    my $section_name = $flag_list[$i];

    # do we have a blocker flag?
    if ($section_name eq 'BLOCKER') {

      # if yes, we need to consider the milestone
      for (my $i = 1; $i < @milestones; $i++) {
        push @section_list, 'BLOCKER-' . $milestones[$i];
      }
      push @section_list, 'BLOCKER-' . $milestones[0];
    }
    else {
      push @section_list, $section_name;
    }
  }

  for (my $i = 0; $i < @ack_releases; $i++) {
    my $release = $ack_releases[$i];
    my %release_data;
    $release_data{'release'}     = $release;
    $release_data{'show_totals'} = 1;
    my @lists = ();
    foreach my $list (@section_list) {
      if ($self->{config}->{HIDE_UNUSED} && !$values{($release . '_' . $list)}) {
        next;
      }
      my %list_data;
      my $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&mode=$customer_mode$x_params$y_params\">$list</a>";
      $list_data{'list'} = $link;
      my @types = ();
      foreach my $type (@type_list) {
        my $base_id = $release . '_' . $list . '_' . $type;
        my %type_data;
        my $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&mode=$customer_mode$x_params$y_params\">$type</a>";
        $type_data{'type'} = $link;
        my @column_values = ();
        my @quota_values  = ();
        my $all_in        = 0;
        my $j             = 0;
        my $all_in_id     = $base_id . '_in_all';
        my $all_out_id    = $base_id . '_out_all';
        my $raw_in_id     = $base_id . '_in_raw';
        my $raw_out_id    = $base_id . '_out_raw';
        my $all           = ($values{$all_in_id} ? $values{$all_in_id} : 0)
          + ($values{$all_out_id} ? $values{$all_out_id} : 0);
        my $raw = ($values{$raw_in_id} ? $values{$raw_in_id} : 0)
          + ($values{$raw_out_id} ? $values{$raw_out_id} : 0);
        my $multiplier    = $raw ? ($bugs_scale / $raw) : 0;
        my $all_out_quota = 0;

        foreach my $column (@ack_columns) {
          my %column_data;
          my %quota_data;
          my $id = $base_id . '_in_' . $column;
          my $in = $values{$id} ? $values{$id} : 0;
          $all_in += $in;
          my $link
            = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&resolved=1&column=$column&mode=$customer_mode$x_params$y_params\">$in</a>";
          if (!$hide_resolved_info) {
            $column_data{'in'} = $values{$id} ? $link : 0;
          }
          $id = $release . '_' . $list . '_in_' . $column;
          if (defined $values{$id}) {
            $values{$id} += $in;
          }
          else {
            $values{$id} = $in;
          }
          $id = $base_id . '_out_' . $column;
          my $out = $values{$id} ? $values{$id} : 0;
          $link
            = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&resolved=0&column=$column&mode=$customer_mode$x_params$y_params\">$out</a>";
          if (!$hide_resolved_info) {
            $column_data{'out'} = $values{$id} ? $link : 0;
          }
          $id = $release . '_' . $list . '_out_' . $column;
          if (defined $values{$id}) {
            $values{$id} += $out;
          }
          else {
            $values{$id} = $out;
          }
          $quota_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[$j];
          my $out_quota = int($out * $multiplier);
          if ($out && ($out_quota == 0)) { $out_quota++; }
          $quota_data{'value'} = $out_quota;
          $all_out_quota += $out_quota;
          my $value = $in + $out;
          $link
            = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&column=$column&mode=$customer_mode$x_params$y_params\">$value</a>";
          $column_data{'all'} = $value ? $link : 0;

          if ($hide_resolved_info) {
            $column_data{'hide_resolved_info'} = 1;
          }
          push(@column_values, \%column_data);
          push(@quota_values,  \%quota_data);
          $j++;
        }
        if (@ack_columns) {
          my %in_data;
          $in_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[$j];
          $in_data{'value'}
            = ($all_in && ($bugs_scale > $all_out_quota))
            ? ($bugs_scale - $all_out_quota)
            : 0;
          unshift(@quota_values, \%in_data);
        }
        else {
          my %out_data;
          $out_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[0];
          my $id        = $base_id . '_out_all';
          my $all_out   = $values{$id} ? $values{$id} : 0;
          my $out_quota = int($all_out * $multiplier);
          if ($all_out && ($out_quota == 0)) { $out_quota++; }
          $out_data{'value'} = $out_quota;
          push(@quota_values, \%out_data);
          my %in_data;
          $in_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[$#ack_colors];
          $id = $base_id . '_in_all';
          my $all_in = $values{$id} ? $values{$id} : 0;
          $in_data{'value'}
            = ($all_in && ($bugs_scale > $out_quota)) ? ($bugs_scale - $out_quota) : 0;
          unshift(@quota_values, \%in_data);
        }
        if (!$hide_resolved_info) {
          $type_data{'quota'} = \@quota_values;
        }
        $type_data{'columns'} = \@column_values;
        my $id = $base_id . '_in_all';
        my $in = $values{$id} ? $values{$id} : 0;
        $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&resolved=1&mode=$customer_mode$x_params$y_params\">$in</a>";
        if (!$hide_resolved_info) {
          $type_data{'in'} = $values{$id} ? $link : 0;
        }
        $id = $release . '_' . $list . '_in_all';
        if (defined $values{$id}) {
          $values{$id} += $in;
        }
        else {
          $values{$id} = $in;
        }
        $id = $base_id . '_out_all';
        my $out = $values{$id} ? $values{$id} : 0;
        $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&resolved=0&mode=$customer_mode$x_params$y_params\">$out</a>";
        if (!$hide_resolved_info) {
          $type_data{'out'} = $values{$id} ? $link : 0;
        }
        $id = $release . '_' . $list . '_out_all';
        if (defined $values{$id}) {
          $values{$id} += $out;
        }
        else {
          $values{$id} = $out;
        }
        my $value = $in + $out;
        $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&type=$type&mode=$customer_mode$x_params$y_params\">$value</a>";
        $type_data{'all'} = $value ? $link : 0;
        if ($hide_resolved_info) {
          $type_data{'hide_resolved_info'} = 1;
        }
        push(@types, \%type_data);
      }
      $list_data{'types'} = \@types;

      my @column_values = ();

      #            my @quota_values  = ();
      my $all_in        = 0;
      my $j             = 0;
      my $base_id       = $release . '_' . $list;
      my $all_in_id     = $base_id . '_in_all';
      my $all_out_id    = $base_id . '_out_all';
      my $all           = $values{$all_in_id} + $values{$all_out_id};
      my $multiplier    = $all ? ($bugs_scale / $all) : 0;
      my $all_out_quota = 0;
      foreach my $column (@ack_columns) {
        my %column_data;
        my %quota_data;
        my $id = $base_id . '_in_' . $column;
        my $in = $values{$id} ? $values{$id} : 0;
        $all_in += $in;
        my $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&resolved=1&column=$column&mode=$customer_mode$x_params$y_params\">$in</a>";
        if (!$hide_resolved_info) {
          $column_data{'in'} = $values{$id} ? $link : 0;
        }
        $id = $base_id . '_out_' . $column;
        my $out = $values{$id} ? $values{$id} : 0;
        $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&resolved=0&column=$column&mode=$customer_mode$x_params$y_params\">$out</a>";
        if (!$hide_resolved_info) {
          $column_data{'out'} = $values{$id} ? $link : 0;
        }
        $quota_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[$j];
        my $out_quota = int($out * $multiplier);
        if ($out && ($out_quota == 0)) { $out_quota++; }
        $quota_data{'value'} = $out_quota;
        $all_out_quota += $out_quota;
        my $value = $in + $out;
        $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&column=$column&mode=$customer_mode$x_params$y_params\">$value</a>";
        $column_data{'all'} = $value ? $link : 0;

        if ($hide_resolved_info) {
          $column_data{'hide_resolved_info'} = 1;
        }
        push(@column_values, \%column_data);

        #                push( @quota_values,  \%quota_data );
        $j++;
      }
      if (@ack_columns) {
        my %in_data;
        $in_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[$j];
        $in_data{'value'}
          = ($all_in && ($bugs_scale > $all_out_quota))
          ? ($bugs_scale - $all_out_quota)
          : 0;

        #                unshift( @quota_values, \%in_data );
      }
      else {
        my %out_data;
        $out_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[0];
        my $id        = $base_id . '_out_all';
        my $all_out   = $values{$id} ? $values{$id} : 0;
        my $out_quota = int($all_out * $multiplier);
        if ($all_out && ($out_quota == 0)) { $out_quota++; }
        $out_data{'value'} = $out_quota;

        #                push( @quota_values, \%out_data );
        my %in_data;
        $in_data{'image'} = $self->{config}->{IMAGE_PATH} . $ack_colors[$#ack_colors];
        $id = $base_id . '_in_all';
        my $all_in = $values{$id} ? $values{$id} : 0;
        $in_data{'value'}
          = ($all_in && ($bugs_scale > $out_quota)) ? ($bugs_scale - $out_quota) : 0;

        #                unshift( @quota_values, \%in_data );
      }
      if (!$hide_resolved_info) {

        #                $list_data{'quota'} = \@quota_values;
      }
      $list_data{'columns'} = \@column_values;
      my $id = $base_id . '_in_all';
      my $in = $values{$id} ? $values{$id} : 0;
      $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&resolved=1&mode=$customer_mode$x_params$y_params\">$in</a>";
      if (!$hide_resolved_info) {
        $list_data{'in'} = $values{$id} ? $link : 0;
      }
      $id = $base_id . '_out_all';
      my $out = $values{$id} ? $values{$id} : 0;
      $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&resolved=0&mode=$customer_mode$x_params$y_params\">$out</a>";
      if (!$hide_resolved_info) {
        $list_data{'out'} = $values{$id} ? $link : 0;
      }
      my $value = $in + $out;
      $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=$list&mode=$customer_mode$x_params$y_params\">$value</a>";
      $list_data{'all'} = $value ? $link : 0;

      if ($hide_resolved_info) {
        $list_data{'hide_resolved_info'} = 1;
      }

      push(@lists, \%list_data);
    }
    $release_data{'lists'} = \@lists;

    # add totals for the ack list, excluding fast

    my @column_totals = ();
    foreach my $column (@ack_columns) {
      my %column_data;
      my $id = $release . '_in_' . $column . '_ack';
      $column_data{'total_in'} = $values{$id} ? $values{$id} : 0;
      $id = $release . '_out_' . $column . '_ack';
      $column_data{'total_out'} = $values{$id} ? $values{$id} : 0;
      $column_data{'total_all'}
        = $column_data{'total_in'} + $column_data{'total_out'};

      if ($column_data{'total_in'} > 0) {
        my $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=TOTAL&resolved=1&mode=$customer_mode&column=$column$x_params$y_params\">"
          . $column_data{'total_in'} . "</a>";
        $column_data{'total_in'} = $link;
      }
      if ($column_data{'total_out'} > 0) {
        my $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=TOTAL&resolved=0&mode=$customer_mode&column=$column$x_params$y_params\">"
          . $column_data{'total_out'} . "</a>";
        $column_data{'total_out'} = $link;
      }
      if ($column_data{'total_all'} > 0) {
        my $link
          = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=TOTAL&mode=$customer_mode&column=$column$x_params$y_params\">"
          . $column_data{'total_all'} . "</a>";
        $column_data{'total_all'} = $link;
      }
      push(@column_totals, \%column_data);
    }
    $release_data{'column_totals'} = \@column_totals;

    # count ack list totals, which must be counted vertically

    $release_data{'total_in'}  = 0;
    $release_data{'total_out'} = 0;

    foreach my $list (@section_list) {
      next if $list eq 'FAST';    # do not include FAST bugs in ack total
      $release_data{'total_in'} += $values{$release . '_' . $list . '_in_all'}
        if $values{$release . '_' . $list . '_in_all'};
      $release_data{'total_out'} += $values{$release . '_' . $list . '_out_all'}
        if $values{$release . '_' . $list . '_out_all'};
    }

    $release_data{'total_all'}
      = $release_data{'total_in'} + $release_data{'total_out'};

    if ($release_data{'total_in'} > 0) {
      my $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=TOTAL&resolved=1&mode=$customer_mode$x_params$y_params\">"
        . $release_data{'total_in'} . "</a>";
      $release_data{'total_in'} = $link;
    }
    if ($release_data{'total_out'} > 0) {
      my $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=TOTAL&resolved=0&mode=$customer_mode$x_params$y_params\">"
        . $release_data{'total_out'} . "</a>";
      $release_data{'total_out'} = $link;
    }
    if ($release_data{'total_all'} > 0) {
      my $link
        = "<a href=\"page.cgi?id=releases/ack_list.html&update=$release&list=TOTAL&mode=$customer_mode$x_params$y_params\">"
        . $release_data{'total_all'} . "</a>";
      $release_data{'total_all'} = $link;
    }

    my @headers = ();
    my @blanks  = ();
    foreach my $column (@ack_columns) {
      my %header_data;
      my %blank_data;
      $header_data{'column'} = $column;
      if ($hide_resolved_info) {
        $header_data{'hide_resolved_info'} = 1;
        $blank_data{'hide_resolved_info'}  = 1;
      }
      push(@headers, \%header_data);
      push(@blanks,  \%blank_data);
    }
    $release_data{'columns'}       = \@headers;
    $release_data{'column_blanks'} = \@blanks;
    if ($i == $#ack_releases) {
      $release_data{'last'} = 1;
    }
    if ($hide_resolved_info) {
      $release_data{'hide_resolved_info'} = 1;
    }
    if (!@ack_columns) {
      $release_data{'all_only'} = 1;
    }

    push(@rows, \%release_data);
  }

  $vars->{rows} = \@rows;

  if (!$hide_resolved_info) {
    my @legend = ();
    if (@ack_columns) {
      my $j = 0;
      my %in_data;
      $in_data{'image'}
        = $self->{config}->{IMAGE_PATH} . $ack_colors[$#ack_columns + 1];
      $in_data{'column'} = 'In';
      push(@legend, \%in_data);
      foreach my $column (@ack_columns) {
        my %row_data;
        $row_data{'image'}  = $self->{config}->{IMAGE_PATH} . $ack_colors[$j];
        $row_data{'column'} = $column . ' Out';
        push(@legend, \%row_data);
        $j++;
      }
    }
    else {
      my %in_data;
      $in_data{'image'}  = $self->{config}->{IMAGE_PATH} . $ack_colors[$#ack_colors];
      $in_data{'column'} = 'Bugs In';
      push(@legend, \%in_data);
      my %out_data;
      $out_data{'image'}  = $self->{config}->{IMAGE_PATH} . $ack_colors[0];
      $out_data{'column'} = 'Bugs Out';
      push(@legend, \%out_data);
    }

    $vars->{legend} = \@legend;

    $vars->{resolved_list}   = $resolved_list;
    $vars->{unresolved_list} = $unresolved_list;
  }
  else {
    $vars->{hide_resolved_info} = 1;
  }

  return;
}

#################################################

1;    # Magic true value required at end of module
__END__

=head1 NAME

BugzillaMetrics::WebApp::ACKView - [One line description of module's purpose here]


=head1 VERSION

This document describes BugzillaMetrics::WebApp::ACKView version 0.0.1


=head1 SYNOPSIS

    use BugzillaMetrics::WebApp::ACKView;

=for author to fill in:
    Brief code example(s) here showing commonest usage(s).
    This section will be as far as many users bother reading
    so make it as educational and exeplary as possible.


=head1 DESCRIPTION

=for author to fill in:
    Write a full description of the module and its features here.
    Use subsections (=head2, =head3) as appropriate.


=head1 INTERFACE 

=for author to fill in:
    Write a separate section listing the public components of the modules
    interface. These normally consist of either subroutines that may be
    exported, or methods that may be called on objects belonging to the
    classes provided by the module.


=head1 DIAGNOSTICS

=for author to fill in:
    List every single error and warning message that the module can
    generate (even the ones that will "never happen"), with a full
    explanation of each problem, one or more likely causes, and any
    suggested remedies.

=over

=item C<< Error message here, perhaps with %s placeholders >>

[Description of error here]

=item C<< Another error message here >>

[Description of error here]

[Et cetera, et cetera]

=back


=head1 CONFIGURATION AND ENVIRONMENT

=for author to fill in:
    A full explanation of any configuration system(s) used by the
    module, including the names and locations of any configuration
    files, and the meaning of any environment variables or properties
    that can be set. These descriptions must also include details of any
    configuration language used.

BugzillaMetrics::WebApp::ACKView requires no configuration files or environment variables.


=head1 DEPENDENCIES

=for author to fill in:
    A list of all the other modules that this module relies upon,
    including any restrictions on versions, and an indication whether
    the module is part of the standard Perl distribution, part of the
    module's distribution, or must be installed separately. ]

None.


=head1 INCOMPATIBILITIES

=for author to fill in:
    A list of any modules that this module cannot be used in conjunction
    with. This may be due to name conflicts in the interface, or
    competition for system or program resources, or due to internal
    limitations of Perl (for example, many modules that use source code
    filters are mutually incompatible).

None reported.


=head1 BUGS AND LIMITATIONS

=for author to fill in:
    A list of known problems with the module, together with some
    indication Whether they are likely to be fixed in an upcoming
    release. Also a list of restrictions on the features the module
    does provide: data types that cannot be handled, performance issues
    and the circumstances in which they may arise, practical
    limitations on the size of data sets, special cases that are not
    (yet) handled, etc.

No bugs have been reported.

Please report any bugs or feature requests to
C<engineering-services@redhat.com>, or through the web interface at
L<https://engineering.redhat.com/rt3>.


=head1 AUTHOR

Bernd Groh  C<< <bgroh@redhat.com> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2007, Red Hat, Inc. All rights reserved.



=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
