# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;
use 5.10.1;

package Bugzilla::Extension::Releases::Release;

use base qw(Bugzilla::Object);

use Bugzilla::Constants;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::User;
use Bugzilla::FlagType;

###############################
####    Initialization     ####
###############################

use constant DB_TABLE => 'rh_release';

use constant DB_COLUMNS => qw(
  id
  type_id
  description
  sortkey
  is_active
);

use constant LIST_ORDER => 'sortkey';

use constant REQUIRED_CREATE_FIELDS => qw(
  type_id
  description
);

use constant UPDATE_COLUMNS => qw(
  type_id
  description
  sortkey
  is_active
);

use constant VALIDATORS => {
  type_id     => \&_check_type_id,
  description => \&_check_description,
  sortkey     => \&_check_sortkey,
  is_active   => \&_check_is_active,
};

use constant UPDATE_VALIDATORS => {
  type_id     => \&_check_type_id,
  description => \&_check_description,
  sortkey     => \&_check_sortkey,
  is_active   => \&_check_is_active,
};

###############################

sub new {
  my $class = shift;
  my $param = shift;

  my $flagtype;
  if (ref $param) {
    if ($param->{'release_id'}) {
      $flagtype = Bugzilla::FlagType->new($param->{'release_id'});
      $flagtype
        || ThrowCodeError("flag_type_nonexistent", {id => $param->{'release'}});
      Bugzilla->user->can_see_flag($flagtype)
        || ThrowUserError('flag_access_denied', {name => $flagtype->name});
    }
    elsif ($param->{'release'}) {
      $flagtype = Bugzilla::FlagType->new({name => $param->{'release'}});
      $flagtype
        || ThrowCodeError("flag_type_nonexistent", {id => $param->{'release'}});
      Bugzilla->user->can_see_flag($flagtype)
        || ThrowUserError('flag_access_denied', {name => $flagtype->name});
    }
    else {
      ThrowCodeError('bad_arg', {argument => 'release', function => "${class}::new"});
    }
    my $condition = 'type_id = ?';
    my @values    = ($flagtype->id);
    $param = {condition => $condition, values => \@values};
  }

  return $class->SUPER::new($param);
}

sub remove_from_db {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  $dbh->bz_start_transaction();
  $dbh->do('DELETE FROM rh_release WHERE id = ?', undef, $self->id);
  $dbh->bz_commit_transaction();

  return;
}

################################
# Validators
################################

sub _check_type_id {
  my ($invocant, $type_id) = @_;
  $type_id || ThrowCodeError('param_required', {param => 'type_id'});
  my $flagtype = Bugzilla::FlagType->new($type_id);
  $flagtype || ThrowCodeError('flag_type_nonexistent', {id => $type_id});
  return $type_id;
}

sub _check_description {
  my ($invocant, $description) = @_;
  $description = trim($description);
  $description || ThrowCodeError('param_required', {param => 'description'});
  return $description;
}

sub _check_sortkey {
  my ($invocant, $sortkey) = @_;
  my $skey = $sortkey;
  if (!defined $skey || $skey eq '') {
    ($sortkey)
      = Bugzilla->dbh->selectrow_array('SELECT MAX(sortkey) + 100 FROM rh_release')
      || 100;
  }
  detaint_natural($sortkey)
    || ThrowUserError('field_invalid_sortkey', {sortkey => $skey});
  return $sortkey;
}

sub _check_is_active { return $_[1] ? 1 : 0; }

###############################
####       Methods         ####
###############################

sub set_type_id     { $_[0]->set('type_id',     $_[1]); return; }
sub set_description { $_[0]->set('description', $_[1]); return; }
sub set_sortkey     { $_[0]->set('sortkey',     $_[1]); return; }
sub set_is_active   { $_[0]->set('is_active',   $_[1]); return; }

###############################
####      Accessors        ####
###############################

sub id          { return $_[0]->{'id'}; }
sub type_id     { return $_[0]->{'type_id'}; }
sub description { return $_[0]->{'description'}; }
sub sortkey     { return $_[0]->{'sortkey'}; }
sub is_active   { return $_[0]->{'is_active'}; }

sub flag_type {
  my ($self) = @_;
  return $self->{'flag_type'} if $self->{'flag_type'};
  $self->{'flag_type'} = Bugzilla::FlagType->new($self->type_id);
  return $self->{'flag_type'};
}

###############################
####      Subroutines      ####
###############################

1;

__END__

=head1 NAME

Release - Red Hat Bugzilla Release class.

=head1 SYNOPSIS

    use Bugzilla::Extension::Releases::Release;

    my $release = new Bugzilla::Extension::Releases::Release($release_id);
    my $release = new Bugzilla::Extension::Releases::Release({ release => $release });

    my $release = Bugzilla::Extension::Releases::Release->check({ release => $release });

    my $release =
      Bugzilla::Extension::Releases::Release->create({ type_id     => $flagtype->id,
                                                  description => $description });

    $release->set_description($new_description);
    $release->set_type_id($flagtype->id);
    $release->update();

    $release->remove_from_db;

=head1 DESCRIPTION

Release.pm represents a Red Hat Bugzilla Release object.

=head1 METHODS

=over

=item C<new($param)>

 Description: The constructor is used to load an existing Release 
              by passing a release ID or a hash with the release
              name.

 Params:      $param - If you pass an integer, the integer is the
                       release ID from the database that we want to
                       read in. If you pass in a hash with the 'release'
                       key, then the value of the release
                       key is the release name.

 Returns:     A Release object.

=item C<set_type_id($new_type_id)>

 Description: Changes the flag type id of the release.

 Params:      $new_type_id - new type_id of the release (integer).

 Returns:     Nothing.

=item C<set_description($new_desc)>

 Description: Changes the description of the release.

 Params:      $new_desc - new description of the release(string).

 Returns:     Nothing.

=item C<update()>

 Description: Write changes made to the release into the DB.

 Params:      none.

 Returns:     A hashref with changes made to the release object.

=item C<remove_from_db()>

 Description: Deletes the current release from the DB. The object itself
              is not destroyed.

 Params:      none.

 Returns:     Nothing.

=back

=head1 CLASS METHODS

=over

=item C<create(\%params)>

 Description: Create a new release.

 Params:      The hashref must have the following keys:
              type_id         - flag type id for the new release (integer). 
              description     - description of the new release (string).

 Returns:     A Release object.

=back

=cut
