# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::Job;

use strict;
use warnings;
use 5.10.1;

BEGIN {
  ## no critic (ProhibitStringyEval)
  # safe and avoids having to conditionally import if TheSchwarts is missing
  eval "use base qw(TheSchwartz::Worker Bugzilla::Extension)";
  ## use critic
}
use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Extension::RuleEngine::Rule;
use Bugzilla::Extension::RuleEngine::RuleState;
use Bugzilla::Mailer;
use Bugzilla::Util qw(datetime_from diff_arrays format_time);
use DateTime::Format::DBI;
use Bugzilla::Field;

use base qw(Exporter);
@Bugzilla::Extension::RuleEngine::Job::EXPORT
  = qw(_custom_search_match _get_flag_change_date _get_flag_group_match);

use Clone qw(clone);
use Data::Dumper;
use DateTime;
use List::Util qw(sum);
use List::MoreUtils qw(indexes last_index uniq any none);
use Try::Tiny;

# The longest we expect a job to possibly take, in seconds.
use constant grab_for => 300;

# We don't want the Rules Engine to fail permanently very easily. Retry for
# 6 hours.
use constant max_retries => 36;

# If it fails, retry after 10 mintes
use constant retry_delay => 600;

use constant BUG_FIELD_MAP => {
  assigned_to               => {key  => 'login'},
  assigned_to_realname      => {func => 'assigned_to', key => 'name'},
  'attach_data.thedata'     => {func => 'attachments', key => 'data'},
  'attachments.description' => {func => 'attachments', key => 'description'},
  'attachments.filename'    => {func => 'attachments', key => 'filename'},
  'attachments.isobsolete'  => {func => 'attachments', key => 'isobsolete'},
  'attachments.ispatch'     => {func => 'attachments', key => 'ispatch'},
  'attachments.isprivate'   => {func => 'attachments', key => 'isprivate'},
  'attachments.mimetype'    => {func => 'attachments', key => 'contenttype'},
  'attachments.submitter'   => {func => 'attachments', key => 'attacher/login'},
  bug_id                    => {type => 'number'},
  bug_group                 => {func => 'groups_in', key => 'name'},
  cf_clone_of               => {type => 'number'},
  cf_qe_conditional_nak     => {type => 'multi'},
  cf_conditional_nak        => {type => 'multi'},
  cf_verified               => {type => 'multi'},
  cf_partner                => {type => 'multi'},
  cf_last_closed            => {type => 'date'},
  cf_pm_score               => {type => 'number'},
  commenter                 => {func => 'comments', key => 'author/login'},
  creation_ts               => {type => 'date'},
  component                 => {func => 'all_components'},
  days_elapsed              => {type => 'date', func => 'delta_ts'},
  deadline                  => {type => 'date'},
  delta_ts                  => {type => 'date'},
  docs_contact              => {key  => 'login'},
  docs_contact_realname     => {func => 'docs_contact', key => 'name'},
  estimated_time            => {type => 'number'},
  'ext_bz_bug_map.ext_bz_bug_id' =>
    {func => 'external_bugs', key => 'ext_bz_bug_id'},
  'external_bugzilla.description' =>
    {func => 'external_bugs', key => 'type/description'},
  'external_bugzilla.url' => {func => 'external_bugs', key => 'type/url'},
  'flagtypes.name'        => {func => 'flags'},
  keywords                => {func => 'keyword_objects', key => 'name'},
  lastdiffed              => {type => 'date'},
  longdesc                => {func => 'comments', key => 'body'},
  'longdescs.count'       => {func => 'comments'},
  'longdescs.isprivate'   => {func => 'comments', key => 'is_private'},
  percentage_complete     => {type => 'number'},
  qa_contact              => {key  => 'login'},
  qa_contact_realname     => {func => 'qa_contact', key => 'name'},
  remaining_time          => {type => 'number'},
  reporter                => {key  => 'login'},
  reporter_realname       => {func => 'docs_contact', key => 'name'},
  'requestees.login_name' => {func => 'flags', key => 'requestee/login'},
  see_also                => {func => 'see_also', key => 'name'},
  'setters.login_name'    => {func => 'flags', key => 'setter/login'},
  version                 => {func => 'all_versions'},
  work_time               => {func => 'comments', type => 'number'},
  ## AgileTools bug fields
  'agile_pool.name'           => {func => 'pool',       key  => 'name'},
  'bug_agile_pool.pool_id'    => {func => 'pool_id',    type => 'number'},
  'bug_agile_pool.pool_order' => {func => 'pool_order', type => 'number'},
};

sub work {
  my ($class, $job) = @_;
  my $action      = $job->arg->{action};
  my $bug_ids     = $job->arg->{bug_ids};
  my $delta_ts    = $job->arg->{delta_ts};
  my $do_periodic = $job->arg->{do_periodic} || 0;
  my $rule_id     = $job->arg->{rule_id};
  my $who         = $job->arg->{who};

  # Be system user
  Bugzilla->set_user(Bugzilla::User->new(
    {name => Bugzilla->params->{system_user}}));

  # With no arg this reports global kill switches
  if (Bugzilla::Extension::RuleEngine::RuleState->disabled_count) {

    # Declining doesn't affect the max_retries value, so we can decline
    # for an indefinite amount of time.
    bre_debug(
      "Declined job " . $job->jobid . " because the global kill switch is active");
    $job->declined();
    return;
  }

  ## REDHAT EXTENSION START 1128496
  # We need to check if there are earlier (lower ts_job.id) rules with
  # the same bug id, and if there is, decline this job
  my @other_jobs
    = Bugzilla->job_queue->list_jobs({
    funcname => 'Bugzilla::Extension::RuleEngine::Job', limit => MAX_INT_32,
    });

  foreach my $other_job (@other_jobs) {
    next if $other_job->jobid >= $job->jobid;
    my $other_bug_ids = $other_job->arg->{bug_ids};
    next if not defined $other_bug_ids;
    foreach my $other_bug_id (@$other_bug_ids) {
      if (any { $_ == $other_bug_id } @$bug_ids) {
        bre_debug("Declined job "
            . $job->jobid
            . " because job "
            . $other_job->jobid
            . " must run first");
        $job->declined();
        return;
      }
    }
  }

  ## REDHAT EXTENSION END 1128496

  my $args = {
    action      => $action,
    bug_ids     => $bug_ids,
    delta_ts    => $delta_ts,
    do_periodic => $do_periodic,
    rule_id     => $rule_id,
    who         => $who,
  };

  my $success = eval { $class->do_job($args); };

  if (!$success) {
    unless ($@ =~ /Mismatched bug count. Got \d+, expected \d+/
      || $@ =~ /Bugs?[\d\s]+(is|are) earlier than delta_ts/
      || $@ =~ /database reported a query serialization error/
      || $@ =~ /could not serialize access due to concurrent update/
      || $@ =~ /Deadlock found when trying to get lock/
      || $@ =~ /deadlock detected/
      || $@ =~ /Declined/)
    {
      my $err_str = $@;

      # If we don't get one of the above errors, chances are there is a
      # problem with the code. In these cases, we need to trip the kill
      # switch.
      my $rule;
      $rule = Bugzilla::Extension::RuleEngine::Rule->new($rule_id)
        if ($rule_id && $rule_id > 0);

      kill_me($rule, $err_str, 1);
    }

    if ($@ =~ /Declined/) {

      # Declining doesn't affect the max_retries value, so we can decline
      # for an indefinite amount of time.
      bre_debug($@);
      $job->declined();
    }
    else {
      $job->failed($@);
    }
    undef $@;
  }
  else {
    $job->completed;
  }

  # We know that Bugzilla->dbh exists (it is created when we check the Rule
  # Engine is enabled), but it doesn't hurt to check. In any case, calling
  # Bugzilla->dbh would create it :)
  if (Bugzilla->dbh) {
    Bugzilla->dbh->disconnect;
  }

  return;
}

sub do_job {
  my $self = shift;
  my $args = shift;
  my $dbh  = Bugzilla->dbh;

  my $action      = $args->{action};            # create, updated or dryrun
  my $bug_ids     = $args->{bug_ids} || [];     # arrayref of bugs to change
  my $delta_ts    = $args->{delta_ts} || '';    # set for create and update
  my $do_periodic = $args->{do_periodic};       # do we run periodic rules
  my $rule_id     = $args->{rule_id} || -1;     # for dryrun, rule to run
  my $who         = $args->{who} || '';         # for dryrun, user id to report to

  my @result = ();                              # Used for dry run

  bre_debug("DO_JOB STARTING\n" . Dumper($args));

  # Get a list of bugs
  my $bugs = Bugzilla::Bug->new_from_list($bug_ids);
  if ($#$bugs != $#$bug_ids) {

    # This can happen if a newly created bug doesn't exist yet
    my $bug_count = scalar(@$bugs);
    my $id_count  = scalar(@$bug_ids);
    die "Mismatched bug count. Got $bug_count, expected $id_count\n";
  }

  # If delta_ts is set, check that
  if ($delta_ts) {
    my @premature = grep { $_->delta_ts lt $delta_ts } @$bugs;
    if (scalar(@premature)) {
      if (scalar(@premature) == 1) {
        die "Bug " . $premature[0]->id . " is earlier than delta_ts ($delta_ts)";
      }
      else {
        die "Bugs "
          . join(', ', map { $_->id } @premature)
          . " are earlier than delta_ts ($delta_ts)";
      }
    }
  }

  my %changes = ();

  # Lets do the work
  foreach my $bug (@$bugs) {
    unless ($bug->product_obj->rule_group) {
      bre_debug("Skipping bug because product has no BRE group: " . $bug->id);
      next;
    }

    if ($bug->product_obj->rule_group_obj->disabled_count) {
      die(  "Declined job because the kill switch is active for Rule Group '"
          . $bug->product_obj->rule_group_obj->name
          . "'");
    }

    my $uname = $bug->product_obj->rule_group_obj->user_name;

    unless ($uname && $uname ne '') {
      my $msg
        = "The BRE user is not set for the "
        . $bug->product_obj->rule_group_obj->name
        . " Rule Group.";
      bre_debug($msg);
      my $rules = Bugzilla::Extension::RuleEngine::Rule->match_rules({
        inactive      => 1,
        isactive      => 1,
        rule_group_id => $bug->product_obj->rule_group,
      });

      # add a kill switch if there are any active rules
      if (@$rules) {
        kill_me($rules->[0], $msg);
      }

      next;
    }

    my $user = Bugzilla::User->new({name => $uname});

    unless ($user && $user->can_see_bug($bug)) {
      bre_debug(
        "Skipping because the BRE user for this Group cannot see bug: " . $bug->id);
      next;
    }

    # switch to BRE user for this group
    Bugzilla->set_user($user);

    # Get the list of rules for thos product's BRE Group
    my %args = (
      rule_group_id => $bug->product_obj->rule_group,
      inactive      => 1,
      isactive      => 1,
    );

    if ($do_periodic) {
      $args{do_periodic} = 1;
    }
    else {
      $args{is_bug_change} = 1;
    }

    my @rules = @{Bugzilla::Extension::RuleEngine::Rule->match_rules(\%args)};

    unless (@rules) {
      bre_debug("No rules found for bug: " . $bug->id . "\n" . Dumper(\%args));
      next;
    }

    $dbh->bz_start_transaction;
    my ($changes, $notifies, $minor_update) = $self->do_job_on_bug($bug, \@rules);

    if ($action eq 'dryrun') {

      # We ignore the notifies for dry runs
      $changes{$bug->id} = $changes;
      $dbh->bz_rollback_transaction;
    }
    elsif (scalar(@$changes) == 0) {
      bre_debug("No changes!");

      # Nothing changed, move along
      $dbh->bz_rollback_transaction;
    }
    else {
      if (scalar(@$notifies)) {
        $self->_send_notifies($notifies);
      }
      $dbh->bz_commit_transaction;

      # Send the bug mail. Ideally we'd have all the changes
      # in the param, but that just isn't worth it.
      $bug->send_changes($changes->[-1]{changes}, undef, $minor_update);
    }
  }

  # Change back to super user for remaining work
  Bugzilla->set_user(Bugzilla::User->new(
    {name => Bugzilla->params->{system_user}}));

  if ($action eq 'dryrun') {

    # Send out change notification
    $self->_send_dryrun_feedback({
      rule_id => $rule_id, who => $who, changes => \%changes,
    });
  }

  # No error, the job will be marked as completed
  return 1;
}

sub do_job_on_bug {
  my $self  = shift;
  my $bug   = shift;
  my $rules = shift;

  my $minor_update = 1;

  bre_debug("Processing bug " . $bug->id);

  # We skip rules once they've been run once. ($skip_rules{id} = 1)
  # This prevents rules cancelling each other out. More importantly, it
  # prevents badly written rules from running multiple times for the
  # same update.
  my %skip_rules = ();

  # Notifies are done later, so record them now
  # @notifies = [ { rule_name => 'x', bug_id => x, email => 'x' }, ... ]
  my @notifies = ();

  # The rules that were excuted [ { rule_name => 'x', changes => array } ]
  my @rule_changes = ();

  # Did a rule get run this round? Set to one to ensure it runs at least once
  my $rule_change = 1;

  while ($rule_change) {
    $rule_change = 0;
    foreach my $rule (@$rules) {
      bre_debug("Processing rule " . $rule->name);

      if ($skip_rules{$rule->id}) {
        bre_debug("Skipping");
        next;
      }

      # Does the rule match? Skip if it doesn't
      next unless $self->_rule_is_match($bug, $rule);

      # Make the changes.
      my ($changes, $extra_changes) = $self->_rule_do_action($bug, $rule, \@notifies);

      # Record the changes
      push @rule_changes,
        {
        rule_name     => $rule->name,
        changes       => $changes,
        extra_changes => $extra_changes,
        };

      # We don't want this rule to run again for this change
      $skip_rules{$rule->id} = 1;

      # We reload the bug object, to get the updated values
      $bug = $bug->new($bug->id);

      # Force the rules to run again
      $rule_change = 1;

      if ($minor_update && !$rule->minor_update()) {
        $minor_update = 0;
      }
    }
  }

  return (\@rule_changes, \@notifies, $minor_update);
}

sub _rule_is_match {
  my $self    = shift;
  my $bug     = shift;
  my $rule    = shift;
  my $matches = shift;
  my $dbh     = Bugzilla->dbh;

  $matches = $rule->current->match if ($rule);
MATCH: foreach my $match (keys %$matches) {
    bre_debug("Checking for match on '$match' on bug " . $bug->id);
    my $match_values = $matches->{$match};
    if ($match =~ /_word$/) {

      # We'll match _word field on the word
      next;
    }
    elsif ($match eq 'flag_types') {
      my $flags = $bug->flags;
    FLAG: foreach my $row (@$match_values) {
        my $name   = $row->{name};
        my $status = $row->{status};
        bre_debug("Checking flag '$name' '" . join('', @$status) . "'");

        if (any { $_ eq 'X' } @$status) {

          # For unset flags, we need to check the flag doesn't exist
          unless (any { $_->name eq $name } @$flags) {
            next FLAG;
          }
        }

        foreach my $flag (@$flags) {
          if ($flag->name eq $name && (any { $_ eq $flag->status } @$status)) {

            # Match!
            next FLAG;
          }
        }

        # No match
        bre_debug('no match');
        return 0;
      }
    }
    elsif ($match eq 'custom') {
      foreach my $row (@$match_values) {
        my $negate = $row->{negate};
        my $field  = $row->{field};
        my $op     = $row->{op};
        my $value  = $row->{value};
        bre_debug("Checking custom field '"
            . ($field || "no field") . "' '"
            . ($op    || "no op") . "' '"
            . ($value || "no value")
            . "'");
        my $bug_values = $self->_get_bug_value($bug, $field);
        bre_debug("Value from bug is " . Dumper($bug_values));

        ## REDHAT EXTENSION BEGIN 1081769
        # Modified from Bugzilla::Search->_flagtypes_nonchanged
        if ($field eq 'flagtypes.name' && $op =~ /^not(.+)/) {

          # For 'not' flag operators, we need to negate the whole term.
          # If you search for "Flags" (does not contain) "approval+" we actually want
          # to return *bugs* that don't contain an approval+ flag.  Without rewriting
          # the negation we'll search for *flags* which don't contain approval+.
          $op     = $1;
          $negate = !$negate;
          bre_debug("Flag change now: '$negate', '$field' '$op' '$value'");
        }

        ## REDHAT EXTENSION END 1081769
        elsif ($field eq 'days_elapsed') {

          # in $value we have the 'days since last bug change' arg supplied
          # by the user as an integer.  We need to convert it into a DateTime.

          my $dateval = DateTime->now;
          $dateval->add(days => $value);
          $value = $dateval;
        }
        elsif ($field eq 'creation_ts' && (!defined $value || $value eq '')) {
          kill_me($rule,
            "In rule '" . $rule->name . "', $field is empty on bug " . $bug->id)
            if ($rule);
        }

        my $found;
        my $type = 'text';
        if (defined BUG_FIELD_MAP->{$field} && defined BUG_FIELD_MAP->{$field}{type}) {
          $type = BUG_FIELD_MAP->{$field}{type};
        }
        else {
          my $field_object = Bugzilla::Field->new({name => $field});
          if ( $field_object->type == Bugzilla::Constants::FIELD_TYPE_DATE
            || $field_object->type == Bugzilla::Constants::FIELD_TYPE_DATETIME)
          {
            $type = 'date';
          }
        }
        eval {
          $found = _custom_search_match({
            bug    => $bug,
            op     => $op,
            value  => $value,
            type   => $type,
            values => $bug_values,
          });
        };
        if ($@) {
          bre_debug("Rule "
              . ($rule ? $rule->name : "New rule")
              . " errored when matching '$field' against "
              . Dumper($value, $bug_values)
              . $@);
          kill_me($rule, "$@ in rule '" . $rule->name . "', field: '$field'") if ($rule);
        }

        # So now we know if we have a match or not
        bre_debug("negate: $negate, found: $found");
        if (($negate && $found) || (!$negate && !$found)) {
          bre_debug('no match');
          return 0;
        }
      }
    }
    elsif ($match eq 'bug_status'
      && scalar(@{$match_values->{values}}) == 1
      && $match_values->{values}[0] eq 'All open')
    {
      bre_debug('Checking special status');

      # Special case for the open bug status
      unless ($bug->status->is_open) {
        bre_debug('no match');
        return 0;
      }
    }
    elsif ($match eq 'component_list') {
      my $bug_components = $bug->all_components;
      my $type           = $match_values->{'word'};
      my $value          = $match_values->{values}[0];

      # Get a list of all the components for this release flag
      my $query = q{
                SELECT c.name
                FROM components c
                    JOIN rh_release_components rc ON c.id = rc.component_id
                    JOIN rh_release r ON rc.release_id = r.id
                    JOIN flagtypes f ON r.type_id = f.id
                WHERE f.name = ?
            };

      my $component_list;
      if ($type eq 'approved' or $type eq 'capacity') {
        $query .= " AND rc.component_type = ?";
        $component_list = $dbh->selectcol_arrayref($query, undef, $value, $type);
      }
      else {
        $component_list = $dbh->selectcol_arrayref($query, undef, $value);
      }

      bre_debug("Type: $type, Bug components / Release components"
          . Dumper($bug_components, $component_list));

      # See if there is a match.
      my $match = 0;
      foreach my $component (@$bug_components) {
        if ($type eq 'neither') {

          # For neither, we check for a negative match
          unless (any { $_ eq $component } @$component_list) {
            $match = 1;
          }
        }
        else {
          if (any { $_ eq $component } @$component_list) {
            $match = 1;
          }
        }
      }

      unless ($match) {
        bre_debug('no match');
        return 0;
      }
    }
    elsif ($match eq 'flags_order') {
      my $before = $match_values->{before};
      my $after  = $match_values->{after};

      my $before_date = _get_flag_change_date($bug, $before);
      my $after_date  = _get_flag_change_date($bug, $after);

      # If $before_date is after $after_date, then there is no match
      if ( !defined $before_date
        || !defined $after_date
        || $before_date gt $after_date)
      {
        bre_debug('no match');
        return 0;
      }
    }
    elsif ($match eq 'flag_groups') {
      foreach my $details (@$match_values) {
        if (!_get_flag_group_match($bug, $details)) {
          bre_debug('no match');
          return 0;
        }
      }
    }
    else {
      # Get the values from bugs, from the match, and the type
      my $bug_values = $self->_get_bug_value($bug, $match);
      bre_debug("Value from bug is " . Dumper($bug_values));
      my $values = $match_values->{values};
      my $type   = $match_values->{'word'} || 'is any of';
      bre_debug("Type from rule is '$type', value is " . Dumper($values));

      # Count the number of times they match
      my $found = 0;

      ## REDHAT EXTENSION START 1313572
      # Hack, can't figure out where the bogus fields are being added, see 1360606
      if (not defined $values) {
        bre_debug("value is not defined for $match");
        $found++;
      }

      ## REDHAT EXTENSION END 1313572
      foreach my $value (@$values) {
        if (scalar(@$bug_values) == 1 && !defined($bug_values->[0])) {
          $found++ if $value eq '---';
        }
        else {
          $found++ if any { $_ eq $value } @$bug_values;
        }
      }

      # If we didn't get the expected number, there is no match
      bre_debug("Found is '$found'");
      if ( ($type eq 'is any of' and $found == 0)
        || ($type eq 'is all of' and $found != scalar(@$values))
        || ($type eq 'is not any of' and $found > 0))
      {
        bre_debug('no match');
        return 0;
      }

    }
  }

  # If we get here, the rule matches
  bre_debug("This bug matches!");
  return 1;
}

sub _get_flag_change_date {
  my ($bug, $flag) = @_;

  # We need to get the latest date that the flag was changed
  # First we check in the bugs_activity log
  my $query = q{
        SELECT MAX(a.bug_when)
        FROM bugs_activity a JOIN fielddefs f ON (a.fieldid = f.id)
        WHERE a.bug_id = ?
        AND a.added = ?
        AND f.name = 'flagtypes.name'
    };

  my $change_date
    = Bugzilla->dbh->selectrow_array($query, undef, $bug->id, $flag);
  return $change_date if $change_date;

  # So the flag _may_ have been set at bug_creation time, if it is
  # still currently set.
  my ($name, $status) = ($flag =~ /^(.+)(.)$/);
  if (
    any { $_->type->name eq $name && $_->status eq $status }
    @{$bug->flags}
    )
  {
    return $bug->creation_ts;
  }

  # It could also be set at creation time if it had been removed at one point
  $query = q{
        SELECT 1
        FROM bugs_activity a JOIN fielddefs f ON (a.fieldid = f.id)
        WHERE a.bug_id = ?
        AND a.removed = ?
        AND f.name = 'flagtypes.name'
    };

  my $removed = Bugzilla->dbh->selectrow_array($query, undef, $bug->id, $flag);

  # If removed is true, assume it was created a bug creation time
  # Otherwise, we assume it has never been set, and return undef
  return $removed ? $bug->creation_ts : undef;
}

sub _get_flag_group_match {
  my ($bug, $match_details) = @_;
  my $flag_group
    = Bugzilla::Extension::RuleEngine::FlagGroup->new($match_details->{id});
  my $match_regexp     = $flag_group->current->theregexp;
  my $match_flag_types = $flag_group->current->flag_types;
  my $word             = $match_details->{word};             # 'any' 'all'
  my $status           = $match_details->{status};           # Arrayref of X ? + -

  my $flag_types  = $bug->flag_types; # Get the flag types for this bug
  my $found_match = 0;                # Set when at least one flag type is matched

  foreach my $flag_type (@$flag_types) {

    # See if this flag type is a match (either regexp or flag type list)
    next
      unless ($match_regexp && $flag_type->name =~ qr{$match_regexp})
      || any { $_->name eq $flag_type->name } @$match_flag_types;

    $found_match = 1;

    # For unset flags, we need to check the flag doesn't exist for this type
    if (any { $_ eq 'X' } @$status) {
      if (scalar(@{$flag_type->{flags}}) == 0) {
        return 1 if $word eq 'any';
        next;
      }
    }

    # Otherwise, we get to see if any of the flags of this flag type
    # match one of the status
    foreach my $flag (@{$flag_type->{flags}}) {
      if (any { $_ eq $flag->status } @$status) {

        # Match!
        return 1 if $word eq 'any';
        next;
      }
    }

    # No match for all type search
    return 0 if $word eq 'all';
  }

  # If we got here with an any search, we didn't find any matching flags
  return 0 if $word eq 'any';

  # If we got here with an all search, we return 0 if we found no matching
  # flag types, or 1 otherwise.
  return $found_match;
}

sub _expand_field_wildcards {
  my ($bug, $value, $type) = @_;

  if ($value && $value =~ m/%%([^%]+)%%/) {
    my $val = $bug->$1 // '';
    $value =~ s/%%([^%]+)%%/$val/eg;
  }

  bre_debug("\n\nMATCHES: $value, type: $type\n");
  if ($type eq 'date' && $value =~ /^([\d\-]+).*?([+-]?)\s*(\d+)([hdwmy])/i) {

    # 2020-01-30 -1m
    my $date    = $1;
    my $sign    = $2 // '+';
    my $period  = $3;
    my $step    = lc($4);
    my %wordmap = (qw(h hours d days w weeks m months y years));
    die("Unable to parse '$value' as a relative date (unknown character '$step')")
      unless exists($wordmap{$step});
    bre_debug("Matched: " . Dumper($date, $sign, $period, $step));
    eval {
      $value = datetime_from($date)->add($wordmap{$step} => "$sign$period")->ymd();
    };
    die "Unable to parse date '$value'" if ($@);
  }
  elsif ($type eq 'date' && $value =~ /^([+-]?)\s*(\d+)([hdwmy])/i) {

    # +1m
    my $sign    = $1 // '+';
    my $period  = $2;
    my $step    = lc($3);
    my %wordmap = (qw(h hours d days w weeks m months y years));
    die("Unable to parse '$value' as a relative date (unknown character '$step')")
      unless exists($wordmap{$step});
    bre_debug("Matched: " . Dumper($sign, $period, $step));
    eval {
      eval { $value = DateTime->now->add($wordmap{$step} => "$sign$period")->ymd(); };
    };
    die "Unable to parse date '$value'" if ($@);
  }

  foreach
    my $field ('assigned_to', 'cc', 'qa_contact', 'docs_contact', 'reporter')
  {
    if ($value && $value =~ m/\%$field\%/) {
      my $user = $bug->$field;
      if ($user && ref($user) eq 'ARRAY') {
        $user = join(', ', @$user);
      }

      $value =~ s/\%${field}\%/$user/g;
    }
  }

  foreach my $role_obj (Bugzilla::Extension::AgileTools::Role->get_all) {
    my $role = $role_obj->name;

    # Add Agile pool roles
    if ($value && $value =~ m/\%pool\.$role\%/ && $bug->pool) {
      my $team = $bug->pool->team;
      if (defined($team->by_roles->{$role})) {
        my $str = join(', ', map { $_->login } @{$team->by_roles->{$role}});
        $value =~ s/\%pool\.${role}\%/$str/g;
      }
    }

    # Add Agile team roles
    if ($value && $value =~ m/\%team\.$role\%/) {
      my $team = $bug->my_team();

      if ($team && defined($team->by_roles->{$role})) {
        my $str = join(', ', map { $_->login } @{$team->by_roles->{$role}});
        $value =~ s/\%team\.${role}\%/$str/g;
      }
    }
  }

  return ($value);
}

# Expand any wildcards such as %cc%, %assigned_to% etc into an array
# of email address string.
#
# On success, returns an arrayref of email address strings.
# On failure, returns undef.
#
# Don't clone $args->{values} as the clone() is done in _custom_search_match
sub _expand_wildcard {
  my $bug      = shift;
  my $wildcard = shift;

  my $result = undef;

  if ($wildcard =~ /^%(assigned_to|cc|qa_contact|docs_contact|reporter)%$/) {
    $result = $bug->$1;
    $result = [$result] if ($result && ref($result) ne 'ARRAY');
  }

  # Add Agile pool roles
  if ($wildcard =~ /^\%pool\.(.*)%$/ && $bug->pool) {
    my $role = $1;
    my $team = $bug->pool->team;

    push(@$result, @{$team->by_roles->{$role}})
      if (defined($team->by_roles->{$role}));
  }

  # Add Agile team roles
  if ($wildcard =~ /^\%team\.(.*)%$/) {
    my $role = $1;
    my $team = $bug->my_team();

    if ($team && defined($team->by_roles->{$role})) {
      push(@$result, @{$team->by_roles->{$role}});
    }
  }

  # Return undef if no match.
  return unless (defined $result);

  # Turn everything into email address strings, if it isn't already.
  my @emails = map { ref($_) eq 'Bugzilla::User' ? $_->login : $_ } @$result;

  return \@emails;
}

# Perform wildcard expansion (if any) prior to calling
# _custom_search_match_real

# WARNING: $args->{bug} is an OPTIONAL parameter.
# It will be null if the caller is the _flag_rule_check function

sub _custom_search_match {
  my $args   = shift;
  my $bug    = $args->{bug};
  my $op     = $args->{op};
  my $value  = $args->{value};     # What we are matching
  my $values = $args->{values};    # What the bug has
  my $type   = $args->{type};

  # Check to see if $value contains a wildcard.  If it does,
  # expand it and match on that.  Some wildcards may expand to
  # multiple values.  If they do, process all values and return
  # true on first match found.
  if ($value && $value =~ /^%(\w+)%$/) {
    my $users = _expand_wildcard($bug, $value);

    # If there is no user (e.g. qa_contact), then there is no match
    return 0 if not defined $users;

    foreach my $email (@$users) {
      my $result
        = _custom_search_match_real({
        bug => $bug, op => $op, value => $email, values => $values, type => $type,
        });

      # If there is a match or error, we can return that
      return $result if $result;
    }

    # There was no match
    return 0;
  }
  elsif ($value && $value =~ /%%(\w+)%%/) {
    $args->{value} = _expand_field_wildcards($bug, $value, $type);
  }

  # The normal case, no wildcards present.  perform normal matching
  return _custom_search_match_real($args);
}

# This determines whether $value $op @$values
# Returns true on match, false if no match.
# dies with an error string on an error.
#
# WARNING: $args->{bug} is an OPTIONAL parameter.
# It will be null if the caller is the _flag_rule_check function
sub _custom_search_match_real {
  my $args   = shift;
  my $bug    = $args->{bug};
  my $op     = $args->{op};
  my $value  = $args->{value};            # What we are matching
  my $values = clone($args->{values});    # What the bug has
  my $type   = $args->{type} || 'text';

  my $found = 0;

  bre_debug(
    "Checking if '" . ($value // "") . "' $op " . join(', ', (@$values) // []));

  # If $type is of type date, we need to create datetime object so we can
  # do the comparison.
  if ($type eq 'date') {
    my $dt_parser = DateTime::Format::DBI->new(Bugzilla->dbh);

    bre_debug(Dumper(@$values));

    # Turn everything into DateTime objects
    if ((scalar(@$values) != 1) || (defined($values->[0]))) {
      @$values = map { $dt_parser->parse_datetime($_) } @$values;
    }

    if ($value && $value eq 'now') {
      $value = DateTime->now;
    }
    elsif ($value && $value =~ /^([+-]?)\s*?(\d+)(\w)$/) {
      my $sign    = $1 // '+';
      my $period  = $2;
      my $step    = lc($3);
      my %wordmap = (qw(h hours d days w weeks m months y years));
      die("Unable to parse '$value' as a relative date (unknown character '$step')")
        unless $wordmap{$step};
      eval { $value = DateTime->now->add($wordmap{$step} => "$sign$period"); };
      die "Unable to parse date '$value'" if ($@);

    }
    elsif ($value) {
      my $date = $value;
      $value = datetime_from($value);
      die "Unable to parse date '$date'" unless $value;
    }
    else {
      $value = datetime_from('1900-01-01');
    }
  }

  if ($op eq 'equals') {
    $found
      = $type eq 'text'
      ? grep { $_ eq $value } @$values
      : grep { $_ == $value } @$values;
  }
  elsif ($op eq 'notequals') {
    $found
      = $type eq 'text'
      ? grep { $_ ne $value } @$values
      : grep { $_ != $value } @$values;
  }
  elsif ($op eq 'lessthan') {
    $found
      = $type eq 'text'
      ? grep { $_ lt $value } @$values
      : grep { $_ < $value } @$values;
  }
  elsif ($op eq 'lessthaneq') {
    $found
      = $type eq 'text'
      ? grep { $_ le $value } @$values
      : grep { $_ <= $value } @$values;
  }
  elsif ($op eq 'greaterthan') {
    $found
      = $type eq 'text'
      ? grep { $_ gt $value } @$values
      : grep { $_ > $value } @$values;
  }
  elsif ($op eq 'greaterthaneq') {
    $found
      = $type eq 'text'
      ? grep { $_ ge $value } @$values
      : grep { $_ >= $value } @$values;
  }
  elsif ($op eq 'regexp') {
    $found = grep { $_ =~ $value } @$values;
  }
  elsif ($op eq 'notregexp') {
    $found = grep { $_ !~ $value } @$values;
  }
  elsif ($op eq 'substring') {
    $found = grep { index(lc($_), lc($value)) != -1 } @$values;
  }
  elsif ($op eq 'notsubstring') {
    $found = grep { index(lc($_), lc($value)) == -1 } @$values;
  }
  elsif ($op eq 'anyexact') {
    my @splits = split /[\s,]/, $value;
    foreach my $string (@splits) {
      if (any { $string eq $_ } @$values) {
        $found = 1;
        last;
      }
    }
  }
  elsif ($op =~ /^(any|all|no)wordssubstr$/) {
    my @splits = map {lc} split /[,\s]+/, $value;
    my $count  = 0;
    foreach my $string (@splits) {
      ++$count if any { index(lc($_), $string) != -1 } @$values;
    }
    if ( ($op eq 'anywordssubstr' and $count != 0)
      || ($op eq 'allwordssubstr' and $count == scalar(@splits))
      || ($op eq 'nowordssubstr'  and $count == 0))
    {
      $found = 1;
    }
  }
  elsif ($op eq 'isempty') {
    if ($type eq 'text' && ref $values ne 'ARRAY') {
      $found = (length($values) == 0);
    }
    else {
      if (scalar(@$values) == 0) {
        $found = 1;
      }
      if (scalar(@$values) == 1) {
        if (!defined($values->[0]) || $values->[0] eq '' || $values->[0] eq '---') {
          $found = 1;
        }
      }
    }
  }
  elsif ($op eq 'isnotempty') {
    if ($type eq 'text' && ref $values ne 'ARRAY') {
      $found = (length($values) > 0);
    }
    else {
      if (scalar(@$values) > 1) {
        $found = 1;
      }
      if (scalar(@$values) == 1) {
        if (defined($values->[0]) && $values->[0] ne '' && $values->[0] ne '---') {
          $found = 1;
        }
      }
    }
  }
  else {
    # The op isn't valid. Return an error
    die "Don't know op '$op'";
  }

  if ($found > 1 && defined($bug)) {
    Bugzilla->logger->error("Got found value of $found on Bug "
        . $bug->id
        . " for match '$op' on '$value' against "
        . Dumper($values));
  }

  return $found;
}

sub _get_bug_value {
  my $self   = shift;
  my $bug    = shift;
  my $column = shift;

  my $sub  = BUG_FIELD_MAP->{$column}{func} || $column;
  my @keys = split '/', (BUG_FIELD_MAP->{$column}{key} || '');
  bre_debug("Calling get bug value column = '$column', sub = '$sub', key = '"
      . join(' / ', @keys)
      . "'");

  # Get the value and turn it into an array if it isn't already
  my $values = $bug->$sub;
  $values = [$values] if ref $values ne 'ARRAY';

  # Do we need to do anything to the value?
  # Some fields need to be handled specially
  if ($column eq 'flagtypes.name') {
    $values = [map { $_->type->name . $_->status } @$values];
  }
  elsif ($column eq 'work_time') {

    # Yes, return an array with one value (the total work time)
    $values = [sum 0, @$values];
  }
  elsif ($column eq 'longdescs.count') {

    # Yes, return an array with one value (the number of rows)
    return [scalar(@$values)];
  }
  elsif (scalar @keys) {

    # A lot of sub return an object, and we want a single value from it
    # Sometime we need to look at a value of an object of an object
    foreach my $key (@keys) {
      $values = [map { defined $_ ? $_->$key : undef } @$values];
    }
  }

  return $values;
}

sub _rule_do_action {
  my $self     = shift;
  my $bug      = shift;
  my $rule     = shift;
  my $notifies = shift;    # A reference to an array. Add to it if needed.

  my %set_all       = ();  # Values sent to Bugzilla::Bug->set_all
  my @set_flags     = ();  # Flag changes
  my %extra_changes = ();  # Things we need to add to $bug->update

  while (my ($field, $value) = each %{$rule->current->action}) {
    bre_debug("performing action '$field'");
    if ($field eq 'comment') {
      $extra_changes{comment_text}    = $value->{text};
      $extra_changes{comment_private} = $value->{private};
      $bug->add_comment($value->{text}, {isprivate => $value->{private}});
    }
    elsif ($field eq 'notify') {

      # Expand possible wildcards for the notify action.
      my @expanded_results;
      foreach my $email (@$value) {
        if ($email =~ /^%(\w+)%$/) {

          # This is a wildcard value, expand and add values.
          my $r = _expand_wildcard($bug, $email);
          push @expanded_results, @$r if defined $r;
        }
        else {
          # A regular email address.
          push @expanded_results, $email;
        }
      }

      # Strip out any duplicate email addresses, validate them against the
      # database and convert them back into plain emails.
      my @validated_emails
        = map { $_->login }
        grep  { defined $_ }
        map   { Bugzilla::User->new({name => $_}) } uniq @expanded_results;

      if (scalar(@validated_emails)) {
        $extra_changes{notify} = \@validated_emails;
        push @$notifies,
          {bug_id => $bug->id, rule_name => $rule->name, email => \@validated_emails};
      }
    }
    elsif ($field eq 'flag_types') {
      my $flag_changes = clone($value);
      my $flag_types   = $bug->flag_types;
      foreach my $flag_type (@$flag_types) {
        next unless $flag_changes->{$flag_type->name};
        my $info      = delete $flag_changes->{$flag_type->name};
        my $condition = $info->{condition};
        my $status    = $info->{status};
        my $requestee = $info->{requestee} || '';

        if ($requestee ne '') {
          my $r = _expand_wildcard($bug, $requestee);
          if ($r && @$r) {
            ## skip requesteeable flags if target is disabled
            my $r_user = Bugzilla::User->new({name => $r->[0]});
            next unless ($r_user->is_enabled);
            $requestee = $r->[0];
          }
        }

        if (scalar(@{$flag_type->{flags}}) == 0) {

          # The flag type isn't used by this bug. Create a new one
          if ($condition eq '' || $condition eq 'Exists' || index($condition, 'X') != -1)
          {
            push @set_flags,
              {type_id => $flag_type->id, status => $status, requestee => $requestee};
          }
        }
        else {
          # The flag type is used, update existing flags
          foreach my $flag (@{$flag_type->{flags}}) {
            if ( $condition eq ''
              || $condition eq 'Exists'
              || index($condition, $flag->status) != -1)
            {
              push @set_flags, {id => $flag->id, status => $status, requestee => $requestee};
            }
          }
        }
      }

      # Are there flags left? It means this product / component doesn't
      # offer this flag (or the rules engine can't see it)
      foreach my $flag_type (keys %$flag_changes) {
        unless ($flag_changes->{$flag_type}{ignore}) {
          kill_me($rule,
                "While running rule '"
              . $rule->name
              . "', I was unable to set flag $flag_type on bug "
              . $bug->id);
        }
      }
    }
    elsif ($field eq 'target_release') {

      # $value is an arrayref
      # We need to manually adjust the target_release values
      my $target_release = $bug->target_release;

      if ($value->{action} eq 'set') {
        $target_release = $value->{values};
      }
      elsif ($value->{action} eq 'add') {
        foreach my $v (@${$value->{values}}) {
          push @$target_release, $v unless any { $_ eq $v } @$target_release;
        }
      }
      elsif ($value->{action} eq 'remove') {
        foreach my $v (@{$value->{values}}) {
          @$target_release = grep { $_ ne $v } @$target_release;
        }
      }
      elsif ($value->{action} eq 'unset') {

        # Bugzilla::Bug::_check_target_release will set to default value
        $target_release = [];
      }
      $set_all{$field} = $target_release;

    }
    elsif ($field eq 'bug_group') {
      if ($value->{action} eq 'set') {

        # We need to manually adjust the groups if it set
        my @current_groups = map { $_->name } @{$bug->groups_in};
        my ($removed, $added) = diff_arrays(\@current_groups, $value->{values});
        $set_all{groups}{remove} = $removed if scalar(@$removed);
        $set_all{groups}{add}    = $added   if scalar(@$added);
      }
      elsif ($value->{action} eq 'unset') {
        $set_all{groups}{remove} = $bug->groups_in if scalar(@{$bug->groups_in});
      }
      else {
        # Add and remove is handled correctly by set_all
        $set_all{groups}{$value->{action}} = $value->{values};
      }
    }
    elsif (any { $field eq $_ } qw(assigned_to qa_contact docs_contact)) {
      if ($value->{action} eq 'reset to default') {
        my $assignee;

        if ($field eq 'assigned_to') {
          $assignee = $bug->component_obj->default_assignee;
        }
        elsif ($field eq 'qa_contact') {
          $assignee = $bug->component_obj->default_qa_contact;
        }
        elsif ($field eq 'docs_contact') {
          $assignee = $bug->component_obj->default_docs_contact;
        }
        else {
          kill_me($rule, "Unknown field type '$field' when reassigning bug");
        }

        $set_all{$field} = $assignee->login;
      }
      else {
        my $val   = $value->{values}->[0];
        my $users = _expand_wildcard($bug, $val);
        unless ($users) {
          kill_me($rule,
                "While running rule '"
              . $rule->name
              . "', I was unable to find any users in '$val' to use for $field on bug "
              . $bug->id);
        }
        my $login;
        foreach my $mail (@$users) {
          my $r_user = Bugzilla::User->new({name => $mail});
          next unless ($r_user->is_enabled);
          $login = $r_user->login;
          last;
        }
        if ($login) {
          $set_all{$field} = $login;
          Bugzilla->logger->info(Dumper($value, $set_all{$field}));
        }
        else {
          kill_me($rule,
                "While running rule '"
              . $rule->name
              . "', I was unable to find any active users in '$val' to use for $field on bug "
              . $bug->id);
        }
      }
    }
    elsif ($field eq 'bug_agile_pool.pool_id') {
      my $pool_id;

      if ($value->{action} eq 'reset to default') {
        $pool_id = $bug->default_pool_obj ? $bug->default_pool_obj->id : 0;
      }
      elsif ($value->{action} eq 'Remove') {
        $pool_id = 0;
      }
      else {
        $pool_id = $value->{values}->[0];
      }

      bre_debug("Setting pool_id to '$pool_id'");
      $set_all{'pool_id'} = $pool_id;
    }
    elsif (ref $value eq 'HASH') {

      # $value is an arrayref
      if ($value->{action} eq 'unset') {

        # Unset is special
        $set_all{$field}{set} = [];
      }
      else {
        my @results;
        foreach my $val (@{$value->{values}}) {
          my $users = _expand_wildcard($bug, $val);
          unless ($users) {
            bre_debug("There where no matches for wildcard: $val");
            push(@results, $val);
            next;
          }
          foreach my $mail (@$users) {
            my $r_user = Bugzilla::User->new({name => $mail});
            next unless ($r_user->is_enabled);
            push(@results, $r_user);
          }
        }
        $set_all{$field}{$value->{action}} = \@results;
      }
    }
    else {
      my $type;
      $type = BUG_FIELD_MAP->{$field}{type} if (defined(BUG_FIELD_MAP->{$field}));
      unless ($type) {
        my $field_object = Bugzilla::Field->new({name => $field});
        if ($field_object->type == Bugzilla::Constants::FIELD_TYPE_DATE) {
          $type = 'date';
        }
      }
      $type = 'text' unless ($type);

      bre_debug("value: $value, type: $type");
      $set_all{$field} = _expand_field_wildcards($bug, $value, $type);
      bre_debug("value 2: " . $set_all{$field});
    }
  }

  # Update values
  try {
    $bug->set_all(\%set_all);
  }
  catch {
    kill_me($rule,
          "While running rule '"
        . $rule->name
        . "' on bug "
        . $bug->id
        . ", an error occurred when trying to set a value: $_");
  };

  try {
    $bug->set_flags(\@set_flags, []);
  }
  catch {
    kill_me($rule,
          "While running rule '"
        . $rule->name
        . "' on bug "
        . $bug->id
        . ", an error occurred when trying to set the flags: $_");
  };

  # 'Commit' changes (we are still in an outer transaction though)
  $bug->{_rh_rule} = $rule;
  my $changes;
  try {
    $changes = $bug->update();
  }
  catch {
    # Check to see if a deadlock happened.
    if ( $_ =~ /Deadlock found when trying to get lock/
      || $_ =~ /deadlock detected/
      || $_ =~ /database reported a query serialization error/
      || $_ =~ /could not serialize access due to concurrent update/)
    {
      Bugzilla->logger->info('BRE: DB transient error when updating bug '
          . $bug->id
          . ' hit by rule \''
          . $rule->name
          . '\'');

      my $dbh = Bugzilla->dbh;
      $dbh->bz_rollback_transaction() if $dbh->bz_in_transaction();

      # Abort the run.
      exit 1;
    }
    else {
      # Not a deadlock exception - rethrow the error.
      kill_me($rule, "For Bug " . $bug->id . " error: " . $_);
    }
  };
  return ($changes, \%extra_changes);
}

sub _send_notifies {
  my $self     = shift;
  my $notifies = shift;    # Array of hashes, keys: bug_id, rule_name, email

  # The email client will display the Date: header in the desired timezone,
  # so we can always use UTC here.
  my $timestamp = Bugzilla->dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
  $timestamp = format_time($timestamp, '%a, %d %b %Y %T %z', 'UTC');

  foreach my $notify (@$notifies) {
    foreach my $email (uniq @{$notify->{email}}) {

      # Since these aren't going to users, we will send it in the sites language
      my $default_lang = Bugzilla::User->new()->setting('lang');
      my $template     = Bugzilla->template_inner($default_lang);

      my $message;
      my %vars = (to => $email, date => $timestamp, %$notify);
      $template->process("email/rule-notify.txt.tmpl", \%vars, \$message)
        || ThrowTemplateError($template->error());

      MessageToMTA($message);
    }
  }

  return;
}

sub _send_dryrun_feedback {
  my $self = shift;
  my $args = shift;

  my $who     = $args->{who};
  my $rule_id = $args->{rule_id};
  my $changes = $args->{changes};

  # Use the users preferred language
  my $to_user = Bugzilla::User->new($who);
  my $rule    = Bugzilla::Extension::RuleEngine::Rule->new($rule_id);
  my $vars    = {rule => $rule, to_user => $to_user, changes => $changes,};

  my $template = Bugzilla->template_inner($to_user->setting('lang'));
  my ($msg_text, $msg_html, $msg_header);
  $template->process("email/dryrun-header.txt.tmpl", $vars, \$msg_header)
    || ThrowTemplateError($template->error());
  $template->process("email/dryrun.txt.tmpl", $vars, \$msg_text)
    || ThrowTemplateError($template->error());

  my @parts = (Email::MIME->create(
    attributes => {content_type => "text/plain",},
    body       => $msg_text,
  ));

  if ($to_user->setting('email_format') eq 'html') {

    # Add rowspan to the $changes values
    foreach my $bug_id (keys %$changes) {
      my $bugcount = 0;
      foreach my $rule (@{$changes->{$bug_id}}) {
        my $count = scalar(keys %{$rule->{changes}});
        ++$count if $rule->{extra_changes}{comment_text};
        ++$count if $rule->{extra_changes}{notify};
        $rule->{rowspan} = $count;
        $bugcount += $count;
      }
      $vars->{rowspan}{$bug_id} = $bugcount || 1;
    }

    $template->process("email/dryrun.html.tmpl", $vars, \$msg_html)
      || ThrowTemplateError($template->error());
    push @parts,
      Email::MIME->create(
      attributes => {content_type => "text/html",},
      body       => $msg_html,
      );
  }

  # TT trims the trailing newline, and threadingmarker may be ignored.
  my $email = new Email::MIME("$msg_header\n");
  if (scalar(@parts) == 1) {
    $email->content_type_set($parts[0]->content_type);
  }
  else {
    $email->content_type_set('multipart/alternative');

    # Some mail clients need same encoding for each part, even empty ones.
    $email->charset_set('UTF-8') if Bugzilla->params->{'utf8'};
  }
  $email->parts_set(\@parts);
  MessageToMTA($email);

  return;
}

sub bre_debug {
  my $msg  = shift;
  my $line = defined caller() ? (caller())[2] : 'no line';
  Bugzilla->logger->debug("[BRE] ($line) $msg\n");

  return;
}

sub kill_me {
  my $rule    = shift;
  my $msg     = shift;
  my $no_exit = shift // 0;
  my $dbh     = Bugzilla->dbh;
  bre_debug('ENABLING KILL SWITCH');
  bre_debug($msg);

  # Firstly, we need to rollback any transaction we are in
  $dbh->bz_rollback_transaction() if $dbh->bz_in_transaction();

  # Now enable the kill switch
  my $args = {disabled_by => Bugzilla->user->id, reason => $msg,};

  $args->{rule_id} = $rule->id if defined $rule;

  Bugzilla::Extension::RuleEngine::RuleState->create($args);

  unless ($no_exit) {
    exit 1;
  }

  return;
}

1;
