# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::RuleDetail;

use strict;
use warnings;
use 5.10.1;

use Bugzilla::Constants;
use Bugzilla::Field;
use Bugzilla::Util;
use Bugzilla::Error;

use Bugzilla::Extension::RuleEngine::Rule;

use base qw(Bugzilla::Object);

use JSON;
use JSON::XS;

###############################
####    Initialization     ####
###############################

use constant DB_TABLE   => 'rh_rule_detail';
use constant LIST_ORDER => 'creation_ts, id';

use constant DB_COLUMNS => qw(
  id
  rule_id
  creation_ts
  creator_id
  match_json
  action_json
  isactive
);

# Nothing gets updated. Ever.
use constant UPDATE_COLUMNS => qw( );

use constant VALIDATORS => {
  creation_ts => \&_check_creation_ts,
  rule_id     => \&_check_rule_id,
  isactive    => \&Bugzilla::Object::check_boolean,
};

###############################
####      Validators       ####
###############################

sub _check_creation_ts {
  return Bugzilla->dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
}

sub _check_rule_id {
  my ($invocant, $id) = @_;
  $id = trim($id);

  Bugzilla::Extension::RuleEngine::Rule->check({id => $id});

  return $id;
}

###############################
####       Methods         ####
###############################

sub create {
  my $class  = shift;
  my $params = shift;

  $params->{match_json}  = to_json(delete $params->{match});
  $params->{action_json} = to_json(delete $params->{action});

  my $self = $class->SUPER::create($params);
  return $self;
}

sub rule {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{rule}) {
    $self->{rule} = Bugzilla::Extension::RuleEngine::Rule->new($self->rule_id);
  }
  return $self->{rule};
}

sub creator {
  my ($self) = @_;
  return $self->{creator} if exists $self->{creator};
  $self->{creator} = new Bugzilla::User($self->creator_id);
  return $self->{creator};
}

###############################
####      Accessors        ####
###############################

sub rule_id     { return $_[0]->{rule_id}; }
sub creation_ts { return $_[0]->{creation_ts}; }
sub creator_id  { return $_[0]->{creator_id}; }
sub match_json  { return $_[0]->{match_json}; }
sub action_json { return $_[0]->{action_json}; }
sub is_active   { return $_[0]->{isactive}; }

sub match {
  my $self = shift;

  if (!defined $self->{match}) {
    $self->{match} = from_json($self->match_json);
  }

  return $self->{match};
}

sub action {
  my $self = shift;

  if (!defined $self->{action}) {
    $self->{action} = from_json($self->action_json);
  }

  return $self->{action};
}

sub match_pretty {
  my $self = shift;

  my $json = JSON::XS->new->ascii->pretty->allow_nonref;
  $json->canonical(1);

  return ($json->encode($self->match));
}

sub action_pretty {
  my $self = shift;

  my $json = JSON::XS->new->ascii->pretty->allow_nonref;
  $json->canonical(1);

  return ($json->encode($self->action));
}

###############################
####       Helpers         ####
###############################

1;

__END__

=head1 NAME

Bugzilla::Extension::RuleEngine::RuleDetail - Bugzilla Rule Detail class.

=head1 SYNOPSIS

    use Bugzilla::Extension::RuleEngine::RuleDetail;

    my $rule_detail = new Bugzilla::Extension::RuleEngine::RuleDetail(1);

    my $id          = $rule_group->id;
    my $rule_id     = $rule_group->rule_id;
    my $creation_ts = $rule_group->creation_ts;
    my $is_active   = $rule_group->is_active;

=head1 DESCRIPTION

Bugzilla::Extension::RuleEngine::RuleDetail represents a version of a rule
object. It is an implementation of L<Bugzilla::Object>, and thus provides all
methods that L<Bugzilla::Object> provides.

The methods that are specific to L<Bugzilla::Extension::RuleEngine::RuleDetail> are listed
below.

=head1 METHODS

=over

=item C<creator>

=over

=item B<Description>

The person that created this version of the rule

=item B<Params>

none.

=item B<Returns>

A L<Bugzilla::User> object.

=back

=item C<rule>

=over

=item B<Description>

Returns the rule that this detail belongs to.

=item B<Params>

none.

=item B<Returns>

A L<Bugzilla::Extension::RuleEngine::Rule> object.

=back

=item C<match>

=over

=item B<Description>

Returns the details of the criteria to match a bug against
this version of a rule

=item B<Params>

none.

=item B<Returns>

A hash with the following keys:

=over

=item B<custom> (array of hashes)
Reflects custom searches. The keys will be:
negate (boolean): Whether the search is negated
field (string): The fielddefs.name being searched on (only selected values are allowed)
op (string): The operation to be performed. A sub set of the official list is supported
value (string): The value to search for. (can be empty if op is 'null' or 'notnull')

=item B<flag_types> (array of hashes)

Each hash has two keys name (string): the flag type name
status (array of strings): status to match for. One or more of: X, ?, + or -

=item B<flag_groups> (array of hashes)
Each hash has three keys id (integer): the flag group id word (string): 'any' or 'all'.
If all, all flags matches by the group must be set to one of the status(es) status
(array of strings): status to match for. One or more of: X, ?, + or -

=item B<anything else>
is a field name which is a hash with one or two keys>
values (array of strings): Values to match (e.g. product names)
word (string): One of 'is any of', 'is all of' or 'is not any of'. If not specified then 'is any of' is assumed.

=back

=back

=item C<action>

=over

=item B<Description>

Returns the details of what changes to make if the rule detail
is successfully matched

=item B<Params>

none.

=item B<Returns>

A hash with the following keys:

=over

=item B<bug_status / resolution / target_milestone / component> (string)
The new value for the respective field

=item B<keywords / target_release> (hash)

Two keys. action is 'set', 'add' or 'remove', values is an array of strings with the value(s)
to set, add or remove

=item B<flag_types> (hash of hashes)

The flags that need to be updated. The key is the flag type name, the value is a hash with two
keys, condition and status. condition is NULL for always change, 'Exists' to check that the
flag exists, or a combination of 'X?+-' to check the current status. status is the new status
of the flag.

=item B<comment> (hash)

A comment to be made. Keys are 'private' (1 or 0) and text (the text of the comment).

=item B<notify> (string)

The e-mail address(es) to send a notification to.

=back

=back

=back

=cut
