# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::WebService::ManageComponents;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::WebService::Util qw(validate);
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Object;
use Data::Dumper;
use Bugzilla::Extension::Workflows::ManageComponents;
use List::MoreUtils qw(uniq);
use List::Compare;

use constant PUBLIC_METHODS => qw(
  update_components
  get_components
);


sub update_components {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $dbh = Bugzilla->dbh;

  my $comps = $args->{components}
    // ThrowCodeError('param_required', {param => 'components'});

  my @results;

  $dbh->bz_start_transaction();

  foreach my $comp_hash (@$comps) {
    my $type = trim($comp_hash->{'type'} || 'component');

    my $class = 'Bugzilla::Component';
    $class = 'Bugzilla::Extension::SubComponents::RHSubComponent'
      if ($type eq 'sub_component');

    my $component;
    if (defined($comp_hash->{name})) {
      $component = $class->check({name => $comp_hash->{name}});
    }
    elsif (defined($comp_hash->{id})) {
      $component = $class->check({id => $comp_hash->{id}});
    }
    else {
      ThrowCodeError('params_required', {params => ['name', 'id']});
    }

    $user->check_can_manage_comp($component);

    $component->set_default_assignee(trim($comp_hash->{'default_assignee'}))
      if (defined $comp_hash->{'default_assignee'});

    $component->set_default_qa_contact(trim($comp_hash->{'default_qa_contact'}))
      if (defined $comp_hash->{'default_qa_contact'});

    $component->set_default_docs_contact(trim($comp_hash->{'default_docs_contact'}))
      if (defined $comp_hash->{'default_docs_contact'});

    $component->set_description(trim($comp_hash->{'description'}))
      if (defined $comp_hash->{'description'});

    $component->set_default_pool(trim($comp_hash->{'default_pool'}))
      if (defined $comp_hash->{'default_pool'});

    $component->set_agile_team(trim($comp_hash->{'agile_team'}))
      if (defined $comp_hash->{'agile_team'});

    if (defined $comp_hash->{'initial_cc'}) {
      my @initial_cc;
      if (ref $comp_hash->{'initial_cc'} eq 'HASH') {
        my %hash = %{$comp_hash->{'initial_cc'}};

        if (exists $hash{'set'}) {
          @initial_cc = split ',\s*', $hash{'set'};
        }

        if (exists $hash{'add'}) {
          my @current_cc = split ',\s*', $component->cc_list_string;
          my @add_ccs    = split ',\s*', $hash{'add'};
          @initial_cc = uniq(@current_cc, @add_ccs);
        }

        if (exists $hash{'remove'}) {
          my @current_cc = split ',\s*', $component->cc_list_string;
          my @rem_ccs    = split ',\s*', $hash{'remove'};
          my $lc = List::Compare->new(\@current_cc, \@rem_ccs);
          @initial_cc = $lc->get_unique();
        }
      }
      else {
        @initial_cc = split ',\s*', $comp_hash->{'initial_cc'};
      }

      $component->set_cc_list(\@initial_cc);
    }

    if (defined $comp_hash->{'template'}) {
      my $bvp_template_id = trim($comp_hash->{'template'});
      $component->set_template_id($bvp_template_id)
        if ($component->can('set_template_id'));
    }

    my $returned_changes = $component->update();

    push(@results, _comp_2_hash($component, $user)) if ($returned_changes);
  }

  $dbh->bz_commit_transaction();

  return ({components => \@results});
}

# This also does sub_components
sub _comp_2_hash {
  my $component = shift;
  my $user      = shift;

  my %out = (
    id                 => $component->id,
    name               => $component->full_name_pretty,
    type               => $component->type,
    description        => $component->description,
    default_assignee   => $component->default_assignee->login,
    initial_cc         => $component->cc_list_string,
    default_pool       => '',
    default_qa_contact => (
      defined $component->default_qa_contact ? $component->default_qa_contact->login
      : ''
    ),
    default_docs_contact => (
      defined $component->default_docs_contact
      ? $component->default_docs_contact->login
      : ''
    ),
  );

  $out{'default_pool'}
    = defined $component->default_pool_obj
    && $component->default_pool_obj->team->user_can_view($user)
    ? _pool_2_hash($component->default_pool_obj)
    : '';

  $out{'agile_team'}
    = defined $component->my_team
    ? _team_2_hash($component->my_team, $component->agile_team)
    : '';

  $out{'can_manage'} = $user->can_manage_comp($component) if ($user);

  # sub components don't have templates
  $out{'template'}
    = $component->can('template_name') ? $component->template_name : '';

  $out{'product'}
    = $component->can('product')
    ? $component->product->name
    : $component->component->product->name;

  return (\%out);
}

sub _pool_2_hash {
  my $pool = shift;

  my %hash = (
    id   => $pool->id,
    name => $pool->name,
    team => {id => $pool->team->id, name => $pool->team->name,}
  );

  return \%hash;
}

sub _team_2_hash {
  my $team       = shift;
  my $agile_team = shift;

  my %hash = (
    id     => $team->id,
    name   => $team->name,
    direct => ($agile_team ? $team->id == $agile_team : 0),
  );

  return \%hash;
}

sub get_components {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $dbh = Bugzilla->dbh;
  my (@results, @components);
  my %params;
  my @valid_args = qw/pools teams show_nested product/;

  foreach my $arg (@valid_args) {
    if (defined($args->{$arg})) {
      my $val = $args->{$arg};
      trick_taint($val) unless (ref $val);
      $params{$arg} = $val;
    }
  }

  component_search(\%params, \@components);

  foreach my $comp (@components) {
    push(@results, _comp_2_hash($comp, $user));
  }

  return ({components => \@results});
}

1;

__END__

=head1 NAME

Bugzilla::Extension::Workflows::WebService::ManageComponents - API for handling
Administration actions for the L<Bugzilla::Extension::Workflows::ManageComponents> extension.

=head1 DESCRIPTION

This part of the Bugzilla API allows you to manage roles on, and basic data for, components and sub_components.

=head1 METHODS

See L<Bugzilla::WebService> for a description of what B<STABLE>, B<UNSTABLE>,
and B<EXPERIMENTAL> mean, and for more description about error codes.

=head2 get_components B<EXPERIMENTAL>

=head3 Description

Search for components and sub_components using the supplied parameters.

=head3 Params

pools:
  B<array> of B<strings> of pool names to search on.
teams:
  B<array> of B<strings> of team names to search on.
show_nested:
  B<boolean> Include all sub_components when a parent matches.
product:
  B<string> Product name to search on.

=head4 Example call

Get all the components my teams owns, including any sub_components owned by
other teams, or no teams.

  {
    "method":"ManageComponents.get_components",
    "params":[
      {
        "teams": ["My first team", "My second team"],
        "show_nested": 1
      }
    ]
  }

=head3 Returns

An array of hashes of components and sub_components.

  "components": [
    {
      "can_manage": 1,
      "default_assignee": "fu@example.com",
      "default_docs_contact": "",
      "default_pool": {
        "id": 2
        "name": "A backlog",
        "team": {
          "id": 1,
          "name": "A Team"
        },
      },
      "default_qa_contact": "qe@example.com",
      "description": "This is a very important component",
      "id": 17
      "initial_cc": "",
      "name": "my_component",
      "template": "Default",
      "type": "component",
    }
  ]

=head2 update_components B<EXPERIMENTAL>

=head3 Description

Update components and sub components.

=head3 Params

This API expects and array , named components, containg a hash for each component
or sub component you wish to edit. Fields that are absent from the has are not
updated.

name:
  B<string> The name of the component being modified. Either this or ID must be
  set.
id:
  B<int> The ID of the component being modified. Either this or name must be set.
  For sub_components it is best to use the ID as names are not unique.
type:
  B<string> What type of component is is, 'component' or 'sub_component'.
  Defaults to 'component'.
default_assignee:
  B<string> The login name of the new default assignee.
default_qa_contact:
  B<string> The login name of the new default qa contact.
default_docs_contact:
  B<string> The login name of the new default docs contact.
description:
  B<string> The new description for the component.
template:
  B<int|string> The name or ID of the comment template to use on the new bug page.
default_pool:
  B<int|string> The name or ID of the new default pool for this component.
agile_team:
  B<int|string> The name or ID of the agile team to use as the sub-team for this
  component.
initial_cc:
  Either B<string> A comma separated list of user logins to add to the CCs when a new bug
  is opened for this component.
  Or a hash with the same in a key names "add", "set", or "remove".

=head4 Example call

Change the default assignee for a component.

  {
    "method": "ManageComponents.update_components",
    "params": [
      {
        "components": [
          {
            "default_assignee": "dave@example.com",
            "id": 11,
            "type": "component"
          }
        ]
      }
    ]
  }

=head3 Returns

An array of hashes of components and sub_components that were modified. If a
supplied component was not updated then it will not be returned.

  "components": [
    {
      "agile_team": {
        "id": 2,
        "name": "B Team"
      }
      "can_manage": 1,
      "default_assignee": "fu@example.com",
      "default_docs_contact": "",
      "default_pool": {
        "id": 2
        "name": "A backlog",
        "team": {
          "id": 1,
          "name": "A Team"
        },
      },
      "default_qa_contact": "qe@example.com",
      "description": "This is a very important component",
      "id": 17
      "initial_cc": "",
      "name": "my_component",
      "template": "Default",
      "type": "component",
    }
  ]

=cut
